#include <string.h>
#include "config-cgroups.h"
#include "spec-cgroup.h"
#include "error.h"
#include "client-message.h"

typedef struct _ConfigCgroupDialog {
    Conc *conc;
    gchar *name; /* NULL if adding new group */
    GladeXML *xml; /* NULL if removing group */
    GtkTreeView *treeview;
    GtkListStore *members;
} ConfigCgroupDialog;

enum {
    COLUMN_MEMBER,
    COLUMN_LAST
};

/*
 * Glade widget names:
 */
#define CONFIG_CGROUP_DIALOG			"config cgroup dialog"
#define CONFIG_CGROUP_DIALOG_NAME		"ccgdname"
#define CONFIG_CGROUP_DIALOG_MEMBERS		"ccgdmembers"
#define CONFIG_CGROUP_DIALOG_CONSOLE		"ccgdconsole"
#define CONFIG_CGROUP_DIALOG_ADD		"ccgdadd"
#define CONFIG_CGROUP_DIALOG_REMOVE		"ccgdremove"

/*
 * Glade handler names:
 */
#define CONFIG_CGROUP_DIALOG_RESPONSE		"response"
#define CONFIG_CGROUP_DIALOG_MEMBERS_CHANGED	"members changed"
#define CONFIG_CGROUP_DIALOG_CONSOLE_CHANGED	"console changed"
#define CONFIG_CGROUP_DIALOG_ADD_CONSOLE	"add console"
#define CONFIG_CGROUP_DIALOG_REMOVE_CONSOLE	"remove console"

static void
config_cgroups_dialog_destroy(ConfigCgroupDialog *cgd) {
    if (cgd->xml) {
	gtk_widget_destroy(glade_xml_get_widget(cgd->xml, CONFIG_CGROUP_DIALOG));
    }
    if (cgd->name) {
	g_free(cgd->name);
    }
    if (cgd->members) {
	g_object_unref(cgd->members);
    }
    g_free(cgd);
}

static void
config_cgroups_done(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    ConfigCgroupDialog *cgd = data;
    GError *e = message_arg_error_check(arg);
    gchar *failed_title = _("Failed to update console group");
    gchar *failed_text = _("Could not modify console group (%s)");

    if (cgd->xml == NULL) {
	failed_title = _("Failed to remove console group");
	failed_text = _("Could not remove console group (%s)");
    }
    if (cgd->name == NULL) {
	failed_title = _("Failed to add console group");
	failed_text = _("Could not add console group (%s)");
    }
    if (e != NULL) {
	warning_dialog(cgd->conc, failed_title, failed_text, e->message);
	g_clear_error(&e);
	if (cgd->xml != NULL) {
	    return;
	}
    }
    config_cgroups_dialog_destroy(cgd);
}


static void
config_cgroups_dialog_response(GtkDialog *dialog,
	gint arg1,
	gpointer user_data) {
    ConfigCgroupDialog *cgd = user_data;

    if (arg1 == GTK_RESPONSE_OK) {
	const gchar *name;
	gchar *command;
	gchar *member;
	SpecCgroup *sg;
	MessageArg *a;
	GError *e = NULL;
	GtkTreeIter iter;

	name = gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(cgd->xml,
			CONFIG_CGROUP_DIALOG_NAME)));
	if (name == NULL || strlen(name) == 0) {
	     warning_dialog(cgd->conc, _("Cannot add console group"),
		     _("No name was given"));
	     return;
	}
	sg = g_new0(struct _SpecCgroup, 1);
	sg->name = g_strdup(name);
	sg->consoles = NULL;
	if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(cgd->members),
		    &iter) == TRUE) {
	    do {
		gtk_tree_model_get(GTK_TREE_MODEL(cgd->members), &iter,
			COLUMN_MEMBER, &member, -1);
		sg->consoles = g_list_prepend(sg->consoles, member);
	    } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(cgd->members),
			&iter) == TRUE);
	}
	a = spec_cgroup_message_arg("spec", sg);
	spec_cgroup_free(sg);

	if (cgd->name == NULL) {
	    command = CLIENT_MESSAGE_CONFIG_CGROUP_ADD;
	} else {
	    MessageArg *spec = a;

	    command = CLIENT_MESSAGE_CONFIG_CGROUP_EDIT;
	    a = message_arg_list("", NULL);
	    a->list = g_list_prepend(a->list,
		    message_arg_string("name", cgd->name));
	    a->list = g_list_prepend(a->list, spec);
	}
	if (message_send(cgd->conc->msg_q, command, a, config_cgroups_done,
		    cgd, &e) == FALSE) {
	    error_dialog(cgd->conc, _("Internal error"),
		    _("Failed to send message (%s)"), e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
	return;
    }
    config_cgroups_dialog_destroy(cgd);
}

static void
config_cgroups_members_changed(GtkTreeView *treeview,
	gpointer user_data) {
    ConfigCgroupDialog *cgd = user_data;
    GtkTreeIter iter;
    gboolean sensitive = FALSE;

    if (gtk_tree_selection_get_selected(
		gtk_tree_view_get_selection(cgd->treeview), NULL, &iter)
	    == TRUE) {
	sensitive = TRUE;
    }
    g_object_set(glade_xml_get_widget(cgd->xml, CONFIG_CGROUP_DIALOG_REMOVE),
	    "sensitive", sensitive, NULL);
}

static void
config_cgroups_console_changed(GtkComboBox *combobox,
	gpointer user_data) {
    ConfigCgroupDialog *cgd = user_data;
    gchar *selection;
    GtkTreeIter iter;
    gchar *member;
    gboolean sensitive = TRUE;

    selection = get_combobox_text(cgd->xml, CONFIG_CGROUP_DIALOG_CONSOLE,
	    GTK_TREE_MODEL(cgd->conc->all_consoles), CONSOLES_COLUMN_NAME);
    if (selection != NULL &&
	    gtk_tree_model_get_iter_first(GTK_TREE_MODEL(cgd->members),
		&iter) == TRUE) {
	do {
	    gtk_tree_model_get(GTK_TREE_MODEL(cgd->members), &iter,
		    COLUMN_MEMBER, &member, -1);
	    if (strcmp(selection, member) == 0) {
		sensitive = FALSE;
		g_free(member);
		break;
	    }
	    g_free(member);
	} while (gtk_tree_model_iter_next(GTK_TREE_MODEL(cgd->members),
		    &iter) == TRUE);
    }
    if (selection == NULL) {
	sensitive = FALSE;
    }
    g_object_set(glade_xml_get_widget(cgd->xml, CONFIG_CGROUP_DIALOG_ADD),
	    "sensitive", sensitive, NULL);
    if (selection != NULL) {
	g_free(selection);
    }
}

static void
config_cgroups_console_remove(GtkButton *button,
	gpointer user_data) {
    ConfigCgroupDialog *cgd = user_data;
    GtkTreeIter iter;

    if (gtk_tree_selection_get_selected(
		gtk_tree_view_get_selection(cgd->treeview), NULL, &iter)
	    == TRUE) {
	gtk_list_store_remove(cgd->members, &iter);
    }
    config_cgroups_members_changed(NULL, cgd);
    config_cgroups_console_changed(NULL, cgd);
}

static void
config_cgroups_console_add(GtkButton *button,
	gpointer user_data) {
    ConfigCgroupDialog *cgd = user_data;
    gchar *selection;
    GtkTreeIter iter;

    selection = get_combobox_text(cgd->xml, CONFIG_CGROUP_DIALOG_CONSOLE,
	    GTK_TREE_MODEL(cgd->conc->all_consoles), CONSOLES_COLUMN_NAME);
    if (selection == NULL) {
	return;
    }
    gtk_list_store_append(cgd->members, &iter);
    gtk_list_store_set(cgd->members, &iter,
	    COLUMN_MEMBER, selection, -1);
    config_cgroups_members_changed(NULL, cgd);
    config_cgroups_console_changed(NULL, cgd);
}

static void
config_cgroups_dialog(ConfigCgroupDialog *cgd) {
    GtkCellRenderer *text_rend;
    GtkTreeViewColumn *col;

    cgd->xml = glade_xml_new(cgd->conc->glade_file, CONFIG_CGROUP_DIALOG, NULL);
    if (cgd->name != NULL) {
	gchar *title = g_strdup_printf(_("Edit Console Group: %s"), cgd->name);

	g_object_set(glade_xml_get_widget(cgd->xml, CONFIG_CGROUP_DIALOG),
		"title", title, NULL);
	g_free(title);
    }
    text_rend = gtk_cell_renderer_text_new();
    g_object_set(glade_xml_get_widget(cgd->xml, CONFIG_CGROUP_DIALOG_CONSOLE),
	    "model", cgd->conc->all_consoles, NULL);
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(glade_xml_get_widget(cgd->xml,
		    CONFIG_CGROUP_DIALOG_CONSOLE)), text_rend, TRUE);
    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(glade_xml_get_widget(cgd->xml,
		    CONFIG_CGROUP_DIALOG_CONSOLE)), text_rend, "text",
	    CONSOLES_COLUMN_NAME);
    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(cgd->members),
	    COLUMN_MEMBER, GTK_SORT_ASCENDING);
    cgd->treeview = GTK_TREE_VIEW(glade_xml_get_widget(cgd->xml,
		CONFIG_CGROUP_DIALOG_MEMBERS));
    g_object_set(cgd->treeview,
	    "model", cgd->members,
	    "rules-hint", TRUE,
	    "headers-clickable", TRUE,
	    "enable-search", TRUE,
	    "search-column", COLUMN_MEMBER,
	    NULL);
    col = gtk_tree_view_column_new_with_attributes("Console", text_rend,
	    "text", COLUMN_MEMBER, NULL);
    gtk_tree_view_append_column(cgd->treeview, col);
    g_object_set(glade_xml_get_widget(cgd->xml, CONFIG_CGROUP_DIALOG_ADD),
	    "sensitive", FALSE, NULL);
    g_object_set(glade_xml_get_widget(cgd->xml, CONFIG_CGROUP_DIALOG_REMOVE),
	    "sensitive", FALSE, NULL);
    glade_xml_signal_connect_data(cgd->xml,
	    CONFIG_CGROUP_DIALOG_CONSOLE_CHANGED,
	    (GCallback)config_cgroups_console_changed, cgd);
    glade_xml_signal_connect_data(cgd->xml,
	    CONFIG_CGROUP_DIALOG_ADD_CONSOLE,
	    (GCallback)config_cgroups_console_add, cgd);
    glade_xml_signal_connect_data(cgd->xml,
	    CONFIG_CGROUP_DIALOG_MEMBERS_CHANGED,
	    (GCallback)config_cgroups_members_changed, cgd);
    glade_xml_signal_connect_data(cgd->xml,
	    CONFIG_CGROUP_DIALOG_REMOVE_CONSOLE,
	    (GCallback)config_cgroups_console_remove, cgd);
    gtk_combo_box_set_active(GTK_COMBO_BOX(glade_xml_get_widget(cgd->xml,
		    CONFIG_CGROUP_DIALOG_CONSOLE)), 0);
    glade_xml_signal_connect_data(cgd->xml,
	    CONFIG_CGROUP_DIALOG_RESPONSE,
	    (GCallback)config_cgroups_dialog_response, cgd);
}

static void
make_list(ConfigCgroupDialog *cgd,
	GList *members) {
    GList *l;
    GtkTreeIter iter;

    cgd->members = gtk_list_store_new(COLUMN_LAST, G_TYPE_STRING);
    for (l = members; l != NULL; l = g_list_next(l)) {
	gtk_list_store_append(cgd->members, &iter);
	gtk_list_store_set(cgd->members, &iter,
		COLUMN_MEMBER, l->data, -1);
    }
}

extern void
config_cgroups_add(Conc *conc) {
    ConfigCgroupDialog *cgd = g_new0(struct _ConfigCgroupDialog, 1);

    cgd->conc = conc;
    cgd->name = NULL;
    make_list(cgd, NULL);
    config_cgroups_dialog(cgd);
}

static void
config_cgroups_edit_data(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    ConfigCgroupDialog *cgd = data;
    GError *e = message_arg_error_check(arg);
    SpecCgroup *sg;

    if (e != NULL) {
	warning_dialog(cgd->conc, _("Failed to edit console group"),
		_("Could not edit console group (%s)"), e->message);
	g_clear_error(&e);
	config_cgroups_dialog_destroy(cgd);
	return;
    }
    sg = spec_cgroup_new_from_arg(arg);
    g_assert(sg);
    make_list(cgd, sg->consoles);
    config_cgroups_dialog(cgd);
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(cgd->xml,
		    CONFIG_CGROUP_DIALOG_NAME)), sg->name);
    spec_cgroup_free(sg);
}


extern void
config_cgroups_edit(Conc *conc,
	const gchar *name) {
    ConfigCgroupDialog *cgd = g_new0(struct _ConfigCgroupDialog, 1);
    MessageArg *a;
    GError *e = NULL;

    cgd->conc = conc;
    cgd->name = g_strdup(name);
    cgd->xml = NULL;
    cgd->members = NULL;
    a = message_arg_string("name", cgd->name);
    if (message_send(cgd->conc->msg_q, CLIENT_MESSAGE_CONFIG_CGROUP_SPEC,
		a, config_cgroups_edit_data, cgd, &e) == FALSE) {
	error_dialog(cgd->conc, _("Internal error"),
		_("Failed to send message (%s)"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
}


extern void
config_cgroups_remove(Conc *conc,
	const gchar *name) {
    ConfigCgroupDialog *cgd = g_new0(struct _ConfigCgroupDialog, 1);
    MessageArg *a;
    GError *e = NULL;

    cgd->conc = conc;
    cgd->name = g_strdup(name);
    cgd->xml = NULL;
    cgd->members = NULL;
    a = message_arg_string("name", name);
    if (message_send(cgd->conc->msg_q, CLIENT_MESSAGE_CONFIG_CGROUP_REMOVE,
		a, config_cgroups_done, cgd, &e) == FALSE) {
	error_dialog(cgd->conc, _("Internal error"),
		_("Failed to send message (%s)"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
}

