#include <string.h>
#include "config-servers.h"
#include "spec-server.h"
#include "error.h"
#include "client-message.h"

typedef struct _ConfigServerDialog {
    Conc *conc;
    gchar *name; /* NULL if adding new group */
    GladeXML *xml; /* NULL if removing group */
} ConfigServerDialog;

/*
 * Glade widget names:
 *
 */
#define CONFIG_SERVER_DIALOG			"config server dialog"
#define CONFIG_SERVER_DIALOG_NAME		"csdname"
#define CONFIG_SERVER_DIALOG_TYPE		"csdtype"
#define CONFIG_SERVER_DIALOG_EXEC		"csdexec"
#define CONFIG_SERVER_DIALOG_EXEC_PATH		"csdexec path"
#define CONFIG_SERVER_DIALOG_CONNECT		"csdconnect"
#define CONFIG_SERVER_DIALOG_CONNECT_HOST	"csdconnect host"
#define CONFIG_SERVER_DIALOG_CONNECT_PORT	"csdconnect port"
#define CONFIG_SERVER_DIALOG_SSH		"csdssh"
#define CONFIG_SERVER_DIALOG_SSH_HOST		"csdssh host"
#define CONFIG_SERVER_DIALOG_SSH_USER		"csdssh user"
#define CONFIG_SERVER_DIALOG_SSH_COMMAND	"csdssh command"
#define CONFIG_SERVER_DIALOG_SSH_IDENTITY	"csdssh identity"

/*
 * Glade handler names:
 *
 */
#define CONFIG_SERVER_DIALOG_RESPONSE		"response"
#define CONFIG_SERVER_DIALOG_TYPE_CHANGED	"type changed"

static void
config_servers_dialog_destroy(ConfigServerDialog *csd) {
    if (csd->xml) {
	gtk_widget_destroy(glade_xml_get_widget(csd->xml, CONFIG_SERVER_DIALOG));
    }
    if (csd->name) {
	g_free(csd->name);
    }
    g_free(csd);
}

static void
config_servers_done(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    ConfigServerDialog *csd = data;
    GError *e = message_arg_error_check(arg);
    gchar *failed_title = _("Failed to update server");
    gchar *failed_text = _("Could not modify server (%s)");

    if (csd->xml == NULL) {
	failed_title = _("Failed to remove server");
	failed_text = _("Could not remove server (%s)");
    }
    if (csd->name == NULL) {
	failed_title = _("Failed to add server");
	failed_text = _("Could not add server (%s)");
    }
    if (e != NULL) {
	warning_dialog(csd->conc, failed_title, failed_text, e->message);
	g_clear_error(&e);
	if (csd->xml != NULL) {
	    return;
	}
    }
    config_servers_dialog_destroy(csd);
}

static ServerType
config_servers_type(ConfigServerDialog *csd) {
    gchar *selection;

    selection = get_combobox_text(csd->xml, CONFIG_SERVER_DIALOG_TYPE,
	    NULL, 0);
    if (selection == NULL) {
	return SERVER_TYPE_EXEC;
    }
    if (strcmp(selection, "Connect") == 0) {
	g_free(selection);
	return SERVER_TYPE_CONNECT;
    }
    if (strcmp(selection, "SSH") == 0) {
	g_free(selection);
	return SERVER_TYPE_SSH;
    }
    g_free(selection);
    return SERVER_TYPE_EXEC;
}

static void
config_servers_dialog_response(GtkDialog *dialog,
	gint arg1,
	gpointer user_data) {
    ConfigServerDialog *csd = user_data;

    if (arg1 == GTK_RESPONSE_OK) {
	const gchar *name = NULL;
	const gchar *path = NULL;
	const gchar *host = NULL;
	guint port = 0;
	const gchar *user = NULL;
	const gchar *command = NULL;
	const gchar *identity = NULL;
	ServerType type;
	SpecServer *ss;
	MessageArg *a;
	GError *e = NULL;

	name = gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
			CONFIG_SERVER_DIALOG_NAME)));
	if (name == NULL || strlen(name) == 0) {
	     warning_dialog(csd->conc, _("Cannot add server"),
		     _("No name was given"));
	     return;
	}
	type = config_servers_type(csd);
	switch (type) {
	    case SERVER_TYPE_EXEC:
		path =
		    gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
				   CONFIG_SERVER_DIALOG_EXEC_PATH))); 
		break;
	    case SERVER_TYPE_CONNECT:
		host =
		    gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
				   CONFIG_SERVER_DIALOG_CONNECT_HOST))); 
		if (host == NULL || strlen(host) == 0) {
		    warning_dialog(csd->conc, _("Cannot add server"),
			    _("No host name was given"));
		    return;
		}
		port = g_strtod(gtk_entry_get_text(
			    GTK_ENTRY(glade_xml_get_widget(csd->xml,
				    CONFIG_SERVER_DIALOG_CONNECT_PORT))), NULL);
		break;
	    case SERVER_TYPE_SSH:
		host =
		    gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
				   CONFIG_SERVER_DIALOG_SSH_HOST))); 
		if (host == NULL || strlen(host) == 0) {
		    warning_dialog(csd->conc, _("Cannot add server"),
			    _("No host name was given"));
		    return;
		}
		user =
		    gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
				   CONFIG_SERVER_DIALOG_SSH_USER))); 
		command =
		    gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
				   CONFIG_SERVER_DIALOG_SSH_COMMAND))); 
		identity =
		    gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
				   CONFIG_SERVER_DIALOG_SSH_IDENTITY))); 
		break;
	    default:
		g_assert_not_reached();
	}
	ss = g_new0(struct _SpecServer, 1);
	ss->name = g_strdup(name);
	ss->type = type;
	switch (ss->type) {
	    case SERVER_TYPE_EXEC:
		ss->exec.path = g_strdup(path?path:"");
		break;
	    case SERVER_TYPE_CONNECT:
		ss->connect.host = g_strdup(host);
		ss->connect.port = port;
		break;
	    case SERVER_TYPE_SSH:
		ss->connect.host = g_strdup(host);
		ss->ssh.ssh_program = g_strdup("ssh");
		ss->ssh.ssh_user = g_strdup(user?user:"");
		ss->ssh.ssh_identity = g_strdup(identity?identity:"");
		ss->ssh.remote_cmd = g_strdup(command?command:"");
		break;
	    default:
		g_assert_not_reached();
	}
	a = spec_server_message_arg("spec", ss);
	spec_server_free(ss);

	if (csd->name == NULL) {
	    command = CLIENT_MESSAGE_CONFIG_SERVER_ADD;
	} else {
	    MessageArg *spec = a;

	    command = CLIENT_MESSAGE_CONFIG_SERVER_EDIT;
	    a = message_arg_list("", NULL);
	    a->list = g_list_prepend(a->list, spec);
	    a->list = g_list_prepend(a->list,
		    message_arg_string("name", csd->name));
	}
	if (message_send(csd->conc->msg_q, command, a, config_servers_done,
		    csd, &e) == FALSE) {
	    error_dialog(csd->conc, _("Internal error"),
		    _("Failed to send message (%s)"), e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
	return;
    }
    config_servers_dialog_destroy(csd);
}

static void
config_servers_type_changed(GtkComboBox *combobox,
	gpointer user_data) {
    ConfigServerDialog *csd = user_data;
    gboolean exec_visible = FALSE;
    gboolean connect_visible = FALSE;
    gboolean ssh_visible = FALSE;
    GtkWidget *w;

    switch(config_servers_type(csd)) {
	case SERVER_TYPE_EXEC:
	    exec_visible = TRUE;
	    break;
	case SERVER_TYPE_CONNECT:
	    connect_visible = TRUE;
	    break;
	case SERVER_TYPE_SSH:
	    ssh_visible = TRUE;
	    break;
	default:
	    g_assert_not_reached();
    }
    w = glade_xml_get_widget(csd->xml, CONFIG_SERVER_DIALOG_EXEC);
    if (exec_visible) {
	gtk_widget_show_all(w);
    } else {
	gtk_widget_hide_all(w);
    }
    w = glade_xml_get_widget(csd->xml, CONFIG_SERVER_DIALOG_CONNECT);
    if (connect_visible) {
	gtk_widget_show_all(w);
    } else {
	gtk_widget_hide_all(w);
    }
    w = glade_xml_get_widget(csd->xml, CONFIG_SERVER_DIALOG_SSH);
    if (ssh_visible) {
	gtk_widget_show_all(w);
    } else {
	gtk_widget_hide_all(w);
    }
}

static void
config_servers_dialog(ConfigServerDialog *csd) {
    csd->xml = glade_xml_new(csd->conc->glade_file, CONFIG_SERVER_DIALOG, NULL);
    if (csd->name != NULL) {
	gchar *title = g_strdup_printf(_("Edit Server: %s"), csd->name);

	g_object_set(glade_xml_get_widget(csd->xml, CONFIG_SERVER_DIALOG),
		"title", title, NULL);
	g_free(title);
    }
    glade_xml_signal_connect_data(csd->xml,
	    CONFIG_SERVER_DIALOG_TYPE_CHANGED,
	    (GCallback)config_servers_type_changed, csd);
    gtk_combo_box_set_active(GTK_COMBO_BOX(glade_xml_get_widget(csd->xml,
		    CONFIG_SERVER_DIALOG_TYPE)), 0);
    glade_xml_signal_connect_data(csd->xml,
	    CONFIG_SERVER_DIALOG_RESPONSE,
	    (GCallback)config_servers_dialog_response, csd);
}

extern void
config_servers_add(Conc *conc) {
    ConfigServerDialog *csd = g_new0(struct _ConfigServerDialog, 1);

    csd->conc = conc;
    csd->name = NULL;
    config_servers_dialog(csd);
}

static void
config_servers_edit_data(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    ConfigServerDialog *csd = data;
    GError *e = message_arg_error_check(arg);
    char *s;
    SpecServer *ss;

    if (e != NULL) {
	warning_dialog(csd->conc, _("Failed to edit server"),
		_("Could not edit server (%s)"), e->message);
	g_clear_error(&e);
	config_servers_dialog_destroy(csd);
    }
    ss = spec_server_new_from_arg(arg);
    g_assert(ss);
    config_servers_dialog(csd);
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
		    CONFIG_SERVER_DIALOG_NAME)), ss->name);
    switch (ss->type) {
	case SERVER_TYPE_EXEC:
	    set_combobox_text(csd->xml, CONFIG_SERVER_DIALOG_TYPE, NULL, 0,
		    "Exec");
	    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
			CONFIG_SERVER_DIALOG_EXEC_PATH)), ss->exec.path);
	    break;
	case SERVER_TYPE_CONNECT:
	    set_combobox_text(csd->xml, CONFIG_SERVER_DIALOG_TYPE, NULL, 0,
		    "Connect");
	    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
			CONFIG_SERVER_DIALOG_CONNECT_HOST)), ss->connect.host);
	    s = g_strdup_printf("%d", ss->connect.port);
	    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
			CONFIG_SERVER_DIALOG_CONNECT_PORT)), s);
	    g_free(s);
	    break;
	case SERVER_TYPE_SSH:
	    set_combobox_text(csd->xml, CONFIG_SERVER_DIALOG_TYPE, NULL, 0,
		    "SSH");
	    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
			CONFIG_SERVER_DIALOG_SSH_HOST)), ss->ssh.host);
	    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
			CONFIG_SERVER_DIALOG_SSH_USER)), ss->ssh.ssh_user);
	    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
			CONFIG_SERVER_DIALOG_SSH_IDENTITY)),
		    ss->ssh.ssh_identity);
	    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(csd->xml,
			CONFIG_SERVER_DIALOG_SSH_COMMAND)), ss->ssh.remote_cmd);
	    break;
	default:
	    g_assert_not_reached();
    }
    spec_server_free(ss);
}

extern void
config_servers_edit(Conc *conc,
	const gchar *name) {
    ConfigServerDialog *csd = g_new0(struct _ConfigServerDialog, 1);
    MessageArg *a;
    GError *e = NULL;

    csd->conc = conc;
    csd->name = g_strdup(name);
    csd->xml = NULL;
    a = message_arg_string("name", csd->name);
    if (message_send(csd->conc->msg_q, CLIENT_MESSAGE_CONFIG_SERVER_SPEC,
		a, config_servers_edit_data, csd, &e) == FALSE) {
	error_dialog(csd->conc, _("Internal error"),
		_("Failed to send message (%s)"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
}

extern void
config_servers_remove(Conc *conc,
	const gchar *name) {
    ConfigServerDialog *csd = g_new0(struct _ConfigServerDialog, 1);
    MessageArg *a;
    GError *e = NULL;

    csd->conc = conc;
    csd->name = g_strdup(name);
    csd->xml = NULL;
    a = message_arg_string("name", name);
    if (message_send(csd->conc->msg_q, CLIENT_MESSAGE_CONFIG_SERVER_REMOVE,
		a, config_servers_done, csd, &e) == FALSE) {
	error_dialog(csd->conc, _("Internal error"),
		_("Failed to send message (%s)"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
}

