#include "config-users.h"
#include "client-message.h"
#include "spec-user.h"
#include "error.h"

typedef struct _ConfigUserDialog {
    Conc *conc;
    gchar *name; /* NULL if adding new user */
    GladeXML *xml; /* NULL if removing user */
} ConfigUserDialog;

/* 
 * Glade widget names:
 */
#define CONFIG_USER_DIALOG		"config user dialog"
#define CONFIG_USER_DIALOG_NAME		"cudname"
#define CONFIG_USER_DIALOG_PW1		"cudpw1"
#define CONFIG_USER_DIALOG_PW2		"cudpw2"

/*
 * Glade handler names:
 */
#define CONFIG_USER_DIALOG_RESPONSE	"response"
#define CONFIG_USER_DIALOG_PW_CHANGED	"changed"

static void
config_users_done(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    ConfigUserDialog *cud = data;
    GError *e = message_arg_error_check(arg);
    gchar *failed_title = _("Failed to update user");
    gchar *failed_text = _("Could not modify user (%s)");

    if (cud->xml == NULL) {
	failed_title = _("Failed to remove user");
	failed_text = _("Could not remove user (%s)");
    }
    if (cud->name == NULL) {
	failed_title = _("Failed to add user");
	failed_text = _("Could not add user (%s)");
    }
    if (e != NULL) {
	warning_dialog(cud->conc, failed_title, failed_text, e->message);
	g_clear_error(&e);
	if (cud->xml != NULL) {
	    return;
	}
    }
    if (cud->xml) {
	gtk_widget_destroy(glade_xml_get_widget(cud->xml, CONFIG_USER_DIALOG));
    }
    if (cud->name) {
	g_free(cud->name);
    }
    g_free(cud);
}

static void
config_users_dialog_response(GtkDialog *dialog,
	gint arg1,
	gpointer user_data) {
    ConfigUserDialog *cud = user_data;

    if (arg1 == GTK_RESPONSE_OK) {
	gchar *command;
	SpecUser su;
	MessageArg *a;
	GError *e = NULL;

	su.user = (gchar *)
	    gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(cud->xml,
			CONFIG_USER_DIALOG_NAME)));
	su.password = (gchar *)
	    gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(cud->xml,
			CONFIG_USER_DIALOG_PW1)));
	a = spec_user_message_arg("spec", &su);
	if (cud->name == NULL) {
	    command = CLIENT_MESSAGE_CONFIG_USER_ADD;
	} else {
	    MessageArg *spec = a;

	    command = CLIENT_MESSAGE_CONFIG_USER_EDIT;
	    a = message_arg_list("", NULL);
	    a->list = g_list_prepend(a->list,
		    message_arg_string("name", cud->name));
	    a->list = g_list_prepend(a->list, spec);
	}
	if (message_send(cud->conc->msg_q, command, a, config_users_done,
		    cud, &e) == FALSE) {
	    error_dialog(cud->conc, _("Internal error"),
		    _("Failed to send message (%s)"), e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
	return;
    }
    gtk_widget_destroy(glade_xml_get_widget(cud->xml, CONFIG_USER_DIALOG));
    if (cud->name) {
	g_free(cud->name);
    }
    g_free(cud);
}

static void
config_users_dialog(ConfigUserDialog *cud) {

    cud->xml = glade_xml_new(cud->conc->glade_file, CONFIG_USER_DIALOG, NULL);
    if (cud->name != NULL) {
	gchar *title = g_strdup_printf(_("Edit User: %s"), cud->name);
	
	g_object_set(glade_xml_get_widget(cud->xml, CONFIG_USER_DIALOG),
		"title", title, NULL);
    }
    glade_xml_signal_connect_data(cud->xml, CONFIG_USER_DIALOG_RESPONSE,
	    (GCallback)config_users_dialog_response, cud);
}

extern void
config_users_add(Conc *conc) {
    ConfigUserDialog *cud = g_new0(struct _ConfigUserDialog, 1);

    cud->conc = conc;
    cud->name = NULL;
    config_users_dialog(cud);
}

static void
config_users_edit_data(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    ConfigUserDialog *cud = data;
    GError *e = message_arg_error_check(arg);
    SpecUser *su;
     
    if (e != NULL) {
	warning_dialog(cud->conc, _("Failed to edit user"),
		_("Could not edit user (%s)"), e->message);
	g_clear_error(&e);
	if (cud->name) {
	    g_free(cud->name);
	}
	g_free(cud);
	return;
    }
    config_users_dialog(cud);
    su = spec_user_new_from_arg(arg);
    g_assert(su);
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(cud->xml,
			CONFIG_USER_DIALOG_NAME)), su->user);
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(cud->xml,
			CONFIG_USER_DIALOG_PW1)), "*****");
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(cud->xml,
			CONFIG_USER_DIALOG_PW2)), "*****");
    g_object_set(glade_xml_get_widget(cud->xml, CONFIG_USER_DIALOG_PW1),
	    "sensitive", FALSE, NULL);
    g_object_set(glade_xml_get_widget(cud->xml, CONFIG_USER_DIALOG_PW2),
	    "sensitive", FALSE, NULL);
    spec_user_free(su);
}

extern void
config_users_edit(Conc *conc,
	const gchar *name) {
    ConfigUserDialog *cud = g_new0(struct _ConfigUserDialog, 1);
    MessageArg *a;
    GError *e = NULL;

    cud->conc = conc;
    cud->name = g_strdup(name);
    cud->xml = NULL;
    a = message_arg_string("name", cud->name);
    if (message_send(cud->conc->msg_q, CLIENT_MESSAGE_CONFIG_USER_SPEC,
		a, config_users_edit_data, cud, &e) == FALSE) {
	error_dialog(cud->conc, _("Internal error"),
		_("Failed to send message (%s)"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
}

extern void
config_users_remove(Conc *conc,
	const gchar *name) {
    ConfigUserDialog *cud = g_new0(struct _ConfigUserDialog, 1);
    MessageArg *a;
    GError *e = NULL;

    cud->conc = conc;
    cud->name = g_strdup(name);
    cud->xml = NULL;
    a = message_arg_string("name", name);
    if (message_send(cud->conc->msg_q, CLIENT_MESSAGE_CONFIG_USER_REMOVE,
		a, config_users_done, cud, &e) == FALSE) {
	error_dialog(cud->conc, _("Internal error"),
		_("Failed to send message (%s)"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
}

