#include <string.h>
#include "conc.h"
#include "groups.h"
#include "terminal.h"
#include "log.h"
#include "message.h"
#include "client-message.h"
#include "date.h"

enum {
    COLUMN_NAME,
    COLUMN_STATUS,
    COLUMN_LAST_ACTIVE,
    COLUMN_LAST
};

typedef struct _ConcGroup {
    gchar *name;
    GtkListStore *consoles;
    GtkTreeView *treeview;
    GtkLabel *label;
    GtkLabel *menu_label;
    GtkImage *image;
    guint pageno;
    Conc *conc;
} ConcGroup;

/* 
 * Glade widget names:
 */
#define GROUP_WIDGET			"tab widget"
#define GROUP_WIDGET_TREEVIEW		"gwtreeview"
#define GROUP_LABEL_WIDGET		"tab label widget"
#define GROUP_LABEL_TEXT		"gwlabel"
#define GROUP_LABEL_IMAGE		"gwimage"

/*
 * Glade handler names:
 */
#define GROUP_START_TERMINAL		"start terminal"
#define GROUP_SELECTION_CHANGED		"selection changed"
#define GROUP_BUTTON_PRESS		"button press"
#define GROUP_POPUP_MENU		"popup menu"
#define GROUP_SHOW_BROKEN		"show broken"

static void
group_selection_changed(GtkTreeView *treeview,
	gpointer user_data) {
    ConcGroup *cg = user_data;
    Conc *conc = cg->conc;
    GtkTreeSelection *selection = gtk_tree_view_get_selection(treeview);
    GtkTreeIter iter;

    if (conc->current_console != NULL) {
	g_free(conc->current_console);
	conc->current_console = NULL;
    }
    if (conc->current_group != NULL) {
	g_free(conc->current_group);
	conc->current_group = NULL;
    }
    if (gtk_tree_selection_get_selected(selection, NULL, &iter) == TRUE) {
	conc->current_group = g_strdup(cg->name);
	gtk_tree_model_get(GTK_TREE_MODEL(cg->consoles), &iter,
		COLUMN_NAME, &conc->current_console, -1);
    }
    if (conc->current_group == NULL || conc->current_console == NULL) {
	g_object_set(conc->menu_view_console, "sensitive", FALSE, NULL);
    } else {
	g_object_set(conc->menu_view_console, "sensitive", TRUE, NULL);
    }
}

static void
console_status(Conc *conc,
	const GList *consoles) {
    const GList *l;
    ConcStatus *cs;
    GdkPixbuf *image = NULL;
    GdkPixbuf *old_image = NULL;
    ConcGroup *cg;
    gint pageno = gtk_notebook_get_current_page(conc->notebook);
    GtkTreeIter iter;
    gboolean found;
    gchar *timestr;
    gchar *old_timestr = NULL;

    for (l = conc->groups; l != NULL; l = g_list_next(l)) {
	cg = l->data;
	if (pageno == cg->pageno) {
	    break;
	}
    }
    if (l == NULL ||
	    gtk_tree_model_get_iter_first(GTK_TREE_MODEL(cg->consoles),
		&iter) == FALSE) {
	return;
    }
    for (l = consoles; l != NULL; l = g_list_next(l)) {
	cs = l->data;

	found = FALSE;
	gtk_tree_model_get_iter_first(GTK_TREE_MODEL(cg->consoles), &iter);
	do {
	    gchar *name;

	    gtk_tree_model_get(GTK_TREE_MODEL(cg->consoles), &iter,
		    COLUMN_NAME, &name, -1);
	    if (strcmp(name, cs->name) == 0) {
		g_free(name);
		found = TRUE;
		break;
	    }
	    g_free(name);
	} while (gtk_tree_model_iter_next(GTK_TREE_MODEL(cg->consoles), &iter)
		== TRUE);
	if (found == FALSE) {
	    continue;
	}
	switch (cs->status) {
	    case STATUS_OK:
		image = conc->ok_img;
		break;
	    case STATUS_BROKEN:
		image = conc->failed_img;
		break;
	    case STATUS_ACTIVE:
		image = conc->active_img;
		break;
	    default:
		g_assert_not_reached();
	}
	if (cs->last_active == 0) {
	    timestr = g_strdup("");
	} else {
	    timestr = date_print(cs->last_active);
	}
	gtk_tree_model_get(GTK_TREE_MODEL(cg->consoles), &iter,
		COLUMN_STATUS, &old_image,
		COLUMN_LAST_ACTIVE, &old_timestr,
		-1);
	if (old_timestr == NULL || strcmp(timestr, old_timestr) != 0) {
	    gtk_list_store_set(cg->consoles, &iter,
		    COLUMN_LAST_ACTIVE, timestr,
		    -1);
	}
	if (old_image == NULL || old_image != image) {
	    gtk_list_store_set(cg->consoles, &iter,
		    COLUMN_STATUS, image,
		    -1);
	}
	g_free(timestr);
	if (old_timestr != NULL) {
	    g_free(old_timestr);
	}
	if (old_image != NULL) {
	    g_object_unref(old_image);
	}
    }
}

static void
group_start_terminal(GtkTreeView *treeview,
	GtkTreePath *arg1,
	GtkTreeViewColumn *arg2,
	gpointer user_data) {
    ConcGroup *cg = user_data;
    GtkTreeIter iter;
    gchar *console;

    if (gtk_tree_model_get_iter(GTK_TREE_MODEL(cg->consoles), &iter, arg1)
	    == FALSE) {
	return;
    }
    gtk_tree_model_get(GTK_TREE_MODEL(cg->consoles), &iter,
	COLUMN_NAME, &console,
	-1);
    terminal_open(cg->conc, cg->name, console);
    g_free(console);
}

static void
popup_console(GtkWidget *menuitem,
	gpointer data) {
    ConcGroup *cg = data;
    GtkTreeSelection *selection = gtk_tree_view_get_selection(cg->treeview);
    GtkTreeIter iter;

    if (gtk_tree_selection_get_selected(selection, NULL, &iter) == TRUE) {
	gchar *console;

	gtk_tree_model_get(GTK_TREE_MODEL(cg->consoles), &iter,
		COLUMN_NAME, &console, -1);
	terminal_open(cg->conc, cg->name, console);
	g_free(console);
    }
}

static void
popup_log(GtkWidget *menuitem,
	gpointer data) {
    ConcGroup *cg = data;
    GtkTreeSelection *selection = gtk_tree_view_get_selection(cg->treeview);
    GtkTreeIter iter;

    if (gtk_tree_selection_get_selected(selection, NULL, &iter) == TRUE) {
	gchar *console;

	gtk_tree_model_get(GTK_TREE_MODEL(cg->consoles), &iter,
		COLUMN_NAME, &console, -1);
	log_view(cg->conc, console);
	g_free(console);
    }
}

static void
group_show_popup_menu(ConcGroup *cg,
	GdkEventButton *event) {
    GtkWidget *menu;
    GtkWidget *menuitem;

    menu = gtk_menu_new();
    menuitem = gtk_menu_item_new_with_mnemonic("View _Console");
    g_signal_connect(menuitem, "activate", (GCallback)popup_console, cg);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
    menuitem = gtk_menu_item_new_with_mnemonic("View _Log");
    g_signal_connect(menuitem, "activate", (GCallback)popup_log, cg);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
    gtk_widget_show_all(menu);
    gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
	    event == NULL ?0:event->button,
	    gdk_event_get_time((GdkEvent *)event));
}

static gboolean
group_popup_menu(GtkWidget *treeview,
	gpointer data) {
    ConcGroup *cg = data;
    group_show_popup_menu(cg, NULL);
    return TRUE;
}

static gboolean
group_button_pressed(GtkWidget *treeview,
	GdkEventButton *event,
	gpointer data) {
    ConcGroup *cg = data;

    if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
	GtkTreeSelection *selection;
	GtkTreePath *path;

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cg->treeview));
	if (gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(cg->treeview),
		    (gint)event->x, (gint)event->y, &path, NULL, NULL, NULL)) {
	    gtk_tree_selection_unselect_all(selection);
	    gtk_tree_selection_select_path(selection, path);
	    gtk_tree_path_free(path);
	}
	group_show_popup_menu(cg, event);
	return TRUE;
    }
    return FALSE;
}

static void
set_tooltip(MessageQueue q,
        const MessageArg *arg,
        gpointer data) {
    GtkTooltip *tooltip = data;

    if (arg->type != MESSAGE_ARG_STRING || arg->string->len == 0) {
	gtk_tooltip_set_text(tooltip, NULL);
    } else {
	gtk_tooltip_set_text(tooltip, arg->string->str);
    }
    g_object_unref(tooltip);
}

static gboolean
group_show_tooltip(GtkWidget  *widget,
	gint x,
	gint y,
	gboolean keyboard_mode,
	GtkTooltip *tooltip,
	gpointer data) {
    ConcGroup *cg = data;
    GtkTreeView *tv = GTK_TREE_VIEW(widget);
    GtkTreeModel *model;
    GtkTreeIter iter;
    gchar *name;
    GdkPixbuf *img;
    MessageArg *a;
    GError *e = NULL;

    if (gtk_tree_view_get_tooltip_context(tv, &x, &y, keyboard_mode,
		&model, NULL, &iter) == FALSE) return FALSE;
    gtk_tree_model_get(model, &iter, COLUMN_NAME, &name,
	   COLUMN_STATUS, &img, -1);
    if (name == NULL || img == NULL || img != cg->conc->failed_img) {
	if (name != NULL) g_free(name);
	if (img != NULL) g_object_unref(img);
	return FALSE;
    }
    g_object_unref(img);
    a = message_arg_string("console", name);
    g_free(name);
    if (message_send(cg->conc->msg_q, CLIENT_MESSAGE_CONSOLE_GET_BROKEN, a,
		set_tooltip, tooltip, &e) == FALSE) {
	message_arg_free(a);
	g_printerr("failed to send message (%s)\n", e->message);
	g_clear_error(&e);
	return FALSE;
    }
    message_arg_free(a);
    g_object_ref(tooltip);
    return TRUE;
}

static gboolean
group_subscribe_fe(GtkTreeModel *model,
	GtkTreePath *path,
	GtkTreeIter *iter,
	gpointer data) {
    ConcGroup *cg = data;
    gchar *name;

    gtk_tree_model_get(model, iter, GROUPS_COLUMN_NAME, &name, -1);
    if (name != NULL && strcmp(name, cg->name) == 0) {
	g_free(name);
	gtk_list_store_set(GTK_LIST_STORE(model), iter,
		GROUPS_COLUMN_SUBSCRIBED, TRUE, -1);
	return TRUE;
    }
    g_free(name);
    return FALSE;
}

extern void
group_add(Conc *conc,
	const gchar *name,
	const GList *consoles)
{
    ConcGroup *cg;
    GladeXML *xml;
    GtkWidget *tab, *label;
    gint p;
    const GList *l;
    GtkTreeIter iter;
    GtkCellRenderer *text_rend, *image_rend;
    GtkTreeViewColumn *col;

    cg = g_new0(ConcGroup, 1);
    cg->name = g_strdup(name);
    cg->conc = conc;
    xml = glade_xml_new(conc->glade_file, GROUP_WIDGET, NULL);
    glade_xml_signal_connect_data(xml, GROUP_START_TERMINAL,
	    (GCallback)group_start_terminal, cg);
    glade_xml_signal_connect_data(xml, GROUP_SELECTION_CHANGED,
	    (GCallback)group_selection_changed, cg);
    glade_xml_signal_connect_data(xml, GROUP_BUTTON_PRESS,
	    (GCallback)group_button_pressed, cg);
    glade_xml_signal_connect_data(xml, GROUP_POPUP_MENU,
	    (GCallback)group_popup_menu, cg);
    glade_xml_signal_connect_data(xml, GROUP_SHOW_BROKEN,
	    (GCallback)group_show_tooltip, cg);
    cg->treeview = GTK_TREE_VIEW(glade_xml_get_widget(xml,
		GROUP_WIDGET_TREEVIEW));
    tab = glade_xml_get_widget(xml, GROUP_WIDGET);
    xml = glade_xml_new(conc->glade_file, GROUP_LABEL_WIDGET, NULL);
    label = glade_xml_get_widget(xml, GROUP_LABEL_WIDGET);
    cg->label = GTK_LABEL(glade_xml_get_widget(xml, GROUP_LABEL_TEXT));
    gtk_label_set_text(cg->label, cg->name);
    cg->image = GTK_IMAGE(glade_xml_get_widget(xml, GROUP_LABEL_IMAGE));
    cg->menu_label = GTK_LABEL(gtk_label_new(cg->name));
    p = gtk_notebook_append_page_menu(conc->notebook, tab, label,
	    GTK_WIDGET(cg->menu_label));
    if (p >= 0) {
	cg->pageno = p;
    } else {
	g_printerr(_("could not append notebook page!\n"));
    }
    cg->consoles = gtk_list_store_new(COLUMN_LAST, G_TYPE_STRING,
	    GDK_TYPE_PIXBUF, G_TYPE_STRING);
    for (l = consoles; l != NULL; l = g_list_next(l)) {
	gtk_list_store_append(cg->consoles, &iter);
	gtk_list_store_set(cg->consoles, &iter,
		COLUMN_NAME, l->data,
		COLUMN_STATUS, conc->ok_img,
		-1);
    }
    g_object_set(cg->treeview,
	    "model", cg->consoles,
	    "rules-hint", TRUE,
	    "headers-clickable", TRUE,
	    "reorderable", TRUE,
	    "enable-search", TRUE,
	    "search-column", COLUMN_NAME,
	    NULL);
    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(cg->consoles),
	    COLUMN_NAME, GTK_SORT_ASCENDING);
    text_rend = gtk_cell_renderer_text_new();
    image_rend = gtk_cell_renderer_pixbuf_new();
    col = gtk_tree_view_column_new_with_attributes("Console", text_rend,
	    "text", COLUMN_NAME, NULL);
    gtk_tree_view_append_column(cg->treeview, col);
    col = gtk_tree_view_column_new_with_attributes("Status", image_rend,
	    "pixbuf", COLUMN_STATUS, NULL);
    gtk_tree_view_append_column(cg->treeview, col);
    col = gtk_tree_view_column_new_with_attributes("Last Active", text_rend,
	    "text", COLUMN_LAST_ACTIVE, NULL);
    gtk_tree_view_append_column(cg->treeview, col);
    conc->groups = g_list_append(conc->groups, cg);
    if (gtk_notebook_get_current_page(conc->notebook) == cg->pageno) {
	group_select(conc, cg->pageno);
    }
    gtk_tree_model_foreach(GTK_TREE_MODEL(conc->all_groups),
	    group_subscribe_fe, cg);
}

static gboolean
group_unsubscribe_fe(GtkTreeModel *model,
	GtkTreePath *path,
	GtkTreeIter *iter,
	gpointer data) {
    ConcGroup *cg = data;
    gchar *name;

    gtk_tree_model_get(model, iter, GROUPS_COLUMN_NAME, &name, -1);
    if (name != NULL && strcmp(name, cg->name) == 0) {
	g_free(name);
	gtk_list_store_set(GTK_LIST_STORE(model), iter,
		GROUPS_COLUMN_SUBSCRIBED, FALSE, -1);
	return TRUE;
    }
    g_free(name);
    return FALSE;
}

extern void
group_delete(Conc *conc,
	const gchar *name) {
    GList *l;
    ConcGroup *cg, *ocg;

    for (l = conc->groups; l != NULL; l = g_list_next(l)) {
	cg = l->data;
	if (strcmp(cg->name, name) == 0) {
	    break;
	}
    }
    if (l == NULL) {
	return;
    }
    conc->groups = g_list_remove(conc->groups, cg);
    gtk_notebook_remove_page(conc->notebook, cg->pageno);
    for (l = conc->groups; l != NULL; l = g_list_next(l)) {
	ocg = l->data;
	if (ocg->pageno > cg->pageno) {
	    ocg->pageno -= 1;
	}
    }
    gtk_tree_model_foreach(GTK_TREE_MODEL(conc->all_groups),
	    group_unsubscribe_fe, cg);
    g_free(cg->name);
    g_object_unref(cg->consoles);
    g_free(cg);
    if (gtk_notebook_get_n_pages(conc->notebook) == 0) {
	if (conc->current_console != NULL) {
	    g_free(conc->current_console);
	    conc->current_console = NULL;
	}
	if (conc->current_group != NULL) {
	    g_free(conc->current_group);
	    conc->current_group = NULL;
	}
	g_object_set(conc->menu_view_console, "sensitive", FALSE, NULL);
    }
}

extern void
group_select(Conc *conc,
	guint page) {
    GList *l;

    for (l = conc->groups; l != NULL; l = g_list_next(l)) {
	ConcGroup *cg = l->data;
	MessageArg *a;
	GError *e = NULL;

	if (cg->pageno != page) {
	    continue;
	}
	a = message_arg_string("cgroup", cg->name);
	if (message_send(conc->msg_q, CLIENT_MESSAGE_CURRSUB, a,
		    NULL, NULL, &e) == FALSE) {
	    message_arg_free(a);
	    g_printerr("failed to send message (%s)\n", e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
	group_selection_changed(cg->treeview, cg);
	break;
    }
}

extern void
group_status(Conc *conc,
	const GList *cgroups,
	const GList *consoles) {
    const GList *l;
    GList *p;
    ConcGroup *group;
    ConcStatus *cs;
    GdkPixbuf *image = NULL;

    for (l = cgroups; l != NULL; l = g_list_next(l)) {
	cs = l->data;

	for (p = conc->groups; p != NULL; p = g_list_next(p)) {
	    group = p->data;
	    if (strcmp(cs->name, group->name) == 0) {
		break;
	    }
	}
	if (p == NULL) {
	    continue;
	}
	switch (cs->status) {
	    case STATUS_OK:
		image = conc->ok_img;
		break;
	    case STATUS_BROKEN:
		image = conc->failed_img;
		break;
	    case STATUS_ACTIVE:
		image = conc->active_img;
		break;
	    default:
		g_assert_not_reached();
	}
	if (gtk_image_get_pixbuf(group->image) != image) {
	    gtk_image_set_from_pixbuf(group->image, image);
	}
    }
    console_status(conc, consoles);
}

static ConcGroup *
group_find(Conc *conc,
	const gchar *name) {
    ConcGroup *cg;
    GList *l;

    g_assert(conc);
    g_assert(name);

    for (l = conc->groups; l != NULL; l = g_list_next(l)) {
	cg = l->data;
	if (!strcmp(name, cg->name)) {
	    return cg;
	}
    }
    return NULL;
}

extern void
group_add_console(Conc *conc,
	const gchar *group,
	const gchar *console) {
    ConcGroup *cg;
    GtkTreeIter iter;

    g_assert(conc);
    g_assert(group);
    g_assert(console);

    cg = group_find(conc, group);
    if (cg == NULL) {
	return;
    }
    gtk_list_store_append(cg->consoles, &iter);
    gtk_list_store_set(cg->consoles, &iter,
	    COLUMN_NAME, console,
	    COLUMN_STATUS, conc->ok_img,
	    -1);
}

extern void
group_remove_console(Conc *conc,
	const gchar *group,
	const gchar *console) {
    ConcGroup *cg;
    GtkTreeIter iter;

    g_assert(conc);
    g_assert(group);
    g_assert(console);

    cg = group_find(conc, group);
    if (cg == NULL) {
	return;
    }
    if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(cg->consoles),
		&iter)) {
	do {
	    gchar *c;

	    gtk_tree_model_get(GTK_TREE_MODEL(cg->consoles), &iter,
		    COLUMN_NAME, &c, -1);
	    if (!strcmp(c, console)) {
		g_free(c);
		gtk_list_store_remove(cg->consoles, &iter);
		break;
	    }
	    g_free(c);
	} while (gtk_tree_model_iter_next(GTK_TREE_MODEL(cg->consoles),
		    &iter));
    }
}

extern void
groups_rename_console(Conc *conc,
	const gchar *oldname,
	const gchar *newname) {
    GList *l;
    GtkTreeIter iter;

    g_assert(conc);
    g_assert(oldname);
    g_assert(newname);

    for (l = conc->groups ; l; l = g_list_next(l)) {
	ConcGroup *cg = l->data;
	if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(cg->consoles),
		    &iter)) {
	    do {
		gchar *c;

		gtk_tree_model_get(GTK_TREE_MODEL(cg->consoles), &iter,
			COLUMN_NAME, &c, -1);
		if (!strcmp(c, oldname)) {
		    gtk_list_store_set(cg->consoles, &iter,
			    COLUMN_NAME, newname,
			    -1);
		}
		g_free(c);
	    } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(cg->consoles),
			&iter));
	}
    }
}

extern void
groups_rename_cgroup(Conc *conc,
	const gchar *oldname,
	const gchar *newname) {
    GList *l;

    g_assert(conc);
    g_assert(oldname);
    g_assert(newname);
    
    for (l = conc->groups ; l; l = g_list_next(l)) {
	ConcGroup *cg = l->data;

	if (strcmp(oldname, cg->name) != 0) {
	    continue;
	}
	g_free(cg->name);
	cg->name = g_strdup(newname);
	gtk_label_set_text(cg->label, cg->name);
	gtk_label_set_text(cg->menu_label, cg->name);
    }
    groups_save(conc);
}

extern void
groups_save(Conc *conc) {
    GSList *groups = NULL;
    GList *l;

    for (l = g_list_last(conc->groups); l != NULL; l = g_list_previous(l)) {
	ConcGroup *cg = l->data;

	groups = g_slist_prepend(groups, cg->name);
    }
    gconf_client_set_list(conc->gconf, GCONF_SUBSCRIBED, GCONF_VALUE_STRING,
	    groups, NULL);
    g_slist_free(groups);
}
