#include <string.h>
#include "conc.h"
#include "terminal.h"
#include "error.h"
#include "pipe.h"
#include "client-message.h"
#include <vte/vte.h>

typedef struct _ConcConsole {
    gchar *name;
    GIOChannel *io;
    MplexChannel chan;
} ConcConsole;

typedef struct _ConcTerminal {
    gchar *cgroup;
    ConcConsole *console;
    GtkWindow *window;
    VteTerminal *terminal;
    GIOChannel *rio;
    GIOChannel *io;
    guint reader;
    Conc *conc;
} ConcTerminal;

/* 
 * Glade widget names:
 */
#define TERMINAL_WINDOW			"terminal window"
#define TERMINAL_WINDOW_VBOX		"twvbox"

/*
 * Glade handler names:
 */
#define TERMINAL_CLOSE			"close console"
#define TERMINAL_CLOSED			"console closed"

static void
terminal_feeder_done(gpointer data) {
    ConcTerminal *t = data;

    t->reader = 0;
}

static gboolean
terminal_feeder(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    ConcTerminal *t = data;
    gchar buf[256];
    gsize bytes;
    GError *e = NULL;

    switch (g_io_channel_read_chars(source, buf, sizeof(buf), &bytes, &e)) {
	case G_IO_STATUS_ERROR:
	    /* FIXME: report error? */
	    g_clear_error(&e);
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	default:
	    g_assert_not_reached();
    }
    vte_terminal_feed(t->terminal, buf, bytes);
    return TRUE;
}

static void
terminal_input(VteTerminal *vterminal,
	gchar *arg1,
	guint arg2,
	gpointer user_data) {
    ConcTerminal *t = user_data;

    g_io_channel_write_chars(t->console->io, arg1, arg2, NULL, NULL);
    /*FIXME: check for errors? */
}

static void
terminal_close(ConcTerminal *t) {
    Conc *conc = t->conc;
    ConcConsole *c = t->console;
    GList *l;
    MessageArg *a;
    GError *e = NULL;

    conc->terminals = g_list_remove(conc->terminals, t);
    gtk_widget_destroy(GTK_WIDGET(t->window));
    matrix_expunge(conc->matrix, t->console->io);
    if (t->reader != 0) {
	g_source_remove(t->reader);
    }
    g_io_channel_unref(t->io);
    g_io_channel_unref(t->rio);
    g_free(t->cgroup);
    g_free(t);
    for (l = conc->terminals; l != NULL; l = g_list_next(l)) {
	t = l->data;

	if (t->console == c) {
	    return;
	}
    }
    conc->consoles = g_list_remove(conc->consoles, c);
    a = message_arg_string("console", c->name);
    if (message_send(conc->msg_q, CLIENT_MESSAGE_DETACH, a,
		NULL, NULL, &e) == FALSE) {
	g_printerr("failed to send message (%s)\n", e->message);
	g_clear_error(&e);
    }
    g_free(c->name);
    matrix_expunge(conc->matrix, c->io);
    mplex_close(conc->mplex, c->chan, NULL);
    g_io_channel_unref(c->io);
    g_free(c);
}

static void
terminal_menu_close(GtkMenuItem *item,
	gpointer user_data) {
    ConcTerminal *t = user_data;

    terminal_close(t);
}

static gboolean
terminal_window_closed(GtkWidget *widget,
	GdkEvent *event,
	gpointer user_data) {
    ConcTerminal *t = user_data;

    terminal_close(t);
    return FALSE;
}

static void
terminal_title(VteTerminal *vteterminal,
	gpointer user_data) {
    ConcTerminal *t = user_data;
    gchar *title;

    title = g_strdup_printf("%s:%s %s", t->cgroup, t->console->name,
	    vte_terminal_get_window_title(vteterminal));
    g_object_set(t->window,
	    "title", title,
	    NULL);
    g_free(title);
}

static void
terminal_create(Conc *conc,
	const gchar *cgroup,
	ConcConsole *console) {
    ConcTerminal *t = g_new0(ConcTerminal, 1);
    GladeXML *xml;
    GtkWidget *vbox;
    GIOChannel *pair[2];
    gchar *title;

    conc->terminals = g_list_prepend(conc->terminals, t);
    pipe_new(pair);
    t->conc = conc;
    t->cgroup = g_strdup(cgroup);
    t->console = console;
    t->io = pair[0];
    t->rio = pair[1];
    matrix_connect(conc->matrix, t->console->io, t->rio, NULL);
    xml = glade_xml_new(conc->glade_file, TERMINAL_WINDOW, NULL);
    t->window = GTK_WINDOW(glade_xml_get_widget(xml, TERMINAL_WINDOW));
    t->terminal = VTE_TERMINAL(vte_terminal_new());
    vbox = glade_xml_get_widget(xml, TERMINAL_WINDOW_VBOX);
    gtk_box_pack_end_defaults(GTK_BOX(vbox), GTK_WIDGET(t->terminal));
    gtk_widget_show(GTK_WIDGET(t->terminal));
    title = g_strdup_printf("%s:%s", t->cgroup, t->console->name);
    g_object_set(t->window,
	    "title", title,
	    NULL);
    g_free(title);
    t->reader = g_io_add_watch_full(t->io, G_PRIORITY_DEFAULT, G_IO_IN,
	    terminal_feeder, t, terminal_feeder_done);
    g_signal_connect(t->terminal, "commit", (GCallback)terminal_input, t);
    g_signal_connect(t->terminal, "window-title-changed", (GCallback)terminal_title, t);
    glade_xml_signal_connect_data(xml, TERMINAL_CLOSE,
	    (GCallback)terminal_menu_close, t);
    glade_xml_signal_connect_data(xml, TERMINAL_CLOSED,
	    (GCallback)terminal_window_closed, t);
    g_object_set(t->terminal,
	    "has-focus", TRUE,
	    NULL);
    /* FIXME: read terminal properties from profile */
    vte_terminal_set_font_from_string(t->terminal, "Mono 10");
    vte_terminal_set_backspace_binding(t->terminal, VTE_ERASE_ASCII_BACKSPACE);
}

struct attached_data {
    Conc *conc;
    gchar *cgroup;
    gchar *name;
};

static void
console_attached(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    GError *e = message_arg_error_check(arg);
    struct attached_data *ad = data;
    Conc *conc = ad->conc;
    GIOChannel *io;
    ConcConsole *c;

    if (e != NULL) {
	warning_dialog(conc, _("Failed to attach to console"),
		_("Could not attach to console %s (%s)."),
		ad->name, e->message);
	g_clear_error(&e);
	g_free(ad->cgroup);
	g_free(ad->name);
	g_free(ad);
	return;
    }
    if (arg->type != MESSAGE_ARG_UINT16) {
	g_printerr(_("bad response from manager\n"));
	exit(EXIT_FAILURE);
    }
    io = mplex_open(conc->mplex, arg->uint16, NULL, NULL, NULL, &e);
    if (io == NULL) {
	g_printerr(_("failed to open channel to console %s (%s)\n"),
		ad->name, e->message);
	exit(EXIT_FAILURE);
    }
    c = g_new0(ConcConsole, 1);
    c->name = ad->name;
    c->io = io;
    c->chan = arg->uint16;
    conc->consoles = g_list_prepend(conc->consoles, c);
    terminal_create(conc, ad->cgroup, c);
    g_free(ad->cgroup);
    g_free(ad);
}


extern void
terminal_open(Conc *conc,
	const gchar *cgroup,
	const gchar *name) {
    GList *l;
    GError *e = NULL;
    MessageArg *a;
    struct attached_data *ad;

    for (l = conc->terminals; l != NULL; l = g_list_next(l)) {
	ConcTerminal *t = l->data;

	if (strcmp(t->cgroup, cgroup) == 0 &&
		strcmp(t->console->name, name) == 0) {
	    gtk_window_present(t->window);
	    return;
	}
    }
    for (l = conc->consoles; l != NULL; l = g_list_next(l)) {
	ConcConsole *c = l->data;

	if (strcmp(c->name, name) == 0) {
	    terminal_create(conc, cgroup, c);
	    return;
	}
    }
    ad = g_new0(struct attached_data, 1);
    ad->conc = conc;
    ad->cgroup = g_strdup(cgroup);
    ad->name = g_strdup(name);
    a = message_arg_string("console", ad->name);
    if (message_send(conc->msg_q, CLIENT_MESSAGE_ATTACH, a,
		console_attached, ad, &e) == FALSE) {
	g_printerr("failed to send message (%s)\n", e->message);
	g_clear_error(&e);
	g_free(ad->cgroup);
	g_free(ad->name);
	g_free(ad);
	return;
    }
}

extern void
terminal_detach(Conc *conc,
	const gchar *name) {
    GList *l;
    ConcTerminal *t = NULL;

    do {
	for (l = conc->terminals; l != NULL; l = g_list_next(l)) {
	    t = l->data;

	    if (strcmp(t->console->name, name) == 0) {
		break;
	    }
	}
	if (l != NULL) {
	    terminal_close(t);
	}
    } while (l != NULL);
}

extern void
terminals_rename_console(Conc *conc,
	const gchar *oldname,
	const gchar *newname) {
    GList *l;
    GList *consoles = NULL;

    for (l = conc->consoles; l != NULL; l = g_list_next(l)) {
	ConcConsole *c = l->data;

	if (strcmp(c->name, oldname) == 0) {
	    g_free(c->name);
	    c->name = g_strdup(newname);
	    consoles = g_list_prepend(consoles, c);
	}
    }
    while (consoles != NULL) {
	for (l = conc->terminals; l != NULL; l = g_list_next(l)) {
	    ConcTerminal *t = l->data;

	    if (t->console == consoles->data) {
		terminal_title(t->terminal, t);
	    }
	}
	consoles = g_list_remove(consoles, consoles->data);
    }
}

extern void
terminals_rename_cgroup(Conc *conc,
	const gchar *oldname,
	const gchar *newname) {
    GList *l;

    for (l = conc->terminals; l != NULL; l = g_list_next(l)) {
	ConcTerminal *t = l->data;

	if (strcmp(t->cgroup, oldname) == 0) {
	    g_free(t->cgroup);
	    t->cgroup = g_strdup(newname);
	    terminal_title(t->terminal, t);
	}
    }
}
