#include <string.h>
#include "config.h"
#ifdef CONC_HAS_IPMI
#include "ipmiconsole.h"
#endif
#include "ipmi.h"
#include "stdinout.h"

extern IPMI *
ipmi_new(const gchar *host,
	const gchar *user,
	const gchar *passwd) {
    IPMI *i = g_new0(struct _IPMI, 1);

    g_assert(host);
    g_assert(passwd);
    i->host = g_strdup(host);
    i->passwd = g_strdup(passwd);
    i->user = g_strdup(user == NULL ? "" : user);
    return i;
}

extern void
ipmi_free(IPMI *ipmi) {
    g_assert(ipmi);
    if (ipmi->host != NULL) g_free(ipmi->host);
    if (ipmi->user != NULL) g_free(ipmi->user);
    if (ipmi->passwd != NULL) g_free(ipmi->passwd);
    g_free(ipmi);
}

extern MessageArg *
ipmi_message_arg(const gchar *name,
	const IPMI *ipmi) {
    GList *l = NULL;

    g_assert(name);

    l = g_list_prepend(l, message_arg_string("host", ipmi->host));
    l = g_list_prepend(l, message_arg_string("user", ipmi->user));
    l = g_list_prepend(l, message_arg_string("passwd", ipmi->passwd));
    return message_arg_list(name, l);
}

extern IPMI *
ipmi_new_from_arg(const MessageArg *arg) {
    MessageArg *host;
    MessageArg *user;
    MessageArg *passwd;

    g_assert(arg);

    if (arg->type != MESSAGE_ARG_LIST) return NULL;
    if (message_arg_list_explode(arg,
		"host", MESSAGE_ARG_STRING, &host,
		"user", MESSAGE_ARG_STRING, &user,
		"passwd", MESSAGE_ARG_STRING, &passwd,
		NULL, 0, NULL) == FALSE) {
	return NULL;
    }
    return ipmi_new(host->string->str, user->string->str,
	    passwd->string->str);
}

extern gboolean
ipmi_match(const IPMI *a, const IPMI *b) {
    g_assert(a);
    g_assert(b);
    if ((a->host == NULL && b->host != NULL) ||
	    (a->host != NULL && b->host == NULL) ||
	    (a->host != NULL && b->host != NULL &&
	     strcmp(a->host, b->host) != 0)) {
	return FALSE;
    }
    if ((a->user == NULL && b->user != NULL) ||
	    (a->user != NULL && b->user == NULL) ||
	    (a->user != NULL && b->user != NULL &&
	     strcmp(a->user, b->user) != 0)) {
	return FALSE;
    }
    if ((a->passwd == NULL && b->passwd != NULL) ||
	    (a->passwd != NULL && b->passwd == NULL) ||
	    (a->passwd != NULL && b->passwd != NULL &&
	     strcmp(a->passwd, b->passwd) != 0)) {
	return FALSE;
    }
    return TRUE;
}

extern GIOChannel *
ipmi_open_io_channel(const IPMI *ipmi, GError **error) {
#ifdef CONC_HAS_IPMI
    static gboolean initialised = FALSE;
    struct ipmiconsole_ipmi_config config;
    struct ipmiconsole_protocol_config protocol_config;
    struct ipmiconsole_engine_config engine_config;
    ipmiconsole_ctx_t ctx;
    gint fd;

    if (initialised == FALSE) {
	if (ipmiconsole_engine_init(IPMICONSOLE_THREAD_COUNT_MAX, 0) < 0) {
	    g_set_error(error, IMPI_ERROR, IPMI_ERROR_IPMICONSOLE,
		    "could not initialise libimpiconsole");
	    return NULL;
	}
	initialised = TRUE;
    }
    if (strlen(ipmi->user) > 0) {
	config.username = ipmi->user;
    } else {
	config.username = NULL;
    }
    if (strlen(ipmi->passwd) > 0) {
	config.password = ipmi->passwd;
    } else {
	config.password = NULL;
    }
    config.k_g = NULL;
    config.privilege_level = -1;
    config.cipher_suite_id = -1;
    config.workaround_flags = 0;
    protocol_config.session_timeout_len = -1;
    protocol_config.retransmission_timeout_len = -1;
    protocol_config.retransmission_backoff_count = -1;
    protocol_config.keepalive_timeout_len = -1;
    protocol_config.retransmission_keepalive_timeout_len = -1;
    protocol_config.acceptable_packet_errors_count = -1;
    protocol_config.maximum_retransmission_count = 2;
    engine_config.engine_flags = 0;
    engine_config.behavior_flags = 0;
    engine_config.debug_flags = 0;
    ctx = ipmiconsole_ctx_create(ipmi->host,
	    &config, &protocol_config, &engine_config);
    if (ipmiconsole_engine_submit_block(ctx) < 0) {
	g_set_error(error, IMPI_ERROR, IPMI_ERROR_IPMICONSOLE,
		"%s", ipmiconsole_ctx_strerror(ipmiconsole_ctx_errnum(ctx)));
	return NULL;
    }
    fd = ipmiconsole_ctx_fd(ctx);
    if (fd < 0) {
	g_set_error(error, IMPI_ERROR, IPMI_ERROR_IPMICONSOLE,
		"%s", ipmiconsole_ctx_strerror(ipmiconsole_ctx_errnum(ctx)));
	return NULL;
    }
    return g_io_channel_unix_new(fd);
#else
    g_set_error(error, IMPI_ERROR, IPMI_ERROR_NOT_SUPPORTED,
	    "no IPMI support available");
    return NULL;
#endif
}

extern GQuark
ipmi_error_quark(void) {
    static GQuark error_quark = 0;

    if (error_quark == 0) {
	error_quark = g_quark_from_static_string("ipmi-error");
    }
    return error_quark;
}
