#include <stdarg.h>
#include <string.h>
#include "message.h"
#include "serialize.h"
#include "gi.h"

#define RESPONSE "_R_"
#define ERROR_ARG "_E_"

#define MESSAGE_BUFSIZ 1024

struct _MessageQueue {
    GIOChannel *io;
    gboolean is_running;
    guint reader;
    GString *buf;
    GTree *handlers; /* tree of MessageHandler */
    GTree *messages; /* tree of MessageState */
    MessageID next_id;
};

typedef struct _MessageHandler {
    gchar *name;
    MessageHandlerCB handler;
    gpointer data;
} MessageHandler;

typedef struct _MessageState {
    MessageID id;
    MessageResponseCB cb;
    gpointer data;
} MessageState;

typedef struct _Message {
    gchar *name;
    MessageID id;
    MessageArg *arg;
} Message;

static gint
uint32_compare(gconstpointer a,
	gconstpointer b,
	gpointer data) {
    guint32 id1 = GPOINTER_TO_UINT(a);
    guint32 id2 = GPOINTER_TO_UINT(b);

    return (gint) ((gint64) id2 - (gint64) id1);
}

static gboolean
find_id(MessageQueue q, MessageID *id) {
    MessageID start = g_random_int() >> 1;
	    /* g_tree_lookup seems to go wrong with top-bit-set keys */
    MessageID i = start;

    do {
	if (i >= (G_MAXUINT32 >>1)) {
	    i = 0;
	}
	if (g_tree_lookup(q->messages, GUINT_TO_POINTER(i)) == NULL) {
	    *id = i;
	    return TRUE;
	}
	i++;
    } while (i != start);
    return FALSE;
}

static GString *
put_arg(const MessageArg *a) {
    GString *data = g_string_new("");
    GString *s;
    guint8 t = '\0';

    s = serialize_string_gchar(a->name);
    data = g_string_append_len(data, s->str, s->len);
    g_string_free(s, TRUE);

    switch (a->type) {
	case MESSAGE_ARG_NONE:
	    t = '0';
	    break;
	case MESSAGE_ARG_UINT8:
	    t = '1';
	    break;
	case MESSAGE_ARG_UINT16:
	    t = '2';
	    break;
	case MESSAGE_ARG_UINT32:
	    t = '4';
	    break;
	case MESSAGE_ARG_UINT64:
	    t = '8';
	    break;
	case MESSAGE_ARG_STRING:
	    t = 's';
	    break;
	case MESSAGE_ARG_LIST:
	    t = 'l';
	    break;
	default:
	    g_assert_not_reached();
    }
    s = serialize_uint8(t);
    data = g_string_append_len(data, s->str, s->len);
    g_string_free(s, TRUE);

    switch (a->type) {
	case MESSAGE_ARG_NONE:
	    s = g_string_new("");
	    break;
	case MESSAGE_ARG_UINT8:
	    s = serialize_uint8(a->uint8);
	    break;
	case MESSAGE_ARG_UINT16:
	    s = serialize_uint16(a->uint16);
	    break;
	case MESSAGE_ARG_UINT32:
	    s = serialize_uint32(a->uint32);
	    break;
	case MESSAGE_ARG_UINT64:
	    s = serialize_uint64(a->uint64);
	    break;
	case MESSAGE_ARG_STRING:
	    s = serialize_string(a->string);
	    break;
	case MESSAGE_ARG_LIST:
	    s = serialize_uint32(g_list_length(a->list));
	    break;
	default:
	    g_assert_not_reached();
    }
    data = g_string_append_len(data, s->str, s->len);
    g_string_free(s, TRUE);

    if (a->type == MESSAGE_ARG_LIST) {
	GList *l;

	for (l = a->list; l; l = g_list_next(l)) {
	    MessageArg *m = l->data;

	    s = put_arg(m);
	    data = g_string_append_len(data, s->str, s->len);
	    g_string_free(s, TRUE);
	}
    }

    return data;
}


static gboolean
send_message(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	GError **error)
{
    GString *data = g_string_new("");
    GString *s;
    GIOStatus res;
    guint32 length;

    s = serialize_string_gchar(name);
    data = g_string_append_len(data, s->str, s->len);
    g_string_free(s, TRUE);

    s = serialize_uint32(id);
    data = g_string_append_len(data, s->str, s->len);
    g_string_free(s, TRUE);

    s = put_arg(arg);
    data = g_string_append_len(data, s->str, s->len);
    g_string_free(s, TRUE);

    length = data->len;
    data = g_string_prepend(data, "1234");
    data->str[0] = (guint8)((length >> 24) & 0xff);
    data->str[1] = (guint8)((length >> 16) & 0xff);
    data->str[2] = (guint8)((length >> 8) & 0xff);
    data->str[3] = (guint8)(length & 0xff);

    res = g_io_channel_write_chars(q->io, data->str, data->len, NULL, error);
    g_string_free(data, TRUE);
    return (res != G_IO_STATUS_ERROR);
}

static void
handler_free(gpointer data) {
    MessageHandler *h = data;

    if (h == NULL) {
	return;
    }
    if (h->name != NULL) {
	g_free(h->name);
    }
    g_free(h);
}

static MessageArg *
get_arg(GString **data) {
    MessageArg *a = g_new0(MessageArg, 1);
    GString *s = NULL;
    gchar t;
    gboolean res = FALSE;
    guint32 members = 0;

    if (!deserialize_string(data, &s)) {
	g_printerr(_("get_arg: argument corrupt\n"));
	return NULL;
    }
    a->name = s->str;
    g_string_free(s, FALSE);
    if (!deserialize_uint8(data, (guint8 *) &t)) {
	g_printerr(_("get_arg: argument %s corrupt\n"), a->name);
	g_free(a->name);
	g_free(a);
	return NULL;
    }
    switch (t) {
	case '0':
	    a->type = MESSAGE_ARG_NONE;
	    break;
	case '1':
	    a->type = MESSAGE_ARG_UINT8;
	    break;
	case '2':
	    a->type = MESSAGE_ARG_UINT16;
	    break;
	case '4':
	    a->type = MESSAGE_ARG_UINT32;
	    break;
	case '8':
	    a->type = MESSAGE_ARG_UINT64;
	    break;
	case 's':
	    a->type = MESSAGE_ARG_STRING;
	    break;
	case 'l':
	    a->type = MESSAGE_ARG_LIST;
	    break;
	default:
	    g_printerr(_("get_arg: argument %s corrupt\n"), a->name);
	    g_free(a->name);
	    g_free(a);
	    return NULL;
    }
    switch (a->type) {
	case MESSAGE_ARG_NONE:
	    res = TRUE;
	    break;
	case MESSAGE_ARG_UINT8:
	    res = deserialize_uint8(data, &(a->uint8));
	    break;
	case MESSAGE_ARG_UINT16:
	    res = deserialize_uint16(data, &(a->uint16));
	    break;
	case MESSAGE_ARG_UINT32:
	    res = deserialize_uint32(data, &(a->uint32));
	    break;
	case MESSAGE_ARG_UINT64:
	    res = deserialize_uint64(data, &(a->uint64));
	    break;
	case MESSAGE_ARG_STRING:
	    res = deserialize_string(data, &(a->string));
	    break;
	case MESSAGE_ARG_LIST:
	    res = deserialize_uint32(data, &members);
	    a->list = NULL;
	    if (!res) {
		break;
	    }
	    res = TRUE;
	    while (members > 0) {
		MessageArg *m = get_arg(data);
		if (m == NULL) {
		    GList *l;

		    for (l = a->list; l; l = g_list_next(l)) {
			MessageArg *m = l->data;

			message_arg_free(m);
		    }
		    res = FALSE;
		    break;
		}
		a->list = g_list_append(a->list, m);
		members--;
	    }
	    break;
	default:
	    g_assert_not_reached();
    }
    if (res == FALSE) {
	g_printerr(_("get_arg: argument %s corrupt\n"), a->name);
	g_free(a->name);
	g_free(a);
	return NULL;
    }
    return a;
}

static void
free_message(Message *m) {
    if (m == NULL) {
	return;
    }
    if (m->name != NULL) {
	g_free(m->name);
    }
    message_arg_free(m->arg);
    g_free(m);
}

static void
queue_response(MessageQueue q,
	Message *m)
{
    MessageState *ms = g_tree_lookup(q->messages, GUINT_TO_POINTER(m->id));

    if (ms == NULL) {
    } else {
	g_tree_steal(q->messages, GUINT_TO_POINTER(m->id));
	if (ms->cb) {
	    ms->cb(q, m->arg, ms->data);
	}
	g_free(ms);
    }
    free_message(m);
}

static void
queue_message(MessageQueue q,
	Message *m) {
    MessageHandler *mh = g_tree_lookup(q->handlers, m->name);

    if (mh == NULL) {
	MessageArg *arg = message_arg_error(255, "no handler registered");

	g_printerr(_("no handler registered for message %s\n"), m->name);
	message_respond(q, m->id, arg, NULL);
	message_arg_free(arg);
    } else {
	mh->handler(q, m->name, m->id, m->arg, mh->data);
    }
    free_message(m);
}

static gboolean
message_parse(MessageQueue q) {
    guint32 length = 0;
    GString *ms;
    GString *name;
    MessageID id;

    if (q->buf->len < 4) {
	return FALSE;
    }
    length = (length << 8) + (guint8)q->buf->str[0];
    length = (length << 8) + (guint8)q->buf->str[1];
    length = (length << 8) + (guint8)q->buf->str[2];
    length = (length << 8) + (guint8)q->buf->str[3];
    if (q->buf->len < (length + 4)) {
	return FALSE;
    }
    ms = g_string_new_len(&q->buf->str[4], length);
    q->buf = g_string_erase(q->buf, 0, length + 4);
    if (deserialize_string(&ms, &name)) {
	if (deserialize_uint32(&ms, &id)) {
	    MessageArg *a = get_arg(&ms);

	    if (a) {
		Message *m = g_new0(Message, 1);

		m->name = name->str;
		g_string_free(name, FALSE);
		m->id = id;
		m->arg = a;
		if (!strcmp(m->name, RESPONSE)) {
		    queue_response(q, m);
		} else {
		    queue_message(q, m);
		}
		g_string_free(ms, TRUE);
		return TRUE;
	    } else {
		g_string_free(name, TRUE);
	    }
	}
    }
    g_string_free(ms, TRUE);
    return FALSE;
}

static gboolean
message_reader(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    MessageQueue q = data;
    gchar buf[MESSAGE_BUFSIZ];
    gsize bytes;

    if (condition | G_IO_IN) {
	switch (g_io_channel_read_chars(source, buf,
		    MESSAGE_BUFSIZ, &bytes, NULL)) {
	    case G_IO_STATUS_AGAIN:
		return TRUE;
	    case G_IO_STATUS_EOF:
		return FALSE;
	    case G_IO_STATUS_ERROR:
		return FALSE;
	    default:
		break;
	}
	q->buf = g_string_append_len(q->buf, buf, bytes);
	while (message_parse(q));
	return TRUE;
    }
    return FALSE;
}


extern MessageQueue
message_queue_new(GIOChannel *io,
	GError **error) {
    MessageQueue q = g_new0(struct _MessageQueue, 1);

    q->io = io;
    g_io_channel_ref(io);
    q->handlers = g_tree_new_full((GCompareDataFunc)strcmp, NULL, g_free,
	    handler_free);
    q->messages = g_tree_new_full(uint32_compare, NULL, NULL, g_free);
    q->is_running = FALSE;
    q->buf = g_string_new("");
    return q;
}

extern void
message_queue_start(MessageQueue q) {
    g_assert(q);
    if (q->is_running == TRUE) {
	return;
    }
    q->is_running = TRUE;
    /* FIXME: if we did a full watch, we could do something useful when the
     * channel closed.
     */
    q->reader = g_io_add_watch(q->io, G_IO_IN, message_reader, q);
}

extern void
message_queue_free(MessageQueue q) {
    if (q->is_running) {
	g_source_remove(q->reader);
    }
    g_tree_destroy(q->handlers);
    g_tree_destroy(q->messages);
    g_string_free(q->buf, TRUE);
    g_io_channel_unref(q->io);
    g_free(q);
}

extern gboolean
message_handler_register(MessageQueue q,
	const gchar *name,
	MessageHandlerCB cb,
	gpointer data,
	GError **error) {
    MessageHandler *mh;

    g_assert(q);
    g_assert(name);
    g_assert(cb);
    mh = g_tree_lookup(q->handlers, name);
    if (mh) {
	/* FIXME: is this a programming error? if it is we should assert.
	 */
	g_set_error(error, MESSAGE_QUEUE_ERROR,
		MESSAGE_QUEUE_ERROR_ALREADY_REGISTERED,
		_("a handler is already registered for command %s"), mh->name);
	return FALSE;
    }
    mh = g_new0(MessageHandler, 1);
    mh->name = g_strdup(name);
    mh->handler = cb;
    mh->data = data;
    g_tree_insert(q->handlers, g_strdup(mh->name), mh);
    return TRUE;
}

extern gboolean
message_send(MessageQueue q,
	const gchar *name,
	const MessageArg *arg,
	MessageResponseCB cb,
	gpointer data,
	GError **error) {
    MessageID id;

    g_assert(q);
    g_assert(name);
    g_assert(arg);
    if (find_id(q, &id) == FALSE) {
	/* we've run out of message ID's for this channel - nasty.
	 */
	g_set_error(error, MESSAGE_QUEUE_ERROR,
		MESSAGE_QUEUE_ERROR_NO_MORE_IDS,
		_("no more message IDs available on message channel"));
	return FALSE;
    }
    if (send_message(q, name, id, arg, error) == TRUE) {
	MessageState *ms = g_new0(MessageState, 1);

	ms->id = id;
	ms->cb = cb;
	ms->data = data;
	g_tree_insert(q->messages, GUINT_TO_POINTER(id), ms);
	return TRUE;
    } else {
	return FALSE;
    }
    g_assert_not_reached();
    return FALSE;
}

extern gboolean
message_respond(MessageQueue q,
	MessageID id,
	const MessageArg *arg,
	GError **error) {
    g_assert(q);
    g_assert(arg);

    return send_message(q, RESPONSE, id, arg, error);
}

extern void
message_arg_free(MessageArg *a) {
    if (a == NULL) {
	return;
    }
    if (a->name) {
	g_free(a->name);
    }
    switch (a->type) {
	case MESSAGE_ARG_STRING:
	    if (a->string != NULL) {
		g_string_free(a->string, TRUE);
	    }
	    break;
	case MESSAGE_ARG_LIST:
	    if (a->list != NULL) {
		GList *l;

		for (l = a->list; l; l = g_list_next(l)) {
		    MessageArg *m = l->data;

		    message_arg_free(m);
		}
		g_list_free(a->list);
	    }
	    break;
	default:
	    break;
    }
    g_free(a);
}

extern gboolean
message_arg_list_explode(const MessageArg *a,
	...) {
    va_list ap;
    GList *l;
    gboolean ret = TRUE;

    g_assert(a);
    g_assert(a->type == MESSAGE_ARG_LIST);
    
    va_start(ap, a);
    for (;;) {
	gchar *name = va_arg(ap, gchar *);
	MessageArgType type = va_arg(ap, MessageArgType);
	MessageArg **arg = va_arg(ap, MessageArg **);

	if (name == NULL) {
	    break;
	}
	for (l = a->list; l; l = g_list_next(l)) {
	    MessageArg *larg = l->data;

	    if (!strcmp(larg->name, name) && larg->type == type) {
		*arg = larg;
		break;
	    }
	}
	if (l == NULL) {
	    ret = FALSE;
	}
    }
    va_end(ap);
    return ret;
}

extern MessageArg *
message_arg_none(const gchar *name) {
    MessageArg *arg = g_new0(MessageArg, 1);

    g_assert(name);

    arg->name = g_strdup(name);
    arg->type = MESSAGE_ARG_NONE;
    return arg;
}

extern MessageArg *
message_arg_uint8(const gchar *name,
	guint8 data) {
    MessageArg *arg = g_new0(MessageArg, 1);

    g_assert(name);

    arg->name = g_strdup(name);
    arg->type = MESSAGE_ARG_UINT8;
    arg->uint8 = data;
    return arg;
}

extern MessageArg *
message_arg_uint16(const gchar *name,
	guint16 data) {
    MessageArg *arg = g_new0(MessageArg, 1);

    g_assert(name);

    arg->name = g_strdup(name);
    arg->type = MESSAGE_ARG_UINT16;
    arg->uint16 = data;
    return arg;
}

extern MessageArg *
message_arg_uint32(const gchar *name,
	guint32 data) {
    MessageArg *arg = g_new0(MessageArg, 1);

    g_assert(name);

    arg->name = g_strdup(name);
    arg->type = MESSAGE_ARG_UINT32;
    arg->uint32 = data;
    return arg;
}

extern MessageArg *
message_arg_uint64(const gchar *name,
	guint64 data) {
    MessageArg *arg = g_new0(MessageArg, 1);

    g_assert(name);

    arg->name = g_strdup(name);
    arg->type = MESSAGE_ARG_UINT64;
    arg->uint64 = data;
    return arg;
}

extern MessageArg *
message_arg_string(const gchar *name,
	const gchar *data) {
    MessageArg *arg = g_new0(MessageArg, 1);

    g_assert(name);
    g_assert(data);

    arg->name = g_strdup(name);
    arg->type = MESSAGE_ARG_STRING;
    arg->string = g_string_new(data);
    return arg;
}

extern MessageArg *
message_arg_list(const gchar *name,
	GList *data) {
    MessageArg *arg = g_new0(MessageArg, 1);

    g_assert(name);

    arg->name = g_strdup(name);
    arg->type = MESSAGE_ARG_LIST;
    arg->list = data;
    return arg;
}

extern MessageArg *
message_arg_error(gint code,
	const gchar *format,
	...) {
    GList *l = NULL;
    gchar *msg;
    va_list ap;

    g_assert(format);
    l = g_list_append(l, message_arg_uint8("code", code));
    va_start(ap, format);
    msg = g_strdup_vprintf(format, ap);
    va_end(ap);
    l = g_list_append(l, message_arg_string("msg", msg));
    return message_arg_list(ERROR_ARG, l);
}

extern GError *
message_arg_error_check(const MessageArg *a) {
    MessageArg *code, *msg, *synthetic;
    GQuark domain;
    GError *e;

    g_assert(a);
    if (a->type != MESSAGE_ARG_LIST) {
	return NULL;
    }
    if (strcmp(a->name, ERROR_ARG) != 0) {
	return NULL;
    }
    code = msg = synthetic = NULL;
    message_arg_list_explode(a,
	    "code", MESSAGE_ARG_UINT8, &code,
	    "msg", MESSAGE_ARG_STRING, &msg,
	    "synthetic", MESSAGE_ARG_NONE, &synthetic,
	    NULL, 0, NULL);
    if (code == NULL || msg == NULL) {
	return NULL;
    }
    if (synthetic == NULL) {
	domain = MESSAGE_QUEUE_REMOTE_ERROR;
    } else {
	domain = MESSAGE_QUEUE_ERROR;
    }
    e = g_error_new(domain, code->uint8, "%s", msg->string->str);
    return e;
}

extern GQuark
message_queue_error_quark() {
    static GQuark error_quark = 0;

    if (error_quark == 0) {
	error_quark = g_quark_from_static_string("MESSAGE_QUEUE_ERROR");
    }
    return error_quark;
}

extern GQuark
message_queue_remote_error_quark() {
    static GQuark error_quark = 0;

    if (error_quark == 0) {
	error_quark = g_quark_from_static_string("MESSAGE_QUEUE_REMOTE_ERROR");
    }
    return error_quark;
}
