#ifndef __MESSAGE_H
#define __MESSAGE_H

#include "glib.h"

/*
 * CONC_MSG_CHANNEL - the mplex channel on which message channels are usually
 * started.
 *
 */
#define CONC_MSG_CHANNEL 1

/*
 * MessageQueueError - errors generating whilst using MessageQueues
 *
 */
typedef enum {
    MESSAGE_QUEUE_ERROR_ALREADY_REGISTERED,
    MESSAGE_QUEUE_ERROR_NO_MORE_IDS,
    MESSAGE_QUEUE_ERROR_MESSAGE_TIMEOUT
} MessageQueueError;
#define MESSAGE_QUEUE_ERROR message_queue_error_quark()
extern GQuark message_queue_error_quark(void);

/*
 * MessageQueueRemoteError - errors generated on the remote side of the
 * connection.
 *
 */
#define MESSAGE_QUEUE_REMOTE_ERROR message_queue_remote_error_quark()
extern GQuark message_queue_remote_error_quark(void);

/*
 * MessageQueue - an opaque structure representing a message queue.
 *
 * a message queue allow messages to be sent and received on a GIOChannel.
 * each message is assigned a unique message ID and every message illicits
 * a response that is tagged with this ID.
 *
 */
typedef struct _MessageQueue *MessageQueue;

/*
 * MessageID - a unique number assigned to each message.
 *
 */
typedef guint32 MessageID;

/*
 * MessageArgType - an argument may be one of these types
 */
typedef enum {
    MESSAGE_ARG_NONE,
    MESSAGE_ARG_UINT8,
    MESSAGE_ARG_UINT16,
    MESSAGE_ARG_UINT32,
    MESSAGE_ARG_UINT64,
    MESSAGE_ARG_STRING,
    MESSAGE_ARG_LIST
} MessageArgType;

/*
 * MessageArg - each message supplies exactly one argument (although this
 * could be a MESSAGE_ARG_LIST, which is a list of arguments).
 *
 */
typedef struct _MessageArg {
    gchar *name;
    MessageArgType type;
    union {
	guint8 uint8;
	guint16 uint16;
	guint32 uint32;
	guint64 uint64;
	GString *string;
	GList *list; /* list of MessageArg */
    };
} MessageArg;

/*
 * MessageHandlerCB - called when a message registered by the handler
 * is received.
 *
 */
typedef void (*MessageHandlerCB) (MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data);

/*
 * MessageResponseCB - called when response is received for a previous message.
 *
 */
typedef void (*MessageResponseCB) (MessageQueue q,
	const MessageArg *arg,
	gpointer data);

/*
 * message_queue_new - create a new message queue
 *
 * Creates a new message queue communicating over the GIOChannel.
 * No messages can be received until message_queue_start() is called
 * to allow handlers to be registered.
 * Returns the successfully created message queue or fails, returning
 * NULL and setting an error.
 */
extern MessageQueue message_queue_new(GIOChannel *io,
	GError **error);

/*
 * message_queue_start - allow a queue to receive messages
 *
 * A newly created queue will not receive messages until it has been
 * started.
 */
extern void message_queue_start(MessageQueue q);

/*
 * message_queue_free - destroy a message queue
 *
 * Destroys the message queue and frees all its resoruces.
 * Returns TRUE or fails, setting an error.
 *
 */
extern void message_queue_free(MessageQueue q);

/*
 * message_handler_register - register a handler for a named message
 *
 * If successful, all messages with this name will be passed to this
 * handler. The name and data belong to the caller.
 * Returns TRUE or fails, setting an error.
 *
 */
extern gboolean message_handler_register(MessageQueue q,
	const gchar *name,
	MessageHandlerCB cb,
	gpointer data,
	GError **error);

/*
 * message_send - send a message
 *
 * If successful, a message is sent to the specified message queue.
 * The name, arg, and data arguments belong to the caller.
 * Returns TRUE or fails, setting an error.
 *
 */
extern gboolean message_send(MessageQueue q,
	const gchar *name,
	const MessageArg *arg,
	MessageResponseCB cb,
	gpointer data,
	GError **error);

/*
 * message_respond - send a response message
 *
 * If sucessful, a response message is sent to the specified message queue.
 * The arg belongs to the caller.
 * Returns TRUE or fails, setting an error.
 *
 */
extern gboolean message_respond(MessageQueue q,
	MessageID id,
	const MessageArg *arg,
	GError **error);

/*
 * message_arg_free - destroy an argument
 *
 * Frees all the resources associated with an argument.
 */
extern void message_arg_free(MessageArg *a);

/*
 * message_arg_list_explode - extract data from a list argument
 *
 * Takes arguments in triplets (const gchar *name, MessageArgType type,
 * MessageArg **returnarg), and if an argument with the given name and type
 * exists in the argument list the returnarg is written with a pointer to
 * the matching argument. The argument should be terminated wit the triplet
 * (NULL, 0, NULL).
 * Returns TRUE if all the argument pointers were written.
 *
 * For example:
 *  message_arg_list_explode(arg, "fred", MESSAGE_ARG_STRING,
 *                           &fred_arg, NULL, 0, NULL);
 *
 */
extern gboolean message_arg_list_explode(const MessageArg *a,
	...);

/*
 * message_arg_none
 * message_arg_uint8
 * message_arg_uint16
 * message_arg_uint32
 * message_arg_uint64
 * message_arg_string
 * message_arg_list
 *                    - create a MessageArg of the given type
 *
 * Copies are made of all the data, apart from lists.
 * Returned argument belongs to caller.
 *
 */
extern MessageArg *message_arg_none(const gchar *name);
extern MessageArg *message_arg_uint8(const gchar *name,
	guint8 data);
extern MessageArg *message_arg_uint16(const gchar *name,
	guint16 data);
extern MessageArg *message_arg_uint32(const gchar *name,
	guint32 data);
extern MessageArg *message_arg_uint64(const gchar *name,
	guint64 data);
extern MessageArg *message_arg_string(const gchar *name,
	const gchar *data);
extern MessageArg *message_arg_list(const gchar *name,
	GList *data);

/*
 * message_arg_error - create a standard argument representing an error
 *
 * Returned argument belongs to caller.
 *
 */
extern MessageArg *message_arg_error(gint code,
	const gchar *format,
	...);

/*
 * message_arg_error_check - tests whather a given argument is an error
 *
 * This function extracts an error created with message_arg_error.
 * Return the error (which belings to the caller), or NULL is the
 * argument was not an error.
 *
 */
extern GError *message_arg_error_check(const MessageArg *a);


#endif
