#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#include "mplex.h"
#include "pipe.h"
#include "gi.h"

#define MPLEX_BUFSIZ 1024

#ifndef G_MAXUINT16
#define G_MAXUINT16    ((guint16) 0xffff)
#endif

typedef struct _Channel {
    Mplex mplex;
    MplexChannel chan;
    GIOChannel *io;
    guint reader;
    gboolean *flag;
    MplexChannelClosedCB closed_cb;
    gpointer user_data;
} Channel;

struct _Mplex {
    GIOChannel *io;
    Channel *chan[G_MAXUINT16]; /* would a tree save memory? */
    MplexChannel last;
    GString *buf;
    guint reader;
    gboolean is_running;
    gboolean is_closing;
    MplexClosedCB closed_cb;
    gpointer user_data;
};

static MplexChannel
mplex_next_chan(Mplex mplex) {
    MplexChannel c;
    for (c = mplex->last + 1; c != mplex->last ; c++) {
	if (c != 0 && mplex->chan[c] == NULL) return c;
    }
    return 0;
}

static void
mplex_closed(gpointer data) {
    Mplex mplex = data;

    if (mplex->is_closing == TRUE) {
	/* the mplex is being freed. */
	return;
    }
    mplex->closed_cb(mplex, mplex->user_data);
}

static gboolean
mplex_parse(Mplex mplex)
{
    MplexChannel chan;
    guint len;

    if (mplex->buf->len < 4) {
	return FALSE;
    }
    chan = ((guint8)mplex->buf->str[0] << 8) + (guint8)mplex->buf->str[1];
    len = ((guint8)mplex->buf->str[2] << 8) + (guint8)mplex->buf->str[3];
    if (mplex->buf->len < (len + 4)) {
	return FALSE;
    }
    if (mplex->chan[chan]) {
	if (len == 0) {
	    /* EOF on channel */
	    mplex_close(mplex, chan, NULL);
	} else {
	    if (mplex->chan[chan]->flag != NULL) {
		*(mplex->chan[chan]->flag) = TRUE;
	    }
	    g_io_channel_write_chars(mplex->chan[chan]->io,
			&mplex->buf->str[4], len, NULL, NULL);
	    /*FIXME: do some error checking here*/
	}
    }
    mplex->buf = g_string_erase(mplex->buf, 0, len + 4);
    return TRUE;
}

static gboolean
mplex_reader(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    Mplex mplex = data;
    gchar buf[MPLEX_BUFSIZ];
    gsize bytes;

    switch (g_io_channel_read_chars(source, buf, MPLEX_BUFSIZ, &bytes, NULL)) {
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_ERROR:
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	default:
	    return TRUE;
    }
    mplex->buf = g_string_append_len(mplex->buf, buf, bytes);
    while (mplex_parse(mplex));
    return TRUE;
}

static void
mplex_io_closed(gpointer data) {
    Channel *chan = data;
    gchar buf[4];
    gsize written;
    GError *e = NULL;

    if (chan->mplex->is_closing == TRUE) return;
    chan->reader = 0;
    buf[0] = chan->chan >> 8;
    buf[1] = chan->chan & 0xff;
    buf[2] = 0;
    buf[3] = 0;
    switch (g_io_channel_write_chars(chan->mplex->io, buf,
		4, &written, &e)) {
	case G_IO_STATUS_AGAIN:
	    g_printerr(_("warning: dropping data for mplex channel\n"));
	    /*retry ?*/
	    return;
	case G_IO_STATUS_EOF:
	    return;
	case G_IO_STATUS_ERROR:
	    g_printerr(_("error writing to mplex channel (%s)\n"), e->message);
	    g_clear_error(&e);
	    return;
	default:
	    break;
    }
    if (4 != written) {
	g_printerr(_("warning: short write on mplex channel\n"));
    }
}

static gboolean
mplex_encoder(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    Channel *chan = data;
    gchar buf[MPLEX_BUFSIZ + 4];
    gsize bytes, written;
    GError *e = NULL;

    switch (g_io_channel_read_chars(source, &buf[4],
		MPLEX_BUFSIZ, &bytes, &e)) {
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_ERROR:
	    g_printerr(_("mplex encoder read error (%s)\n"), e->message);
	    g_clear_error(&e);
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	default:
	    return TRUE;
    }
    buf[0] = chan->chan >> 8;
    buf[1] = chan->chan & 0xff;
    buf[2] = bytes >> 8;
    buf[3] = bytes & 0xff;
    switch (g_io_channel_write_chars(chan->mplex->io, buf,
		bytes + 4, &written, &e)) {
	case G_IO_STATUS_AGAIN:
	    g_printerr(_("warning: dropping data for mplex channel\n"));
	    /*retry ?*/
	    return TRUE;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_ERROR:
	    g_printerr(_("error writing to mplex channel (%s)\n"), e->message);
	    g_clear_error(&e);
	    return FALSE;
	default:
	    break;
    }
    if (bytes + 4 != written) {
	g_printerr(_("warning: short write on mplex channel\n"));
    }
    return TRUE;
}

extern Mplex
mplex_new(GIOChannel *io,
	MplexClosedCB closed_cb,
	gpointer user_data,
	GError **error) {
    Mplex mplex = g_new0(struct _Mplex, 1);

    g_assert(io);
    g_assert(closed_cb);

    mplex->buf = g_string_new("");
    mplex->io = io; 
    if (g_io_channel_get_encoding(mplex->io) != NULL &&
	    g_io_channel_set_encoding(mplex->io, NULL, error) !=
	    G_IO_STATUS_NORMAL) {
	g_string_free(mplex->buf, TRUE);
	g_free(mplex);
	return NULL;
    }
    g_io_channel_set_buffered(mplex->io, FALSE);
    g_io_channel_ref(mplex->io);
    mplex->is_running = FALSE;
    mplex->is_closing = FALSE;
    mplex->closed_cb = closed_cb;
    mplex->user_data = user_data;
    return mplex;
}

extern void
mplex_start(Mplex mplex) {
    g_assert(mplex);
    if (mplex->is_running == TRUE) {
	return;
    }
    mplex->reader = g_io_add_watch_full(mplex->io,
	    G_PRIORITY_DEFAULT, G_IO_IN, mplex_reader, mplex, mplex_closed);
    mplex->is_running = TRUE;
}

extern void
mplex_free(Mplex mplex) {
    MplexChannel i;

    g_assert(mplex);
    mplex->is_closing = TRUE;
    if (mplex->is_running == TRUE) {
	g_source_remove(mplex->reader);
    }
    for (i = 0 ; ((guint)i + 1) <= G_MAXUINT16 ; i++) {
	if (mplex->chan[i] == NULL) {
	    continue;
	}
	if (mplex->chan[i]->reader != 0) {
	    g_source_remove(mplex->chan[i]->reader);
	}
	if (mplex->chan[i]->io != NULL) {
	    g_io_channel_unref(mplex->chan[i]->io);
	}
	g_free(mplex->chan[i]);
    }
    g_string_free(mplex->buf, TRUE);
    g_io_channel_unref(mplex->io);
    g_free(mplex);
}

extern GIOChannel *
mplex_open(Mplex mplex,
	MplexChannel chan,
	MplexChannel *retchan,
	MplexChannelClosedCB closed_cb,
	gpointer user_data,
	GError **error) {
    GIOChannel *pair[2];

    g_assert(mplex);
    if (chan == 0) {
	chan = mplex_next_chan(mplex);
    }
    mplex->last = chan;
    if (retchan != NULL) {
	*retchan = chan;
    }
    mplex->chan[chan] = g_new0(Channel, 1);
    mplex->chan[chan]->mplex = mplex;
    mplex->chan[chan]->chan = chan;
    mplex->chan[chan]->closed_cb = closed_cb;
    mplex->chan[chan]->user_data = user_data;
    pipe_new(pair);
    mplex->chan[chan]->io = pair[1];
    mplex->chan[chan]->reader =
	g_io_add_watch_full(mplex->chan[chan]->io, G_PRIORITY_DEFAULT,
		G_IO_IN, mplex_encoder, mplex->chan[chan], mplex_io_closed);
    return pair[0];
}

extern gboolean
mplex_connect(Mplex mplex,
	MplexChannel chan,
	MplexChannel *retchan,
	GIOChannel *io,
	MplexChannelClosedCB closed_cb,
	gpointer user_data,
	GError **error) {
    g_assert(mplex);
    g_assert(io);
    if (chan == 0) {
	chan = mplex_next_chan(mplex);
    }
    mplex->last = chan;
    if (retchan != NULL) {
	*retchan = chan;
    }
    mplex->chan[chan] = g_new0(Channel, 1);
    mplex->chan[chan]->mplex = mplex;
    mplex->chan[chan]->chan = chan;
    mplex->chan[chan]->io = io;
    mplex->chan[chan]->closed_cb = closed_cb;
    mplex->chan[chan]->user_data = user_data;
    if (g_io_channel_get_encoding(mplex->chan[chan]->io) != NULL) {
	if (g_io_channel_set_encoding(mplex->chan[chan]->io, NULL, error) !=
	    G_IO_STATUS_NORMAL) {
	    g_io_channel_unref(mplex->chan[chan]->io);
	    g_free(mplex->chan[chan]);
	    return FALSE;
	}
    }
    g_io_channel_set_buffered(mplex->chan[chan]->io, FALSE);
    if ((g_io_channel_get_flags(mplex->chan[chan]->io) &
		G_IO_FLAG_IS_READABLE) != 0) {
	mplex->chan[chan]->reader = g_io_add_watch_full(mplex->chan[chan]->io,
		G_PRIORITY_DEFAULT, G_IO_IN, mplex_encoder,
		mplex->chan[chan], mplex_io_closed);
    } else {
	mplex->chan[chan]->reader = 0;
    }
    return TRUE;
}

extern gboolean
mplex_close(Mplex mplex,
	MplexChannel chan,
	GError **error) {
    Channel *c;

    g_assert(mplex);
    g_assert(chan > 0);

    c = mplex->chan[chan];
    if (c == NULL) {
	return TRUE;
    }
    mplex->chan[chan] = NULL;
    if (c->closed_cb != NULL) {
	c->closed_cb(mplex, chan, c->user_data);
    }
    if (c->reader > 0) {
	g_source_remove(c->reader);
    }
    if (c->io != NULL) {
	g_io_channel_shutdown(c->io, TRUE, NULL);
	g_io_channel_unref(c->io);
    }
    g_free(c);
    return TRUE;
}

extern gboolean
mplex_notify(Mplex mplex,
	MplexChannel chan,
	gboolean *active_flag,
	GError **error) {
    Channel *c;

    g_assert(mplex);
    g_assert(chan > 0);

    c = mplex->chan[chan];
    g_assert(c);
    c->flag = active_flag;
    return TRUE;
}

