#ifndef __MPLEX_H
#define __MPLEX_H

#include "glib.h"

/*
 * Mplex - an opaque structure representing a multiplexed channel
 *
 */
typedef struct _Mplex *Mplex;

/*
 * MplexChannel - an identifier of a channel that is multiplexed
 * 
 * Zero has a special meaning - usually representing the next available
 * channel.
 *
 */
typedef guint16 MplexChannel;

/*
 * MplexClosedCB - a callback called when an mplex is closed for any reason.
 *
 * Usually the callback will execute mplex_free()
 */
typedef void (*MplexClosedCB) (Mplex mplex, gpointer user_data);

/*
 * MplexChannelClosedCB - a callback called when an mplex channel is closed
 * for any reason.
 *
 */
typedef void (*MplexChannelClosedCB) (Mplex mplex,
	MplexChannel chan,
	gpointer user_data);

/*
 * mplex_new - open a new encrypted multiplexed channel
 *
 * Sets up a new multiplexed channel on the supplied GIOChannel.
 * Nothing can be read from Mplex until it has been started using
 * mplex_start(). This is to allow channels to be opened before the
 * first read.
 * Returns with a valid Mplex, or fails setting error.
 *
 */
extern Mplex mplex_new(GIOChannel *io,
	MplexClosedCB closed_cb,
	gpointer user_data,
	GError **error);

/*
 * mplex_start - start the multiplexed channel
 *
 * No data will be read from the channel until this fucntion is called.
 */
extern void mplex_start(Mplex mplex);

/*
 * mplex_free - close a mulitplex channel and free all its resources
 *
 */
extern void mplex_free(Mplex mplex);

/*
 * mplex_open - open a new channel
 *
 * Opens a channel on the mplex using channel number chan.
 * if chan is 0, the next available channel is used.
 * The channel is returned in retchan (if it is not NULL).
 * Returns an IOChannel for the channel or fails, setting an error.
 *
 */
extern GIOChannel * mplex_open(Mplex mplex,
	MplexChannel chan,
	MplexChannel *retchan,
	MplexChannelClosedCB closed_cb,
	gpointer user_data,
	GError **error);

/*
 * mplex_connect - connect a channel to an GIOChannel
 *
 * Opens a channel and connects it to io. if chan is 0, the next available
 * channel is used. The channel is returned in retchan (if it is not NULL).
 * Returns TRUE if it succeds or fails, setting an error.
 *
 */
extern gboolean mplex_connect(Mplex mplex,
	MplexChannel chan,
	MplexChannel *retchan,
	GIOChannel *io,
	MplexChannelClosedCB closed_cb,
	gpointer user_data,
	GError **error);

/*
 * mplex_close - close an mplex chan
 *
 * Closes a channel previously opened by mplex_open() or mplex_connect(), and
 * frees that channel for reuse.
 * Returns TRUE if successful or fails, setting an error.
 *
 */
extern gboolean mplex_close(Mplex mplex,
	MplexChannel chan,
	GError **error);

/*
 * mplex_notify - request notification of activity on a channel
 *
 * Any data read from the mplex for the given channel results in the
 * active flag being set. This continues until the channel is closed,
 * or mplex_notify is called with a flag address of NULL.
 * Returns TRUE if successful or fails, setting an error.
 *
 */
extern gboolean mplex_notify(Mplex mplex,
	MplexChannel chan,
	gboolean *active_flag,
	GError **error);

#endif
