#include <string.h>
#include "pipe.h"

#define PAIR_BUFSIZ 102400

typedef struct _Pair Pair;

typedef struct _PairChan {
    GIOChannel pad;
    Pair *pair;
    gboolean first;
} PairChan;

struct _Pair {
    PairChan *first;
    PairChan *second;
    GString *first_buf;
    GString *second_buf;
    gboolean is_open;
};

typedef struct _PairChanSource {
    GSource pad;
    PairChan *pc;
    GIOCondition mask;
    GIOCondition cond;
} PairChanSource;

static gboolean
pairchan_check(GSource *source) {
    PairChanSource *pcs = (PairChanSource *)source;
    GIOCondition cond;

    cond = G_IO_OUT;
    if (!pcs->pc->pair->is_open) {
	cond |= G_IO_IN;
    }
    if (pcs->pc->first) {
	if (pcs->pc->pair->second_buf->len > 0) {
	    cond |= G_IO_IN;
	}
    } else {
	if (pcs->pc->pair->first_buf->len > 0) {
	    cond |= G_IO_IN;
	}
    }
    pcs->cond = pcs->mask & cond;
    return (pcs->cond != 0);
}

static gboolean
pairchan_prepare(GSource *source,
	gint *timeout) {
    return pairchan_check(source);
}

static gboolean
pairchan_dispatch(GSource *source,
	GSourceFunc callback,
	gpointer user_data) {
    PairChanSource *pcs = (PairChanSource *)source;

    return ((GIOFunc)callback)((GIOChannel *)pcs->pc, pcs->cond, user_data);
}

static void
pairchan_finalize(GSource *source) {
}

static GSourceFuncs pairchan_source_funcs = {
    pairchan_prepare,
    pairchan_check,
    pairchan_dispatch,
    pairchan_finalize
};

static GIOStatus
pairchan_read(GIOChannel *channel,
	gchar *buf,
	gsize count,
	gsize *bytes_read,
	GError **err) {
    PairChan *pc = (PairChan *)channel;
    GString **b;

    g_assert(channel);
    g_assert(buf);
    if (pc->first) {
	b = &pc->pair->second_buf;
    } else {
	b = &pc->pair->first_buf;
    }
    if ((*b)->len == 0 && !pc->pair->is_open) {
	return G_IO_STATUS_EOF;
    }
    if (count > (*b)->len) { 
	count = (*b)->len;
    }
    g_memmove(buf, (*b)->str, count);
    if (bytes_read != NULL) {
	*bytes_read = count;
    }
    *b = g_string_erase(*b, 0, count);
    return G_IO_STATUS_NORMAL;
}

static GIOStatus
pairchan_write(GIOChannel *channel,
	const gchar *buf,
	gsize count,
	gsize *bytes_written,
	GError **err) {
    PairChan *pc = (PairChan *)channel;
    GString **b;

    g_assert(channel);
    g_assert(buf);
    if (!pc->pair->is_open) {
	return G_IO_STATUS_EOF;
    }
    if (pc->first) {
	b = &pc->pair->first_buf;
    } else {
	b = &pc->pair->second_buf;
    }
    if (count == -1) {
	count = strlen(buf);
    }
    if (((*b)->len + count) > PAIR_BUFSIZ) {
	/* FIXME: set err */
	return G_IO_STATUS_ERROR;
    }
    *b = g_string_append_len(*b, buf, count);
    if (bytes_written != NULL) {
	*bytes_written = count;
    }
    return G_IO_STATUS_NORMAL;
}

static GIOStatus
pairchan_seek(GIOChannel *channel,
	gint64 offset,
	GSeekType type,
	GError **err) {
    g_assert(channel);
    /* FIXME: set err */
    return G_IO_STATUS_ERROR;
}

static GIOStatus
pairchan_close(GIOChannel *channel,
	GError **err) {
    PairChan *pc = (PairChan *)channel;

    g_assert(channel);
    pc->pair->is_open = FALSE;
    return G_IO_STATUS_NORMAL;
}

static GSource *
pairchan_create_watch(GIOChannel *channel,
	GIOCondition condition) {
    PairChan *pc = (PairChan *)channel;
    GSource *source = g_source_new(&pairchan_source_funcs,
	    sizeof(PairChanSource));
    PairChanSource *pcs = (PairChanSource *)source;

    g_assert(channel);
    pcs->pc = pc;
    pcs->mask = condition;
    return source;
}

static void
pairchan_free(GIOChannel *channel) {
    PairChan *pc = (PairChan *)channel;

    g_assert(channel);
    if (pc->first) {
	pc->pair->first = NULL;
    } else {
	pc->pair->second = NULL;
    }
    if (pc->pair->first == NULL && pc->pair->second == NULL) {
	g_string_free(pc->pair->first_buf, TRUE);
	g_string_free(pc->pair->second_buf, TRUE);
	g_free(pc->pair);
    }
    g_free(pc);
}

static GIOStatus
pairchan_set_flags(GIOChannel *channel,
	GIOFlags flags,
	GError **err) {
    g_assert(channel);
    return G_IO_STATUS_NORMAL;
}

static GIOFlags
pairchan_get_flags(GIOChannel *channel) {
    g_assert(channel);
    return (G_IO_FLAG_IS_READABLE | G_IO_FLAG_IS_WRITEABLE);
}

static GIOFuncs pairchan_funcs = {
    pairchan_read,
    pairchan_write,
    pairchan_seek,
    pairchan_close,
    pairchan_create_watch,
    pairchan_free,
    pairchan_set_flags,
    pairchan_get_flags
};

static void
pairchan_new(Pair *p,
	gboolean first) {
    PairChan *pc = g_new0(PairChan, 1);
    GIOChannel *chan = (GIOChannel *)pc;

    if (first) {
	p->first = pc;
    } else {
	p->second = pc;
    }
    pc->pair = p;
    pc->first = first;
    chan->funcs = &pairchan_funcs;
    chan->is_writeable = TRUE;
    chan->is_readable = TRUE;
    g_io_channel_init(chan);
    g_io_channel_set_encoding(chan, NULL, NULL);
    g_io_channel_set_buffered(chan, FALSE);

}

extern void
pipe_new(GIOChannel *pair[2]) {
    Pair *p = g_new0(Pair, 1);

    p->first_buf = g_string_new("");
    p->second_buf = g_string_new("");
    p->is_open = TRUE;
    pairchan_new(p, TRUE);
    pairchan_new(p, FALSE);
    pair[0] = (GIOChannel *) p->first;
    pair[1] = (GIOChannel *) p->second;
}
