#include <string.h>
#include "spec-console.h"

#define SPEC_CONSOLE_NAME "SpecConsole_name"
#define SPEC_CONSOLE_SERVER "SpecConsole_server"
#define SPEC_CONSOLE_TYPE "SpecConsole_type"
#define SPEC_CONSOLE_TERMINAL "SpecConsole_terminal"
#define SPEC_CONSOLE_IPMI "SpecConsole_ipmi"

extern void
spec_console_free(SpecConsole *sc) {
    g_assert(sc);
    if (sc->name) {
	g_free(sc->name);
    }
    if (sc->server) {
	g_free(sc->server);
    }
    if (sc->type == CONSOLE_TYPE_SERIAL && sc->terminal) {
	terminal_free(sc->terminal);
    }
    if (sc->type == CONSOLE_TYPE_IPMI && sc->ipmi) {
	ipmi_free(sc->ipmi);
    }
    g_free(sc);
}

extern SpecConsole *
spec_console_serial_new(const gchar *name,
	const gchar *server,
	const gchar *device,
	guint speed,
	const gchar *bits) {
    SpecConsole *sc = g_new0(struct _SpecConsole, 1);

    g_assert(name);
    g_assert(server);
    g_assert(device);
    g_assert(bits);
    sc->name = g_strdup(name);
    sc->server = g_strdup(server);
    sc->type = CONSOLE_TYPE_SERIAL;
    if (!strcmp(bits, "7E2")) {
	sc->terminal = terminal_new(device, speed, speed, 7, PARITY_EVEN, 2);
    } else {
	sc->terminal = terminal_new(device, speed, speed, 8, PARITY_NONE, 1);
    }
    return sc;
}

extern SpecConsole *
spec_console_ipmi_new(const gchar *name,
	const gchar *server,
	const gchar *host,
	const gchar *user,
	const gchar *passwd) {
    SpecConsole *sc = g_new0(struct _SpecConsole, 1);

    g_assert(name);
    g_assert(server);
    g_assert(host);
    g_assert(passwd);
    sc->name = g_strdup(name);
    sc->server = g_strdup(server);
    sc->type = CONSOLE_TYPE_IPMI;
    sc->ipmi = ipmi_new(host, user, passwd);
    return sc;
}

extern MessageArg *
spec_console_message_arg(const gchar *name,
	const SpecConsole *sc) {
    MessageArg *a;

    g_assert(name);
    g_assert(sc);
    a = message_arg_list(name, NULL);
    a->list = g_list_prepend(a->list,
	    message_arg_string(SPEC_CONSOLE_NAME, sc->name?sc->name:""));
    a->list = g_list_prepend(a->list,
	    message_arg_string(SPEC_CONSOLE_SERVER, sc->server?sc->server:""));
    a->list = g_list_prepend(a->list,
	    message_arg_uint8(SPEC_CONSOLE_TYPE, sc->type));
    switch (sc->type) {
	case CONSOLE_TYPE_SERIAL:
	    a->list = g_list_prepend(a->list,
		    terminal_message_arg(SPEC_CONSOLE_TERMINAL, sc->terminal));
	    break;
	case CONSOLE_TYPE_IPMI:
	    a->list = g_list_prepend(a->list,
		    ipmi_message_arg(SPEC_CONSOLE_IPMI, sc->ipmi));
	    break;
	default:
	    g_assert_not_reached();
    }
    return a;
}

extern SpecConsole *
spec_console_new_from_arg(const MessageArg *arg) {
    MessageArg *name;
    MessageArg *server;
    MessageArg *type;
    MessageArg *ipmi;
    MessageArg *terminal;
    SpecConsole *sc;
    Terminal *t = NULL;
    IPMI *i = NULL;

    g_assert(arg);
    if (arg->type != MESSAGE_ARG_LIST) return NULL;
    if (message_arg_list_explode(arg,
		SPEC_CONSOLE_NAME, MESSAGE_ARG_STRING, &name,
		SPEC_CONSOLE_SERVER, MESSAGE_ARG_STRING, &server,
		SPEC_CONSOLE_TYPE, MESSAGE_ARG_UINT8, &type,
		NULL, 0, NULL) == FALSE) {
	return NULL;
    }
    switch (type->uint8) {
	case CONSOLE_TYPE_SERIAL:
	    if (message_arg_list_explode(arg,
			SPEC_CONSOLE_TERMINAL, MESSAGE_ARG_LIST, &terminal,
			NULL, 0, NULL) == FALSE) {
		return NULL;
	    }
	    t = terminal_new_from_arg(terminal);
	    if (t == NULL) {
		return NULL;
	    }
	    break;
	case CONSOLE_TYPE_IPMI:
	    if (message_arg_list_explode(arg,
			SPEC_CONSOLE_IPMI, MESSAGE_ARG_LIST, &ipmi,
			NULL, 0, NULL) == FALSE) {
		return NULL;
	    }
	    i = ipmi_new_from_arg(ipmi);
	    if (i == NULL) {
		return NULL;
	    }
	    break;
	default:
	    return NULL;
    }
    sc = g_new0(struct _SpecConsole, 1);
    sc->name = g_strdup(name->string->str);
    sc->server = g_strdup(server->string->str);
    sc->type = type->uint8;
    switch (sc->type) {
	case CONSOLE_TYPE_SERIAL:
	    sc->terminal = t;
	    break;
	case CONSOLE_TYPE_IPMI:
	    sc->ipmi = i;
	    break;
	default:
	    g_assert_not_reached();
    }
    return sc;
}
