#include <string.h>
#include <errno.h>
#include <gnutls/gnutls.h>
#include "ssl.h"

#define DH_BITS 1024
#define SSL_BUFSIZ 4096

typedef struct _SSLChannel {
    GIOChannel pad;
    GIOChannel *io;
    gnutls_session session;
    gnutls_connection_end_t type;
    gboolean established;
    union {
	gnutls_anon_client_credentials_t canoncred;
	gnutls_anon_server_credentials_t sanoncred;
    };
} SSLChannel;

typedef struct _SSLChannelSource {
    GSource pad;
    SSLChannel *ssl;
    GPollFD poll;
} SSLChannelSource;


static gboolean
ssl_channel_source_check(GSource *source) {
    SSLChannelSource *s = (SSLChannelSource *)source;

    return (s->poll.events & s->poll.revents) > 0;
}

static gboolean
ssl_channel_source_prepare(GSource *source,
	gint *timeout) {
    return FALSE;
}

static gboolean
ssl_channel_source_dispatch(GSource *source,
	GSourceFunc callback,
	gpointer user_data) {
    SSLChannelSource *s = (SSLChannelSource *)source;

    return ((GIOFunc)callback)((GIOChannel *)s->ssl,
	    s->poll.events & s->poll.revents, user_data);
}

static void
ssl_channel_source_finalise(GSource *source) {
}

static GSourceFuncs ssl_channel_source_funcs = {
    ssl_channel_source_prepare,
    ssl_channel_source_check,
    ssl_channel_source_dispatch,
    ssl_channel_source_finalise
};

static GIOStatus
perform_handshake(SSLChannel *ssl, GError **error) {
    int res;

retry:
    res = gnutls_handshake(ssl->session);
    if (res == GNUTLS_E_AGAIN || res == GNUTLS_E_INTERRUPTED) {
	goto retry;
	return G_IO_STATUS_AGAIN;
    }
    if (res < 0) {
	g_set_error(error, G_IO_CHANNEL_ERROR, G_IO_CHANNEL_ERROR_FAILED,
		"gnutls: %s", gnutls_strerror(res));
	return G_IO_STATUS_ERROR;
    }
    return G_IO_STATUS_NORMAL;
}

static GIOStatus
ssl_channel_read(GIOChannel *channel,
	gchar *buf,
	gsize count,
	gsize *bytes_read,
	GError **error) {
    SSLChannel *ssl = (SSLChannel *)channel;
    gint res;

    if (ssl->io == NULL) return G_IO_STATUS_EOF;
restart:
    if (ssl->established == FALSE) {
	res = perform_handshake(ssl, error);
	if (res != G_IO_STATUS_NORMAL) {
	    return res;
	}
	ssl->established = TRUE;
    }

    res = gnutls_record_recv(ssl->session, buf, count);

    if (res == GNUTLS_E_REHANDSHAKE) {
	ssl->established = FALSE;
	goto restart;
    }
    if (res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN) {
	return G_IO_STATUS_AGAIN;
    }
    if (res < 0) {
	g_set_error(error, G_IO_CHANNEL_ERROR, G_IO_CHANNEL_ERROR_FAILED,
		"gnutls_record_recv failed");
	return G_IO_STATUS_ERROR;
    }
    *bytes_read = res;
    return (res > 0) ? G_IO_STATUS_NORMAL : G_IO_STATUS_EOF;
}

static GIOStatus
ssl_channel_write(GIOChannel *channel,
	const gchar *buf,
	gsize count,
	gsize *bytes_written,
	GError **error) {
    SSLChannel *ssl = (SSLChannel *)channel;
    gint res;


    if (ssl->io == NULL) return G_IO_STATUS_EOF;
restart:
    if (ssl->established == FALSE) {
	res = perform_handshake(ssl, error);
	if (res != G_IO_STATUS_NORMAL) {
	    return res;
	}
	ssl->established = TRUE;
    }

    res = gnutls_record_send(ssl->session, buf, count);

    if (res == GNUTLS_E_REHANDSHAKE) {
	ssl->established = FALSE;
	goto restart;
    }
    if (res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN) {
	return G_IO_STATUS_AGAIN;
    }
    if (res < 0) {
	g_set_error(error, G_IO_CHANNEL_ERROR, G_IO_CHANNEL_ERROR_FAILED,
		"gnutls_record_send failed");
	return G_IO_STATUS_ERROR;
    }
    *bytes_written = res;
    return (res > 0) ? G_IO_STATUS_NORMAL : G_IO_STATUS_EOF;
}

static GIOStatus
ssl_channel_seek(GIOChannel *channel,
	gint64 offset,
	GSeekType type,
	GError **error) {
    SSLChannel *ssl = (SSLChannel *)channel;

    if (ssl->io == NULL) return G_IO_STATUS_EOF;
    return ssl->io->funcs->io_seek(ssl->io, offset, type, error);
}

static GIOStatus
ssl_channel_close(GIOChannel *channel,
	GError **error) {
    SSLChannel *ssl = (SSLChannel *)channel;

    if (ssl->io == NULL) return G_IO_STATUS_EOF;
    while (gnutls_bye(ssl->session, GNUTLS_SHUT_WR) == GNUTLS_E_INTERRUPTED);
    gnutls_deinit(ssl->session);
    if (ssl->type == GNUTLS_CLIENT) {
	gnutls_anon_free_client_credentials(ssl->canoncred);
    } else {
	gnutls_anon_free_server_credentials(ssl->sanoncred);
    }
    g_io_channel_unref(ssl->io);
    ssl->io = NULL;
    return G_IO_STATUS_NORMAL;
}

static GSource *
ssl_channel_create_watch (GIOChannel *channel,
	GIOCondition condition) {
    SSLChannel *ssl = (SSLChannel *)channel;
    GSource *source;
    SSLChannelSource *s;
	
    source = g_source_new(&ssl_channel_source_funcs,
	    sizeof(struct _SSLChannelSource));
    s = (SSLChannelSource *)source;
    s->ssl = ssl;
    s->poll.fd = g_io_channel_unix_get_fd(s->ssl->io);
    s->poll.events = condition;
    s->poll.revents = 0;
    g_source_add_poll(source, &s->poll);
    return source;
}

static void
ssl_channel_free(GIOChannel *channel) {
    SSLChannel *ssl = (SSLChannel *)channel;

    ssl_channel_close(channel, NULL);
    g_free(ssl);
}

static GIOStatus
ssl_channel_set_flags(GIOChannel *channel,
	GIOFlags flags,
	GError **error) {
    SSLChannel *ssl = (SSLChannel *)channel;

    if (ssl->io == NULL) return G_IO_STATUS_EOF;
    return ssl->io->funcs->io_set_flags(ssl->io, flags, error);
}

static GIOFlags
ssl_channel_get_flags(GIOChannel *channel) {
    SSLChannel *ssl = (SSLChannel *)channel;

    if (ssl->io == NULL) return 0;
    return ssl->io->funcs->io_get_flags(ssl->io);
}

static GIOFuncs ssl_channel_funcs = {
    ssl_channel_read,
    ssl_channel_write,
    ssl_channel_seek,
    ssl_channel_close,
    ssl_channel_create_watch,
    ssl_channel_free,
    ssl_channel_set_flags,
    ssl_channel_get_flags,
};

static GIOChannel *
ssl_create(GIOChannel *ssl_io, gnutls_connection_end_t type) {
    GIOChannel *gchan;
    SSLChannel *ssl;
    static gboolean initialised = FALSE;
    static gnutls_dh_params_t dh_params;
    gint res = 0;

    if (initialised == FALSE) {
	gnutls_global_init();
	gnutls_dh_params_init(&dh_params);
	gnutls_dh_params_generate2(dh_params, DH_BITS);
	initialised = TRUE;
    }
    ssl = g_new0(struct _SSLChannel, 1);
    gchan = (GIOChannel *)ssl;
    ssl->io = ssl_io;
    ssl->type = type;
    if (g_io_channel_get_encoding(ssl->io) != NULL &&
	    g_io_channel_set_encoding(ssl->io, NULL, NULL) !=
	    G_IO_STATUS_NORMAL) goto sslerror;
    g_io_channel_set_buffered(ssl->io, FALSE);
    ssl->established = FALSE;
    res = gnutls_init(&ssl->session, type);
    if (res != 0) goto sslerror;
    res = gnutls_priority_set_direct(ssl->session, "NORMAL:+ANON-DH", NULL);
    if (res != 0) goto sslerror;
    if (type == GNUTLS_CLIENT) {
	res = gnutls_anon_allocate_client_credentials(&ssl->canoncred);
	if (res != 0) goto sslerror;
	res = gnutls_credentials_set(ssl->session, GNUTLS_CRD_ANON,
		ssl->canoncred);
	if (res != 0) goto sslerror;
	gnutls_dh_set_prime_bits(ssl->session, DH_BITS);
    } else {
	res = gnutls_anon_allocate_server_credentials(&ssl->sanoncred);
	if (res != 0) goto sslerror;
	gnutls_anon_set_server_dh_params(ssl->sanoncred, dh_params);
	res = gnutls_credentials_set(ssl->session, GNUTLS_CRD_ANON,
		ssl->sanoncred);
	if (res != 0) goto sslerror;
    }
    gnutls_transport_set_ptr(ssl->session,
	    GINT_TO_POINTER(g_io_channel_unix_get_fd(ssl->io)));
    g_io_channel_ref(ssl->io);
    gchan->funcs = &ssl_channel_funcs;
    g_io_channel_init(gchan);
    gchan->is_readable = TRUE;
    gchan->is_writeable = TRUE;
    gchan->use_buffer = FALSE;
    gchan->encoding = NULL;
    return gchan;
sslerror:
    if (ssl->session != NULL) {
	gnutls_deinit(ssl->session);
    }
    if (ssl->type == GNUTLS_CLIENT) {
	if (ssl->canoncred != NULL) {
	    gnutls_anon_free_client_credentials(ssl->canoncred);
	}
    } else {
	if (ssl->sanoncred != NULL) {
	    gnutls_anon_free_server_credentials(ssl->sanoncred);
	}
    }
    if (res != 0) {
	g_printerr("gnutls: %s\n", gnutls_strerror(res));
    }
    g_free(ssl);
    return NULL;
}

extern GIOChannel *
ssl_client(GIOChannel *ssl_io) {
    return ssl_create(ssl_io, GNUTLS_CLIENT);
}

extern GIOChannel *
ssl_server(GIOChannel *ssl_io) {
    return ssl_create(ssl_io, GNUTLS_SERVER);
}

