#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "stdinout.h"

typedef struct _StdInOut {
    GIOChannel pad;
    gint stdin_fd;
    gint stdout_fd;
} StdInOut;

typedef struct _StdInOutSource {
    GSource pad;
    StdInOut *c;
    GPollFD poll_in;
    GPollFD poll_out;
} StdInOutSource;

static gboolean
stdinout_check(GSource *source) {
    StdInOutSource *cs = (StdInOutSource *)source;

    return ((cs->poll_in.revents | cs->poll_out.revents) != 0);
}

static gboolean
stdinout_prepare(GSource *source,
	gint *timeout) {
    *timeout = -1;
    return FALSE;
}

static gboolean
stdinout_dispatch(GSource *source,
	GSourceFunc callback,
	gpointer user_data) {
    StdInOutSource *cs = (StdInOutSource *)source;

    return ((GIOFunc)callback)((GIOChannel *)cs->c,
	    cs->poll_in.revents | cs->poll_out.revents,
	    user_data);
}

static void
stdinout_finalize(GSource *source) {
    StdInOutSource *cs = (StdInOutSource *)source;

    g_io_channel_unref((GIOChannel *)cs->c);
}

static GSourceFuncs stdinout_source_funcs = {
    stdinout_prepare,
    stdinout_check,
    stdinout_dispatch,
    stdinout_finalize
};

static GIOStatus
stdinout_read(GIOChannel *channel,
	gchar *buf,
	gsize count,
	gsize *bytes_read,
	GError **err) {
    StdInOut *c = (StdInOut *)channel;
    gsize bytes;

    g_assert(channel);
    g_assert(buf);
retry:
    bytes = read(c->stdin_fd, buf, count);
    if (bytes == -1) {
	if (bytes_read != NULL) {
	    *bytes_read = 0;
	}
	switch (errno) {
	    case EINTR:
		goto retry;
	    case EAGAIN:
		return G_IO_STATUS_AGAIN;
	    default:
		g_set_error(err, G_IO_CHANNEL_ERROR,
			g_io_channel_error_from_errno(errno),
			g_strerror(errno));
	}
	return G_IO_STATUS_ERROR;
    }
    if (bytes_read != NULL) {
	*bytes_read = bytes;
    }
    if (bytes == 0) {
	return G_IO_STATUS_EOF;
    }
    return G_IO_STATUS_NORMAL;
}

static gint
error_from_errno(gint e) {
    switch (e) {
	case EFBIG:
	    return G_IO_CHANNEL_ERROR_FBIG;
	case EINVAL:
	    return G_IO_CHANNEL_ERROR_INVAL;
	case EIO:
	    return G_IO_CHANNEL_ERROR_IO;
	case EISDIR:
	    return G_IO_CHANNEL_ERROR_ISDIR;
	case ENOSPC:
	    return G_IO_CHANNEL_ERROR_NOSPC;
	case ENXIO:
	    return G_IO_CHANNEL_ERROR_NXIO;
	case EOVERFLOW:
	    return G_IO_CHANNEL_ERROR_OVERFLOW;
	case EPIPE:
	    return G_IO_CHANNEL_ERROR_PIPE;
    }
    return G_IO_CHANNEL_ERROR_FAILED;
}

static GIOStatus
stdinout_write(GIOChannel *channel,
	const gchar *buf,
	gsize count,
	gsize *bytes_written,
	GError **err) {
    StdInOut *c = (StdInOut *)channel;
    gsize bytes;

    g_assert(channel);
    g_assert(buf);
retry:
    bytes = write(c->stdout_fd, buf, count);
    if (bytes == -1) {
	if (bytes_written != NULL) {
	    *bytes_written = 0;
	}
	switch (errno) {
	    case EINTR:
		goto retry;
	    case EAGAIN:
		return G_IO_STATUS_AGAIN;
	    default:
		g_set_error(err, G_IO_CHANNEL_ERROR,
			error_from_errno(errno),
			g_strerror(errno));
	}
	return G_IO_STATUS_ERROR;
    }
    if (bytes_written != NULL) {
	*bytes_written = bytes;
    }
    return G_IO_STATUS_NORMAL;
}

static GIOStatus
stdinout_seek(GIOChannel *channel,
	gint64 offset,
	GSeekType type,
	GError **err) {
    g_assert(channel);
    /* FIXME: set err */
    return G_IO_STATUS_ERROR;
}

static GIOStatus
stdinout_close(GIOChannel *channel,
	GError **err) {
    StdInOut *c = (StdInOut *)channel;

    g_assert(channel);
    close(c->stdin_fd);
    close(c->stdout_fd);
    return G_IO_STATUS_NORMAL;
}

static GSource *
stdinout_create_watch(GIOChannel *channel,
	GIOCondition condition) {
    StdInOut *c = (StdInOut *)channel;
    GSource *source = g_source_new(&stdinout_source_funcs,
	    sizeof(StdInOutSource));
    StdInOutSource *cs = (StdInOutSource *)source;

    g_assert(channel);
    cs->c = c;
    cs->poll_in.fd = c->stdin_fd;
    cs->poll_in.events = condition & ~G_IO_OUT;
    g_source_add_poll(source, &cs->poll_in);
    cs->poll_out.fd = c->stdout_fd;
    cs->poll_out.events = condition & (G_IO_OUT | G_IO_ERR);
    g_source_add_poll(source, &cs->poll_out);
    return source;
}

static void
stdinout_free(GIOChannel *channel) {
    StdInOut *c = (StdInOut *)channel;

    g_assert(channel);
    g_free(c);
}

static GIOStatus
stdinout_set_flags(GIOChannel *channel,
	GIOFlags flags,
	GError **err) {
    g_assert(channel);
    return G_IO_STATUS_NORMAL;
}

static GIOFlags
stdinout_get_flags(GIOChannel *channel) {
    g_assert(channel);
    return (G_IO_FLAG_IS_READABLE | G_IO_FLAG_IS_WRITEABLE);
}

static GIOFuncs stdinout_funcs = {
    stdinout_read,
    stdinout_write,
    stdinout_seek,
    stdinout_close,
    stdinout_create_watch,
    stdinout_free,
    stdinout_set_flags,
    stdinout_get_flags
};

extern GIOChannel *
stdinout_new_with_fds(gint stdin_fd,
	gint stdout_fd) {
    StdInOut *c = g_new0(StdInOut, 1);
    GIOChannel *chan = (GIOChannel *)c;

    c->stdin_fd = stdin_fd;
    c->stdout_fd = stdout_fd;
    chan->funcs = &stdinout_funcs;
    chan->is_writeable = TRUE;
    chan->is_readable = TRUE;
    g_io_channel_init(chan);
    g_io_channel_set_encoding(chan, NULL, NULL);
    g_io_channel_set_buffered(chan, FALSE);
    return chan;
}

extern GIOChannel *
stdinout_new(void) {
    return stdinout_new_with_fds(0, 1);
}
