#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include "terminal.h"

static struct {
    guint speed;
    speed_t bspeed;
} terminal_speeds[] = {
    { 0, B0 },
    { 50, B50 },
    { 75, B75 },
    { 110, B110 },
    { 134, B134 },
    { 150, B150 },
    { 200, B200 },
    { 300, B300 },
    { 600, B600 },
    { 1200, B1200 },
    { 1800, B1800 },
    { 2400, B2400 },
    { 4800, B4800 },
    { 9600, B9600 },
    { 19200, B19200 },
    { 38400, B38400 },
    { 57600, B57600 },
    { 115200, B115200 },
    { 230400, B230400 },
    { -1, -1 }
};

static guint
get_speed(guint s) {
    gint i;

    for (i = 1; terminal_speeds[i].speed >= 0 &&
	    terminal_speeds[i].speed < s; i++);
    return terminal_speeds[i].speed;
}

static speed_t
get_bspeed(guint s) {
    gint i;

    for (i = 0; terminal_speeds[i].speed >= 0; i++) {
	if (s == terminal_speeds[i].speed) {
	    return terminal_speeds[i].bspeed;
	}
    }
    g_assert_not_reached();
    return 0;
}

extern Terminal *
terminal_new(const gchar *device,
	guint ispeed,
	guint ospeed,
	guint8 data_bits,
	TerminalParity parity,
	guint8 stop_bits) {
    Terminal *t = g_new0(Terminal, 1);

    g_assert(device);

    t->device = g_strdup(device);
    t->ispeed = get_speed(ispeed);
    t->ospeed = get_speed(ospeed);
    switch (data_bits) {
	case 7:
	    t->data_bits = 7;
	case 9:
	    t->data_bits = 9;
	default:
	    t->data_bits = 8;
    }
    t->parity = parity;
    switch (stop_bits) {
	case 0:
	    t->stop_bits = 0;
	case 2:
	    t->stop_bits = 2;
	default:
	    t->stop_bits = 1;
    }
    return t;
}

extern TerminalParity
terminal_parity(const gchar *parity) {
    g_assert(parity);
    switch (parity[0]) {
	case 'O':
	case 'o':
	    return PARITY_ODD;
	case 'E':
	case 'e':
	    return PARITY_EVEN;
    }
    return PARITY_NONE;
}

extern G_CONST_RETURN gchar *
terminal_parity_string(TerminalParity parity) {
    switch (parity) {
	case PARITY_ODD:
	    return "odd";
	case PARITY_EVEN:
	    return "even";
	case PARITY_NONE:
	    return "none";
	default:
	    g_assert_not_reached();
    }
    return NULL;
}

extern MessageArg *
terminal_message_arg(const gchar *name,
	Terminal *terminal) {
    GList *l = NULL;

    g_assert(name);

    l = g_list_append(l, message_arg_string("device", terminal->device));
    l = g_list_append(l, message_arg_uint32("ispeed", terminal->ispeed));
    l = g_list_append(l, message_arg_uint32("ospeed", terminal->ospeed));
    l = g_list_append(l, message_arg_uint8("data_bits", terminal->ospeed));
    l = g_list_append(l, message_arg_string("parity",
		terminal_parity_string(terminal->parity)));
    l = g_list_append(l, message_arg_uint8("stop_bits", terminal->ospeed));
    return message_arg_list(name, l);
}

extern Terminal *
terminal_new_from_arg(const MessageArg *arg) {
    MessageArg *device, *ispeed, *ospeed, *data_bits, *parity, *stop_bits;

    g_assert(arg);

    if (arg->type != MESSAGE_ARG_LIST) return NULL;

    if (message_arg_list_explode(arg,
		"device", MESSAGE_ARG_STRING, &device,
		"ispeed", MESSAGE_ARG_UINT32, &ispeed,
		"ospeed", MESSAGE_ARG_UINT32, &ospeed,
		"data_bits", MESSAGE_ARG_UINT8, &data_bits,
		"parity", MESSAGE_ARG_STRING, &parity,
		"stop_bits", MESSAGE_ARG_UINT8, &stop_bits,
		NULL, 0, NULL) == FALSE) {
	return NULL;
    }
    return terminal_new(device->string->str, ispeed->uint32,
	    ospeed->uint32, data_bits->uint8,
	    terminal_parity(parity->string->str), stop_bits->uint8);
}

extern gboolean
terminal_match(Terminal *a,
	Terminal *b) {
    g_assert(a);
    g_assert(b);

    if ((a->device == NULL && b->device != NULL) ||
	    (a->device != NULL && b->device == NULL) ||
	    (a->device != NULL && b->device != NULL
	     && strcmp(a->device, b->device))) {
	return FALSE;
    }
    if (a->ispeed != b->ispeed ||
	    a->ospeed != b->ospeed ||
	    a->data_bits != b->data_bits ||
	    a->parity != b->parity ||
	    a->stop_bits != b->stop_bits) {
	return FALSE;
    }
    return TRUE;
}

extern void
terminal_free(Terminal *terminal) {
    g_free(terminal->device);
    g_free(terminal);
}

extern GIOChannel *
terminal_open_io_channel(Terminal *terminal,
	GError **error) {
    GIOChannel *io;
    struct termios tio;
    gint fd;

    g_assert(terminal);

    io = g_io_channel_new_file(terminal->device, "r+", error);
    if (io == NULL) {
	return NULL;
    }
    if (g_io_channel_get_encoding(io) != NULL) {
	g_io_channel_set_encoding(io, NULL, NULL);
    }
    g_io_channel_set_buffered(io, FALSE);
    fd = g_io_channel_unix_get_fd(io);
    if (isatty(fd) == 0) {
	g_io_channel_unref(io);
	g_set_error(error, TERMINAL_ERROR, TERMINAL_ERROR_NOT_TTY,
		"%s is not a tty", terminal->device);
	return NULL;
    }
    if (tcgetattr(fd, &tio) == -1) {
	g_io_channel_unref(io);
	g_set_error(error, TERMINAL_ERROR, TERMINAL_ERROR_GET_ATTR,
		"%s", g_strerror(errno));
	return NULL;
    }
    cfmakeraw(&tio);
    cfsetispeed(&tio, get_bspeed(terminal->ispeed));
    cfsetospeed(&tio, get_bspeed(terminal->ospeed));
    tio.c_cflag &= ~CSIZE;
    tio.c_cflag |= CLOCAL;
    if (terminal->data_bits == 7) {
	tio.c_cflag |= CS7;
    }
    if (terminal->data_bits == 8) {
	tio.c_cflag |= CS8;
    }
    switch (terminal->parity) {
	case PARITY_NONE:
	    tio.c_cflag &= ~PARENB;
	    break;
	case PARITY_ODD:
	    tio.c_cflag |= PARENB | PARODD;
	case PARITY_EVEN:
	    tio.c_cflag |= PARENB;  
	    tio.c_cflag &= ~PARODD;
    }
    if (terminal->stop_bits == 1) {
	tio.c_cflag &= ~CSTOPB;
    }
    if (terminal->stop_bits == 2) {
	tio.c_cflag |= CSTOPB;
    }
    if (tcsetattr(fd, TCSANOW, &tio) == -1) {
	g_io_channel_unref(io);
	g_set_error(error, TERMINAL_ERROR, TERMINAL_ERROR_SET_ATTR,
		"%s", g_strerror(errno));
	return NULL;
    }
    return io;
}

extern GQuark
terminal_error_quark(void) {
    static GQuark error_quark = 0;

    if (error_quark == 0) {
	error_quark = g_quark_from_static_string("terminal-error");
    }
    return error_quark;
}
