#ifndef __TERMINAL_H
#define __TERMINAL_H

#include "message.h"

/*
 * TerminalError - errors occuring when attempting to open a terminal.
 *
 */
typedef enum {
    TERMINAL_ERROR_NOT_TTY,
    TERMINAL_ERROR_GET_ATTR,
    TERMINAL_ERROR_SET_ATTR
} TerminalError;
#define TERMINAL_ERROR terminal_error_quark()
extern GQuark terminal_error_quark(void);

typedef enum {
    PARITY_NONE,
    PARITY_EVEN,
    PARITY_ODD
} TerminalParity;

/*
 * Terminal - a struct representing a terminal
 *
 * This should be considered read-only.
 */
typedef struct _Terminal {
    gchar *device;
    guint ispeed;
    guint ospeed;
    guint8 data_bits;
    TerminalParity parity;
    guint8 stop_bits;
} Terminal;

/*
 * terminal_new - create a terminal structure
 *
 * Any invalid values will be converted to their nearest valid value.
 * The returned Terminal belongs to the caller.
 *
 */
extern Terminal *terminal_new(const gchar *device,
	guint ispeed,
	guint ospeed,
	guint8 data_bits,
	TerminalParity parity,
	guint8 stop_bits);

/*
 * terminal_parity - generate a TerminalParity from a string
 *
 */
extern TerminalParity terminal_parity(const gchar *parity);

/*
 * terminal_parity_string - return a string representing a parity.
 *
 */
extern G_CONST_RETURN gchar *terminal_parity_string(TerminalParity parity);


/*
 * terminal_message_arg - generate a message arg that encapsultes the Terminal.
 *
 * The returned MessageArg belongs to the caller.
 *
 */
extern MessageArg *terminal_message_arg(const gchar *name,
	Terminal *terminal);

/*
 * terminal_new_from_arg - generate a Terminal struct from a MessageArg
 *
 * The argument should have been created with terminal_message_arg().
 * Returns the Terminal, or NULL if the MEssageArg was invalid.
 * The returned Terminal belongs to the caller.
 *
 */
extern Terminal *terminal_new_from_arg(const MessageArg *arg);

/*
 * terminal_match - compare two terminals
 *
 * Returns TRUE if they match.
 *
 */
extern gboolean terminal_match(Terminal *a,
	Terminal *b);

/*
 * terminal_free - frees a terminal structure and all its resources
 *
 */
extern void terminal_free(Terminal *terminal);

/*
 * terminal_open_io_channel - uses a terminal structure to open and configure
 * a terminal device.
 *
 * Returns a GIOChannel connected to the open terminal or fails,
 * returning NULL an setting an error.
 *
 */
extern GIOChannel *terminal_open_io_channel(Terminal *terminal,
	GError **error);


#endif
