#include <string.h>
#include "mgr.h"
#include "client.h"
#include "ssl.h"
#include "mplex.h"
#include "matrix.h"
#include "message.h"
#include "config.h"
#include "client-message.h"
#include "log-query.h"
#include "user.h"
#include "server.h"
#include "console.h"
#include "cgroup.h"
#include "log.h"
#include "server.h"

struct _Client {
    gchar *user; /* NULL indicates unauthenticated */
    Manager *manager;
    GIOChannel *io;
    GIOChannel *ssl;
    Mplex mplex;
    MessageQueue msg_q;
    GTcpSocket *sock;
    gchar *current;
    GList *subscribed; /* list of gchar * */
    GList *attached; /* list of AttachedConsole */
    GList *logsessions; /* list of LogSession */
};

typedef struct _AttachedConsole {
    gchar *name;
    MplexChannel chan;
    GIOChannel *io;
} AttachedConsole;

typedef struct _LogSession {
    Client c;
    Log l;
    LogQuery *lq;
    MessageQueue mq;
    MessageID mid;
} LogSession;

static void client_close(Client c);

static void
client_send_all(Manager *manager,
	const gchar *command,
	const MessageArg *a) {
    GList *l;
    GError *e = NULL;

    for (l = manager->clients; l != NULL; l = g_list_next(l)) {
	Client c = l->data;

	if (message_send(c->msg_q, command, a, NULL, NULL, &e) == FALSE) {
	    g_printerr("could not send message (%s)", e->message);
	    g_clear_error(&e);
	}
    }
}

static void
client_send_status_fe(Client c) {
    GList *l, *console_list;
    MessageArg *groups, *consoles, *arg, *item;
    CGroup cg;
    Console console;
    GError *e = NULL;

    if (c->subscribed == NULL) {
	return;
    }
    groups = message_arg_list("groups", NULL);
    consoles = message_arg_list("consoles", NULL);
    arg = message_arg_list("", NULL);
    arg->list = g_list_prepend(arg->list, consoles);
    arg->list = g_list_prepend(arg->list, groups);
    for (l = c->subscribed; l != NULL; l = g_list_next(l)) {
	cg = g_hash_table_lookup(c->manager->cgroups, l->data);
	if (cg == NULL) {
	    continue;
	}
	item = message_arg_list("", NULL);
	item->list = g_list_prepend(item->list,
		message_arg_uint64("last_active", 0));
	item->list = g_list_prepend(item->list,
		status_message_arg("status", cgroup_get_status(cg)));
	item->list = g_list_prepend(item->list,
		message_arg_string("cgroup", l->data));
	groups->list = g_list_prepend(groups->list, item);
    }
    if (c->current != NULL &&
	    (cg = g_hash_table_lookup(c->manager->cgroups, c->current)) != NULL) {
	console_list = cgroup_list_consoles(cg);
	for (l = console_list; l != NULL; l = g_list_next(l)) {
	    console = g_hash_table_lookup(c->manager->consoles, l->data);
	    if (console == NULL) {
		g_free(l->data);
		continue;
	    }
	    item = message_arg_list("", NULL);
	    item->list = g_list_prepend(item->list,
		    message_arg_uint64("last_active",
			console_get_last_active(console)));
	    item->list = g_list_prepend(item->list,
		    status_message_arg("status",
			console_get_status(console)));
	    item->list = g_list_prepend(item->list,
		    message_arg_string("console", l->data));
	    g_free(l->data);
	    consoles->list = g_list_prepend(consoles->list, item);
	}
	g_list_free(console_list);
    }
    if (message_send(c->msg_q, CLIENT_MESSAGE_STATUS,
		arg, NULL, NULL, &e) == FALSE) {
	g_printerr("could not send message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(arg);
}

static void
client_detach_console(Client c,
	AttachedConsole *ac) {
    c->attached = g_list_remove(c->attached, ac);
    if (ac->io != NULL) {
	matrix_expunge(c->manager->matrix, ac->io);
    }
    if (c->mplex != NULL) {
	mplex_close(c->mplex, ac->chan, NULL);
    }
    if (ac->io != NULL) {
	g_io_channel_unref(ac->io);
    }
    if (ac->name != NULL) {
	g_free(ac->name);
    }
    g_free(ac);
}

static void
client_console_disconnects(Manager *manager,
	const EventData *event,
	gpointer data) {
    GList *l, *m;

    for (l = manager->clients; l != NULL; l = g_list_next(l)) {
	Client c = l->data;
	AttachedConsole *ac;

	for (m = c->attached; m != NULL; m = g_list_next(m)) {
	    ac = m->data;
	    if (strcmp(ac->name, event->console_disconnects.console) == 0) {
		break;
	    }
	}
	if (m != NULL) {
	    MessageArg *arg = message_arg_string("console", ac->name);
	    GError *e = NULL;

	    if (message_send(c->msg_q, CLIENT_MESSAGE_DETACH,
		    arg, NULL, NULL, &e) == FALSE) {
		g_printerr("could not send message (%s)", e->message);
		g_clear_error(&e);
	    }
	    message_arg_free(arg);
	    client_detach_console(c, ac);
	}
    }
}

static void
client_cgroup_changed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_list("", NULL);

    a->list = g_list_prepend(a->list,
	    message_arg_string("cgroup", event->cgroup_changed.group));
    a->list = g_list_prepend(a->list,
	    message_arg_string("console", event->cgroup_changed.console));
    a->list = g_list_prepend(a->list,
	    message_arg_uint8("status", event->cgroup_changed.added));
    /* FIXME: only send to clients that are subscribed to this group? */
    client_send_all(manager, CLIENT_MESSAGE_CGROUP_CHANGE, a);
    message_arg_free(a);
}

static void
client_cgroup_added(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("cgroup", event->cgroup_added.cgroup);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_CGROUP_ADDED, a);
    message_arg_free(a);
}

static void
client_cgroup_removed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("cgroup",
	    event->cgroup_removed.cgroup);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_CGROUP_REMOVED, a);
    message_arg_free(a);
}

static void
client_cgroup_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_list("", NULL);

    a->list = g_list_prepend(a->list,
	    message_arg_string("cgroup", event->cgroup_renamed.cgroup));
    a->list = g_list_prepend(a->list,
	    message_arg_string("newname", event->cgroup_renamed.newname));
    client_send_all(manager, CLIENT_MESSAGE_CONFIG_CGROUP_RENAMED, a);
    message_arg_free(a);
}

static void
client_console_added(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("console", event->console_added.console);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_CONSOLE_ADDED, a);
    message_arg_free(a);
}

static void
client_console_removed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("console",
	    event->console_removed.console);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_CONSOLE_REMOVED, a);
    message_arg_free(a);
}

static void
client_console_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_list("", NULL);

    a->list = g_list_prepend(a->list,
	    message_arg_string("console", event->console_renamed.console));
    a->list = g_list_prepend(a->list,
	    message_arg_string("newname", event->console_renamed.newname));
    client_send_all(manager, CLIENT_MESSAGE_CONFIG_CONSOLE_RENAMED, a);
    message_arg_free(a);
}

static void
client_log_added(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("log", event->log_added.log);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_LOG_ADDED, a);
    message_arg_free(a);
}

static void
client_log_removed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("log", event->log_removed.log);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_LOG_REMOVED, a);
    message_arg_free(a);
}

static void
client_log_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_list("", NULL);

    a->list = g_list_prepend(a->list,
	    message_arg_string("log", event->log_renamed.log));
    a->list = g_list_prepend(a->list,
	    message_arg_string("newname", event->log_renamed.newname));
    client_send_all(manager, CLIENT_MESSAGE_CONFIG_LOG_RENAMED, a);
    message_arg_free(a);
}

static void
client_user_added(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("user", event->user_added.user);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_USER_ADDED, a);
    message_arg_free(a);
}

static void
client_disconnect_removed_user(gpointer data,
	gpointer user_data) {
    Client c = data;
    gchar *user = user_data;

    if (c->user && strcmp(c->user, user) == 0) {
	client_close(c);
    }
}

static void
client_user_removed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("user", event->user_removed.user);

    g_list_foreach(manager->clients, client_disconnect_removed_user,
	    (gpointer) event->user_removed.user);
    client_send_all(manager, CLIENT_MESSAGE_CONFIG_USER_REMOVED, a);
    message_arg_free(a);
}

static void
client_user_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_list("", NULL);

    a->list = g_list_prepend(a->list,
	    message_arg_string("user", event->user_renamed.user));
    a->list = g_list_prepend(a->list,
	    message_arg_string("newname", event->user_renamed.newname));
    client_send_all(manager, CLIENT_MESSAGE_CONFIG_USER_RENAMED, a);
    message_arg_free(a);
}

static void
client_server_added(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("server", event->server_added.server);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_SERVER_ADDED, a);
    message_arg_free(a);
}

static void
client_server_removed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_string("server", event->server_removed.server);

    client_send_all(manager, CLIENT_MESSAGE_CONFIG_SERVER_REMOVED, a);
    message_arg_free(a);
}

static void
client_server_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    MessageArg *a = message_arg_list("", NULL);

    a->list = g_list_prepend(a->list,
	    message_arg_string("server", event->server_renamed.server));
    a->list = g_list_prepend(a->list,
	    message_arg_string("newname", event->server_renamed.newname));
    client_send_all(manager, CLIENT_MESSAGE_CONFIG_SERVER_RENAMED, a);
    message_arg_free(a);
}

static void
client_console_get_broken(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;
    Console console = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "not logged in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(2, "malformed request");
    }
    if (response == NULL) {
	console = g_hash_table_lookup(c->manager->consoles, arg->string->str);
	if (console == NULL) {
	    response = message_arg_error(0, "no such console");
	}
    }
    if (response == NULL) {
	const gchar *broken = console_get_broken(console);

	if (broken == NULL) {
	    response = message_arg_string("broken", "");
	} else {
	    response = message_arg_string("broken", broken);
	}
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_subscribe(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;
    CGroup cgroup = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "not logged in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(2, "malformed request");
    }
    if (response == NULL) {
	cgroup = g_hash_table_lookup(c->manager->cgroups, arg->string->str);
	if (cgroup == NULL) {
	    response = message_arg_error(0, "no such console group");
	}
    }
    if (response == NULL) {
	GList *consoles = cgroup_list_consoles(cgroup);
	GList *argl = NULL;
	GList *l;

	for (l = consoles; l != NULL; l = g_list_next(l)) {
	    argl = g_list_prepend(argl, message_arg_string("", l->data));
	    g_free(l->data);
	}
	g_list_free(consoles);
	response = message_arg_list("consoles", argl);
	for (l = c->subscribed; l != NULL; l = g_list_next(l)) {
	    if (strcmp(l->data, arg->string->str) == 0) {
		break;
	    }
	}
	if (l == NULL) {
	    c->subscribed = g_list_prepend(c->subscribed,
		    g_strdup(arg->string->str));
	}
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_unsubscribe(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;
    CGroup cgroup = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "not logged in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(2, "malformed request");
    }
    if (response == NULL) {
	cgroup = g_hash_table_lookup(c->manager->cgroups, arg->string->str);
	if (cgroup == NULL) {
	    response = message_arg_error(0, "no such console group");
	}
    }
    if (response == NULL) {
	GList *l;

	response = message_arg_none("");
	for (l = c->subscribed; l != NULL; l = g_list_next(l)) {
	    if (strcmp(l->data, arg->string->str) == 0) {
		break;
	    }
	}
	if (l != NULL) {
	    g_free(l->data);
	    c->subscribed = g_list_remove(c->subscribed, l->data);
	}
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_current(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;
    CGroup cgroup = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "not logged in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(2, "malformed request");
    }
    if (response == NULL) {
	cgroup = g_hash_table_lookup(c->manager->cgroups, arg->string->str);
	if (cgroup == NULL) {
	    response = message_arg_error(0, "no such console group");
	}
    }
    if (response == NULL) {
	GList *l;

	for (l = c->subscribed; l != NULL; l = g_list_next(l)) {
	    if (strcmp(l->data, arg->string->str) == 0) {
		break;
	    }
	}
	if (l == NULL) {
	    response = message_arg_error(1, "not subscribed");
	}
	if (c->current != NULL) {
	    g_free(c->current);
	}
	c->current = g_strdup(arg->string->str);
	client_send_status_fe(c);
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_detach(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "not logged in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(2, "malformed request");
    }
    if (response == NULL) {
	GList *l;
	AttachedConsole *ac;

	for (l = c->attached; l != NULL; l = g_list_next(l)) {
	    ac = l->data;

	    if (strcmp(ac->name, arg->string->str) == 0) {
		break;
	    }
	}
	if (l == NULL) {
	    response = message_arg_error(0, "no such console");
	} else {
	    client_detach_console(c, ac);
	    response = message_arg_none("");
	}
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_attach(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    Console console;
    GIOChannel *console_io;
    GIOChannel *client_io;
    MplexChannel chan;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(2, "malformed request");
    }
    if (response == NULL) {
	console = g_hash_table_lookup(c->manager->consoles, arg->string->str);
	if (console == NULL) {
	    response = message_arg_error(0, "no such console");
	}
    }
    if (response == NULL) {
	console_io = console_get_io_channel(console);
	if (console_io == NULL) {
	    response = message_arg_error(2, "console not available");
	}
    }
    if (response == NULL) {
	client_io = mplex_open(c->mplex, 0, &chan, NULL, NULL, &e);
	if (e == NULL) {
	    matrix_connect(c->manager->matrix, console_io, client_io, &e);
	}
	if (e == NULL) {
	    matrix_connect(c->manager->matrix, client_io, console_io, &e);
	}
	if (e != NULL) {
	    response = message_arg_error(2, "internal error (%s)", e->message);
	}
    }
    if (response == NULL) {
	AttachedConsole *ac = g_new0(AttachedConsole, 1);

	response = message_arg_uint16("channel", chan);
	ac->name = g_strdup(arg->string->str);
	ac->io = client_io;
	ac->chan = chan;
	c->attached = g_list_prepend(c->attached, ac);
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_log_opened(Log l,
	GIOChannel *data_channel,
	gpointer data) {
    LogSession *ls = data;
    MessageArg *response = NULL;
    GIOChannel *client_io;
    MplexChannel chan;
    GError *e = NULL;
    MessageQueue q = ls->mq;
    MessageID id = ls->mid;

    if (data_channel == NULL) {
	response = message_arg_error(2, "log not available");
    } else {
	client_io = mplex_open(ls->c->mplex, 0, &chan, NULL, NULL, &e);
	if (e == NULL) {
	    matrix_connect(ls->c->manager->matrix, data_channel, client_io, &e);
	}
	if (e != NULL) {
	    response = message_arg_error(2, "internal error (%s)", e->message);
	}
    }
    if (response == NULL) {
	response = message_arg_uint16("channel", chan);
    } else {
	ls->c->logsessions = g_list_remove(ls->c->logsessions, ls);
	log_query_free(ls->lq);
	g_free(ls);
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_console_logs(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    Console console;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(2, "malformed request");
    }
    if (response == NULL) {
	console = g_hash_table_lookup(c->manager->consoles, arg->string->str);
	if (console == NULL) {
	    response = message_arg_error(0, "no such console");
	}
    }
    if (response == NULL) {
	GList *logs = log_list_per_console(c->manager, arg->string->str);
	GList *argl = NULL;
	GList *l;

	for (l = logs; l != NULL; l = g_list_next(l)) {
	    argl = g_list_prepend(argl, message_arg_string("", l->data));
	    g_free(l->data);
	}
	g_list_free(logs);
	response = message_arg_list("logs", argl);
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_log(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    LogQuery *lq = NULL;
    MessageArg *response = NULL;
    Log log;
    LogSession *ls;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (lq = log_query_new_from_arg(arg)) == NULL) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL) {
	log = g_hash_table_lookup(c->manager->logs, lq->log);
	if (log == NULL) {
	    response = message_arg_error(0, "no such log");
	}
    }
    if (response == NULL) {
	ls = g_new0(struct _LogSession,1);
	ls->c = c;
	ls->l = log;
	ls->lq = lq;
	ls->mq = q;
	ls->mid = id;
	c->logsessions = g_list_prepend(c->logsessions, ls);
	log_query(ls->l, lq, client_log_opened, ls);
    } else {
	log_query_free(lq);
    }
    if (response != NULL && message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    if (response != NULL ) {
	message_arg_free(response);
    }
}

static void
client_login(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *user;
    MessageArg *password;
    MessageArg *response;
    GError *e = NULL;

    if (message_arg_list_explode(arg,
		"user", MESSAGE_ARG_STRING, &user,
		"password", MESSAGE_ARG_STRING, &password,
		NULL, 0, NULL) == FALSE) {
	response = message_arg_error(0, "malformed request");
    } else {
	if (user_verify(c->manager,
		    user->string->str, password->string->str) == TRUE) {
	    GList *l;
	    GList *lst;
	    GList *cgroups = cgroup_list(c->manager);
	    GList *consoles = console_list(c->manager);
	    GList *logs = log_list(c->manager);
	    GList *servers = server_list(c->manager);
	    GList *users = user_list(c->manager);

	    response = message_arg_list("", NULL);
	    /*cgroups*/
	    lst = NULL;
	    for (l = cgroups; l != NULL; l = g_list_next(l)) {
		lst = g_list_prepend(lst,
			message_arg_string("", l->data));
		g_free(l->data);
	    }
	    g_list_free(cgroups);
	    response->list = g_list_prepend(response->list,
		    message_arg_list("cgroups", lst));
	    /*consoles*/
	    lst = NULL;
	    for (l = consoles; l != NULL; l = g_list_next(l)) {
		lst = g_list_prepend(lst,
			message_arg_string("", l->data));
		g_free(l->data);
	    }
	    g_list_free(consoles);
	    response->list = g_list_prepend(response->list,
		    message_arg_list("consoles", lst));
	    /*logs*/
	    lst = NULL;
	    for (l = logs; l != NULL; l = g_list_next(l)) {
		lst = g_list_prepend(lst,
			message_arg_string("", l->data));
		g_free(l->data);
	    }
	    g_list_free(logs);
	    response->list = g_list_prepend(response->list,
		    message_arg_list("logs", lst));
	    /*servers*/
	    lst = NULL;
	    for (l = servers; l != NULL; l = g_list_next(l)) {
		lst = g_list_prepend(lst,
			message_arg_string("", l->data));
		g_free(l->data);
	    }
	    g_list_free(servers);
	    response->list = g_list_prepend(response->list,
		    message_arg_list("servers", lst));
	    /*users*/
	    lst = NULL;
	    for (l = users; l != NULL; l = g_list_next(l)) {
		lst = g_list_prepend(lst,
			message_arg_string("", l->data));
		g_free(l->data);
	    }
	    g_list_free(users);
	    response->list = g_list_prepend(response->list,
		    message_arg_list("users", lst));

	    c->user = g_strdup(user->string->str);
	} else {
	    response = message_arg_error(0, "bad user name or password");
	}
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_change_password(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *old;
    MessageArg *new;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && message_arg_list_explode(arg,
		"old", MESSAGE_ARG_STRING, &old,
		"new", MESSAGE_ARG_STRING, &new,
		NULL, 0, NULL) == FALSE) {
		response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    user_verify(c->manager, c->user, old->string->str) == FALSE) {
		response = message_arg_error(0, "incorrect password");
    }
    if (response == NULL &&
	    user_change_password(c->manager, c->user, new->string->str) == FALSE) {
		response = message_arg_error(1, "invalid password");
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_config_cgroup_spec(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecCgroup *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    (spec = cgroup_spec(c->manager, arg->string->str, &e)) == NULL) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = spec_cgroup_message_arg("", spec);
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_cgroup_free(spec);
    }
}

static void
client_config_add_cgroup(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecCgroup *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (spec = spec_cgroup_new_from_arg(arg)) == NULL) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL && cgroup_add(c->manager, spec, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_cgroup_free(spec);
    }
}

static void
client_config_edit_cgroup(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecCgroup *spec = NULL;
    MessageArg *orig_name;
    MessageArg *sp;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (message_arg_list_explode(arg,
		"name", MESSAGE_ARG_STRING, &orig_name,
		"spec", MESSAGE_ARG_LIST, &sp,
		NULL, 0, NULL) == FALSE ||
	    (spec = spec_cgroup_new_from_arg(sp)) == NULL)) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    cgroup_edit(c->manager, orig_name->string->str, spec, &e)
	    == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_cgroup_free(spec);
    }
}

static void
client_config_remove_cgroup(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    cgroup_remove(c->manager, arg->string->str, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_config_console_spec(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecConsole *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    (spec = console_spec(c->manager, arg->string->str, &e)) == NULL) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = spec_console_message_arg("", spec);
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_console_free(spec);
    }
}

static void
client_config_add_console(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecConsole *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (spec = spec_console_new_from_arg(arg)) == NULL) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL && console_add(c->manager, spec, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_console_free(spec);
    }
}

static void
client_config_edit_console(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecConsole *spec = NULL;
    MessageArg *orig_name;
    MessageArg *sp;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (message_arg_list_explode(arg,
		"name", MESSAGE_ARG_STRING, &orig_name,
		"spec", MESSAGE_ARG_LIST, &sp,
		NULL, 0, NULL) == FALSE ||
	    (spec = spec_console_new_from_arg(sp)) == NULL)) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    console_edit(c->manager, orig_name->string->str, spec, &e)
	    == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_console_free(spec);
    }
}

static void
client_config_remove_console(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    console_remove(c->manager, arg->string->str, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_config_log_spec(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecLog *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    (spec = log_spec(c->manager, arg->string->str, &e)) == NULL) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = spec_log_message_arg("", spec);
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_log_free(spec);
    }
}

static void
client_config_add_log(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecLog *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (spec = spec_log_new_from_arg(arg)) == NULL) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL && log_add(c->manager, spec, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_log_free(spec);
    }
}

static void
client_config_edit_log(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecLog *spec = NULL;
    MessageArg *orig_name;
    MessageArg *sp;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (message_arg_list_explode(arg,
		"name", MESSAGE_ARG_STRING, &orig_name,
		"spec", MESSAGE_ARG_LIST, &sp,
		NULL, 0, NULL) == FALSE ||
	    (spec = spec_log_new_from_arg(sp)) == NULL)) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    log_edit(c->manager, orig_name->string->str, spec, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_log_free(spec);
    }
}

static void
client_config_remove_log(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    log_remove(c->manager, arg->string->str, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_config_user_spec(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecUser *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    (spec = user_spec(c->manager, arg->string->str, &e)) == NULL) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = spec_user_message_arg("", spec);
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_user_free(spec);
    }
}

static void
client_config_add_user(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecUser *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (spec = spec_user_new_from_arg(arg)) == NULL) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL && user_add(c->manager, spec, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_user_free(spec);
    }
}

static void
client_config_edit_user(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecUser *spec = NULL;
    MessageArg *orig_name;
    MessageArg *sp;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (message_arg_list_explode(arg,
		"name", MESSAGE_ARG_STRING, &orig_name,
		"spec", MESSAGE_ARG_LIST, &sp,
		NULL, 0, NULL) == FALSE ||
	    (spec = spec_user_new_from_arg(sp)) == NULL)) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    user_edit(c->manager, orig_name->string->str, spec, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_user_free(spec);
    }
}

static void
client_config_remove_user(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    user_remove(c->manager, arg->string->str, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_config_server_spec(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecServer *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    (spec = server_spec(c->manager, arg->string->str, &e)) == NULL) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = spec_server_message_arg("", spec);
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_server_free(spec);
    }
}

static void
client_config_add_server(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecServer *spec = NULL;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (spec = spec_server_new_from_arg(arg)) == NULL) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL && server_add(c->manager, spec, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_server_free(spec);
    }
}

static void
client_config_edit_server(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    SpecServer *spec = NULL;
    MessageArg *orig_name;
    MessageArg *sp;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && (message_arg_list_explode(arg,
		"name", MESSAGE_ARG_STRING, &orig_name,
		"spec", MESSAGE_ARG_LIST, &sp,
		NULL, 0, NULL) == FALSE ||
	    (spec = spec_server_new_from_arg(sp)) == NULL)) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    server_edit(c->manager, orig_name->string->str, spec, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
    if (spec != NULL) {
	spec_server_free(spec);
    }
}

static void
client_config_remove_server(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Client c = data;
    MessageArg *response = NULL;
    GError *e = NULL;

    if (c->user == NULL) {
	response = message_arg_error(1, "must log in");
    }
    if (response == NULL && arg->type != MESSAGE_ARG_STRING) {
	response = message_arg_error(0, "malformed request");
    }
    if (response == NULL &&
	    server_remove(c->manager, arg->string->str, &e) == FALSE) {
	response = message_arg_error(0, "%s", e->message);
	g_clear_error(&e);
    }
    if (response == NULL) {
	response = message_arg_none("");
    }
    if (message_respond(q, id, response, &e) == FALSE) {
	g_printerr("could not respond to message (%s)", e->message);
	g_clear_error(&e);
    }
    message_arg_free(response);
}

static void
client_close(Client c) {
    if (c->user != NULL) {
	g_free(c->user);
    }
    while (c->attached) {
	AttachedConsole *ac = c->attached->data;

	client_detach_console(c, ac);
    }
    message_queue_free(c->msg_q);
    mplex_free(c->mplex);
    g_io_channel_unref(c->ssl);
    gnet_tcp_socket_delete(c->sock);
    c->manager->clients = g_list_remove(c->manager->clients, c);
    g_free(c);
}

static void
client_quits(Mplex mplex,
	gpointer data) {
    Client c = data;
    Mplex temp = c->mplex;

    g_printerr("client closed connection\n");
    if (c->user != NULL) {
	g_free(c->user);
    }
    c->mplex = NULL;
    while (c->attached) {
	AttachedConsole *ac = c->attached->data;

	client_detach_console(c, ac);
    }
    message_queue_free(c->msg_q);
    mplex_free(temp);
    g_io_channel_unref(c->ssl);
    gnet_tcp_socket_delete(c->sock);
    c->manager->clients = g_list_remove(c->manager->clients, c);
    g_free(c);
}

static void
accept_con(GTcpSocket *gserver,
	GTcpSocket *sock,
	gpointer user_data) {
    Manager *manager = user_data;
    Client c;
    GError *e = NULL;
    GIOChannel *io;

    if (sock == NULL) {
	/* FIXME: there was an error.
	 * manager->sock will need to be set up again.
	 */
	g_printerr("accept_con is broken\n");
	return;
    }
    c = g_new0(struct _Client, 1);
    c->user = NULL;
    c->attached = NULL;
    c->subscribed = NULL;
    c->current = NULL;
    c->manager = manager;
    c->sock = sock;
    c->ssl = ssl_server(gnet_tcp_socket_get_io_channel(sock));
    c->mplex = mplex_new(c->ssl, client_quits, c, &e);
    if (e != NULL) {
	g_printerr("cannot start mplex! (%s)\n", e->message);
	g_clear_error(&e);
	gnet_tcp_socket_delete(c->sock);
	g_free(c);
	return;
    }
    io = mplex_open(c->mplex, CONC_MSG_CHANNEL, NULL, NULL, NULL, &e);
    if (e != NULL) {
	g_printerr("cannot open message channel! (%s)\n", e->message);
	g_clear_error(&e);
	mplex_free(c->mplex);
	gnet_tcp_socket_delete(c->sock);
	g_free(c);
	return;
    }
    c->msg_q = message_queue_new(io, &e);
    if (e != NULL) {
	g_printerr("cannot open message queue! (%s)\n", e->message);
	g_clear_error(&e);
	mplex_free(c->mplex);
	gnet_tcp_socket_delete(c->sock);
	g_free(c);
	return;
    }
    g_printerr("client connected\n");
    /* add message_handlers */
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_LOGIN,
		client_login, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_ATTACH,
		client_attach, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_DETACH,
		client_detach, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONSOLE_LOGS,
		client_console_logs, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_LOG,
		client_log, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_SUBSCRIBE,
		client_subscribe, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_UNSUBSCRIBE,
		client_unsubscribe, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CURRSUB,
		client_current, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CHANGE_PASSWORD,
		client_change_password, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONSOLE_GET_BROKEN,
		client_console_get_broken, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }

    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_CGROUP_SPEC,
		client_config_cgroup_spec, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_CGROUP_ADD,
		client_config_add_cgroup, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_CGROUP_EDIT,
		client_config_edit_cgroup, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_CGROUP_REMOVE,
		client_config_remove_cgroup, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }

    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_CONSOLE_SPEC,
		client_config_console_spec, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_CONSOLE_ADD,
		client_config_add_console, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_CONSOLE_EDIT,
		client_config_edit_console, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_CONSOLE_REMOVE,
		client_config_remove_console, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }

    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_LOG_SPEC,
		client_config_log_spec, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_LOG_ADD,
		client_config_add_log, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_LOG_EDIT,
		client_config_edit_log, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_LOG_REMOVE,
		client_config_remove_log, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }

    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_USER_SPEC,
		client_config_user_spec, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_USER_ADD,
		client_config_add_user, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_USER_EDIT,
		client_config_edit_user, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_USER_REMOVE,
		client_config_remove_user, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }

    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_SERVER_SPEC,
		client_config_server_spec, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_SERVER_ADD,
		client_config_add_server, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_SERVER_EDIT,
		client_config_edit_server, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }
    if (message_handler_register(c->msg_q, CLIENT_MESSAGE_CONFIG_SERVER_REMOVE,
		client_config_remove_server, c, &e) == FALSE) {
	g_printerr("cannot register message handler! (%s)\n", e->message);
	g_clear_error(&e);
    }

    c->manager->clients = g_list_prepend(c->manager->clients, c);
    message_queue_start(c->msg_q);
    mplex_start(c->mplex);
}

extern void
client_load(Manager *manager) {
    manager->sock = gnet_tcp_socket_server_new_with_port(CONC_PORT_MGR);
    if (manager->sock == NULL) {
	g_printerr(_("error: cannot create server socket!\n"));
	return;
    }
    manager->clients = NULL;
    event_register(manager, EVENT_CONSOLE_DISCONNECTS,
	    client_console_disconnects, NULL);
    event_register(manager, EVENT_CGROUP_CHANGED,
	    client_cgroup_changed, NULL);
    event_register(manager, EVENT_CGROUP_ADDED,
	    client_cgroup_added, NULL);
    event_register(manager, EVENT_CGROUP_REMOVED,
	    client_cgroup_removed, NULL);
    event_register(manager, EVENT_CGROUP_RENAMED,
	    client_cgroup_renamed, NULL);
    event_register(manager, EVENT_CONSOLE_ADDED,
	    client_console_added, NULL);
    event_register(manager, EVENT_CONSOLE_REMOVED,
	    client_console_removed, NULL);
    event_register(manager, EVENT_CONSOLE_RENAMED,
	    client_console_renamed, NULL);
    event_register(manager, EVENT_LOG_ADDED,
	    client_log_added, NULL);
    event_register(manager, EVENT_LOG_REMOVED,
	    client_log_removed, NULL);
    event_register(manager, EVENT_LOG_RENAMED,
	    client_log_renamed, NULL);
    event_register(manager, EVENT_USER_ADDED,
	    client_user_added, NULL);
    event_register(manager, EVENT_USER_REMOVED,
	    client_user_removed, NULL);
    event_register(manager, EVENT_USER_RENAMED,
	    client_user_renamed, NULL);
    event_register(manager, EVENT_SERVER_ADDED,
	    client_server_added, NULL);
    event_register(manager, EVENT_SERVER_REMOVED,
	    client_server_removed, NULL);
    event_register(manager, EVENT_SERVER_RENAMED,
	    client_server_renamed, NULL);
    gnet_tcp_socket_server_accept_async(manager->sock, accept_con, manager);
}

extern void
client_send_status(Manager *manager) {
    GList *l;
    Client c;

    for (l = manager->clients; l != NULL; l = g_list_next(l)) {
	c = l->data;
	client_send_status_fe(c);
    }
}
