#include <string.h>
#include <stdarg.h>
#include <glib/gstdio.h>
#include "config-file.h"

typedef struct _Node Node;
struct _Node {
    ConfigDirective d;
    Node *parent;
    GList *children;	/*list of Node*/
};

struct _ConfigFile {
    gchar *file;
    Node *top;
    GIOChannel *io;
    guint writer;
    GString *wbuf;
};

struct _ConfigPath {
    ConfigFile cf;
    Node *n;
};

typedef struct _ParseState {
    ConfigFile cf;
    Node *n;
} ParseState;


static void
delete_node(Node *n) {
    while (n->children) {
	Node *nn = n->children->data;

	delete_node(nn);
    }
    g_list_free(n->children);
    if (n->d.name != NULL) {
	g_free(n->d.name);
    }
    if (n->d.value != NULL) {
	g_free(n->d.value);
    }
    if (n->parent != NULL) {
	n->parent->children = g_list_remove(n->parent->children, n);
    }
    g_free(n);
}

static Node *
find_child(Node *n, const gchar *name) {
    GList *l;
    Node *nn;

    for (l = n->children; l != NULL; l = g_list_next(l)) {
	nn = l->data;
	if (strcmp(nn->d.name, name) == 0) {
	    break;
	}
    }
    if (l == NULL) {
	return NULL;
    }
    return nn;
}

static void
write_node(GString **s,
	Node *n,
	gint indent) {
    gchar *is = g_strnfill(indent, ' ');
    if (n->d.name != NULL) {
	g_string_append_printf(*s, "%s<%s>", is, n->d.name);
	if (n->d.value != NULL) {
	    *s = g_string_append(*s, n->d.value);
	}
    } else {
	indent -= 4;
    }
    if (n->children != NULL) {
	GList *l;

	if (n->d.name != NULL) {
	    *s = g_string_append_c(*s, '\n');
	}
	for (l = n->children; l != NULL; l = g_list_next(l)) {
	    Node *cn = l->data;

	    write_node(s, cn, indent + 4);
	}
	if (n->d.name != NULL) {
	    *s = g_string_append(*s, is);
	}
    }
    if (n->d.name != NULL) {
	g_string_append_printf(*s, "</%s>\n", n->d.name);
    }
}

static void
parse_start_element(GMarkupParseContext *context,
	const gchar *element_name,
	const gchar **attribute_names,
	const gchar **attribute_values,
	gpointer user_data,
	GError **error) {
    ParseState *ps = user_data;
    Node *n = g_new0(Node, 1);

    n->d.name = g_strdup(element_name);
    n->parent = ps->n;
    ps->n->children = g_list_append(ps->n->children, n);
    ps->n = n;
}

static void
parse_end_element(GMarkupParseContext *context,
	const gchar *element_name,
	gpointer user_data,
	GError **error) {
    ParseState *ps = user_data;

    ps->n = ps->n->parent;
}

static void
parse_text(GMarkupParseContext *context,
	const gchar *text,
	gsize text_len,
	gpointer user_data,
	GError **error) {
    ParseState *ps = user_data;

    if (ps->n->d.value == NULL) {
	ps->n->d.value = g_strdup(text);
	g_strstrip(ps->n->d.value);
    }
}

static GMarkupParser parser_funcs = {
    parse_start_element,
    parse_end_element,
    parse_text,
    NULL,
    NULL
};


extern ConfigFile
config_file_open(const gchar *filename,
	GError **error) {
    GIOChannel *io;
    ConfigFile cf;
    GMarkupParseContext *pc;
    ParseState ps;
    gboolean quit_with_err = FALSE;
    gboolean quit = FALSE;
    struct stat s;

    g_assert(filename);
    cf = g_new0(struct _ConfigFile, 1);
    cf->file = g_strdup(filename);
    cf->top = g_new0(Node, 1);
    cf->top->parent = NULL;
    cf->top->children = NULL;
    cf->top->d.name = NULL;
    cf->top->d.value = NULL;
    if (g_file_test(filename, G_FILE_TEST_EXISTS) == FALSE ||
	(g_stat(filename, &s) == 0 && s.st_size == 0)) {
	return cf;
    }
    io = g_io_channel_new_file(filename, "r", error);
    if (io == NULL) {
	config_file_free(cf);
	return NULL;
    }
    ps.cf = cf;
    ps.n = cf->top;
    pc = g_markup_parse_context_new(&parser_funcs, 0, &ps, g_free);
    while (quit_with_err == FALSE && quit == FALSE) {
	gchar buf[1024];
	gsize bytes;

	switch (g_io_channel_read_chars(io, buf, 1024, &bytes, error)) {
	    case G_IO_STATUS_AGAIN:
		break;
	    case G_IO_STATUS_ERROR:
		quit_with_err = TRUE;
		break;
	    case G_IO_STATUS_EOF:
		if (g_markup_parse_context_end_parse(pc, error) == FALSE) {
		    quit_with_err = TRUE;
		} else {
		    quit = TRUE;
		}
		break;
	    case G_IO_STATUS_NORMAL:
		if (g_markup_parse_context_parse(pc, buf, bytes, error)
			== FALSE) {
		    quit_with_err = TRUE;
		}
	}
    }
    g_io_channel_unref(io);
    if (quit_with_err) {
	config_file_free(cf);
	return NULL;
    }
    return cf;
}

static void
config_file_writer_done(gpointer data) {
    ConfigFile cf = data;

    cf->writer = 0;
    g_io_channel_unref(cf->io);
    cf->io = NULL;
    g_string_free(cf->wbuf, TRUE);
    cf->wbuf = NULL;
}

static gboolean
config_file_writer(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    ConfigFile cf = data;
    gsize count = cf->wbuf->len;
    gsize bytes;
    GError *e = NULL;
    GString *s;

    if (count > 1024) {
	count = 1024;
    }
    switch (g_io_channel_write_chars(source, cf->wbuf->str, count,
		&bytes, &e)) {
	case G_IO_STATUS_ERROR:
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	default:
	    g_assert_not_reached();
    }
    s = g_string_new_len(&cf->wbuf->str[bytes], cf->wbuf->len - bytes);
    g_string_free(cf->wbuf, TRUE);
    cf->wbuf = s;
    if (cf->wbuf->len == 0) {
	return FALSE;
    } else {
	return TRUE;
    }
    return TRUE;
}

extern gboolean
config_file_write(ConfigFile cf,
	GError **error) {

    g_assert(cf);
    if (cf->writer > 0) {
	g_source_remove(cf->writer);
	cf->writer = 0;
    }
    if (cf->wbuf != NULL ) {
	cf->wbuf = g_string_assign(cf->wbuf, "");
    } else {
	cf->wbuf = g_string_new("");
    }
    if (cf->io != NULL) {
	g_io_channel_seek_position(cf->io, 0, G_SEEK_SET, NULL);
    } else {
	cf->io = g_io_channel_new_file(cf->file, "w", error);
	if (cf->io == NULL) {
	    return FALSE;
	}
    }
    write_node(&cf->wbuf, cf->top, 0);
    cf->writer = g_io_add_watch_full(cf->io, G_PRIORITY_HIGH_IDLE, G_IO_OUT,
	    config_file_writer, cf, config_file_writer_done);
    return TRUE;
}

extern void
config_file_free(ConfigFile cf) {
    g_assert(cf);
    delete_node(cf->top);
    g_free(cf->file);
    g_free(cf);
}

extern ConfigPath
config_path(ConfigFile cf,
	...) {
    va_list ap;
    Node *n;
    gchar *s;
    ConfigPath p;

    g_assert(cf);
    n = cf->top;
    va_start(ap, cf);
    for (s = va_arg(ap, gchar *); s != NULL; s = va_arg(ap, gchar *)) {
	n = find_child(n, s);
	if (n == NULL) {
	    return NULL;
	}
    }
    va_end(ap);
    p = g_new0(struct _ConfigPath, 1);
    p->cf = cf;
    p->n = n;
    return p;
}

extern ConfigPath
config_path_create(ConfigFile cf,
	ConfigPath parent,
	const gchar *name) {
    Node *n= g_new0(Node, 1);
    ConfigPath p;

    g_assert(cf);
    g_assert(parent);
    g_assert(name);
    n->parent = parent->n;
    n->d.name = g_strdup(name);
    parent->n->children = g_list_append(parent->n->children, n);
    p = g_new0(struct _ConfigPath, 1);
    p->cf = cf;
    p->n = n;
    return p;
}

extern void
config_path_up(ConfigPath p) {
    g_assert(p);
    if (p->n->parent == NULL) {
	return;
    }
    p->n = p->n->parent;
}

extern void
config_path_free(ConfigPath p) {
    g_free(p);
}

extern G_CONST_RETURN ConfigDirective *
config_get(ConfigPath p) {
    g_assert(p);
    return &p->n->d;
}

extern G_CONST_RETURN gchar *
config_get_value(ConfigPath p,
	const gchar *name) {
    Node *n;

    g_assert(p);
    g_assert(name);
    n = find_child(p->n, name);
    if (n == NULL) {
	return NULL;
    }
    return n->d.value;
}

extern GList *
config_children(ConfigPath p) {
    GList *l;
    GList *res = NULL;

    g_assert(p);
    for (l = p->n->children; l != NULL; l = g_list_next(l)) {
	ConfigPath np = g_new0(struct _ConfigPath, 1);
	np->cf = p->cf;
	np->n = l->data;
	res = g_list_append(res, np);
    }
    return res;
}

extern void
config_delete(ConfigPath p) {
    g_assert(p);
    delete_node(p->n);
    config_path_free(p);
}

extern gboolean
config_set(ConfigPath parent,
	const gchar *name,
	const gchar *value) {
    Node *n;

    g_assert(parent);
    g_assert(name);
    n = g_new0(Node, 1);
    n->parent = parent->n;
    n->children = NULL;
    n->d.name = g_strdup(name);
    if (value == NULL) {
	n->d.value = NULL;
    } else {
	n->d.value = g_strdup(value);
    }
    parent->n->children = g_list_append(parent->n->children, n);
    return TRUE;
}

extern gboolean
config_unset(ConfigPath parent,
	const gchar *name) {
    Node *n;

    g_assert(parent);
    g_assert(name);
    n = find_child(parent->n, name);
    if (n == NULL) {
	return FALSE;
    }
    delete_node(n);
    return TRUE;
}
