#ifndef __CONFIG_FILE_H
#define __CONFIG_FILE_H

#include <glib.h>

/*
 * ConfigFile - opaque structure representing a configuration file.
 *
 */
typedef struct _ConfigFile * ConfigFile;

/*
 * ConfigPath - pointer to an arbitrary poin int the config file
 *
 */
typedef struct _ConfigPath * ConfigPath;

/*
 * ConfigDirective - a name-value pair
 *
 */
typedef struct _ConfigDirective {
    gchar *name;
    gchar *value;
} ConfigDirective;

/*
 * config_file_new - open and parse a config file
 *
 * Returns the ConfigFile, or NULL and sets an error.
 *
 */
extern ConfigFile config_file_open(const gchar *filename,
	GError **error);

/*
 * config_file_write - rewrite the config file
 *
 * Returns TRUE if write successfully started, or FALSE and sets an error.
 *
 */
extern gboolean config_file_write(ConfigFile cf,
	GError **error);

/*
 * config_file_free - free a config file and all its resources
 *
 */
extern void config_file_free(ConfigFile cf);

/*
 * config_path - construct a config path
 *
 * Arguments are a null terminated list of const gchar *.
 * Returns the PATH, or NULL if it were not valid. The path then
 * belong to the caller and should be freed using config_path_free;
 */
extern ConfigPath config_path(ConfigFile cf,
	...);

/*
 * config_path_create - construct a new config path
 *
 * Returns the new path or NULL is the parent was invalid.
 * The path then belongs to the caller and shouls be freed using
 * config_path_free().
 *
 */
extern ConfigPath config_path_create(ConfigFile cf,
	ConfigPath parent,
	const gchar *name);

/*
 * config_path_up - modify a path to point to its parent
 *
 */
extern void config_path_up(ConfigPath p);

/*
 * config_path_free - destroy a path and free all its resources;
 *
 */
extern void config_path_free(ConfigPath p);

/*
 * config_get - get the directive at a given path
 *
 * Returns a pointer to the directive, or NULL if the path was not valid.
 *
 */
extern G_CONST_RETURN ConfigDirective *config_get(ConfigPath p);

/*
 * config_get_value - get the value of a directive.
 *
 * The directive is an immediate child of the path and has the given name.
 * Returns NULL if no such directive exists.
 *
 */
extern G_CONST_RETURN gchar *config_get_value(ConfigPath p,
	const gchar *name);

/*
 * config_children - return a path's immediate children
 *
 * Returns a GList of config paths, or NULL if the path was not valid, or had
 * no children. The paths and list belong to the caller.
 *
 */
extern GList *config_children(ConfigPath p);

/*
 * config_delete - remove a path and all its children.
 *
 * Also frees the path itself (since it is no longer valid).
 *
 */
extern void config_delete(ConfigPath p);

/*
 * config_set - create and set a directive
 *
 * Create a directive with the given name and value as a parent of the
 * given path.
 * Returns TRUE if successful, or FALSE if the parent path was not valid.
 *
 */
extern gboolean config_set(ConfigPath parent,
	const gchar *name,
	const gchar *value);

/*
 * config_unset - remove an existing directive
 *
 * Removes the first directive that is a child of parent with the given name.
 *
 * Returns TRUE if successful, or FALSE if the parent path was not valid
 * or no child with the given name was found.
 *
 */
extern gboolean config_unset(ConfigPath parent,
	const gchar *name);



#endif
