#include <string.h>
#include "console.h"
#include "cgroup.h"
#include "client.h"
#include "config-file.h"
#include "terminal.h"
#include "ipmi.h"
#include "message.h"
#include "server.h"
#include "svr-message.h"
#include "mplex.h"

struct _Console {
    gchar *name;
    Manager *manager;
    ConfigPath config_path;
    gchar *server;
    ConsoleType type;
    union {
	Terminal *t;
	IPMI *i;
    };
    GIOChannel *io;
    Status status;
    gchar *broken;
    guint64 last_active;
    gboolean active;
};

typedef struct _ListPerServer {
    const gchar *server;
    GList *l;
} ListPerServer;

extern G_CONST_RETURN gchar *
console_get_broken(Console c) {
    g_assert(c);
    return c->broken;
}

static void
set_broken(Console c, gchar *text) {
    if (c->broken != NULL) g_free(c->broken);
    if (text == NULL) {
	c->broken = NULL;
    } else {
	c->broken = g_strdup(text);
    }
}


static void
check_activity_fe(gpointer key,
        gpointer value,
        gpointer data) {
    Console c = value;
    GIOChannel *hitfile = data;
    gchar *s;

    if (c->active) {
	c->active = FALSE;
	if (c->status == STATUS_OK) {
	    c->status = STATUS_ACTIVE;
	}
	c->last_active = c->manager->time.tv_sec;
    } else {
	if (c->status == STATUS_ACTIVE) {
	    c->status = STATUS_OK;
	}
    }
    if (hitfile != NULL) {
	s = g_strdup_printf("%s %lld\n", c->name, c->last_active);
	g_io_channel_write_chars(hitfile, s, -1, NULL, NULL);
	g_free(s);
    }
}

static gboolean
check_activity(gpointer data) {
    Manager *manager = data;
    GIOChannel *hitfile;

    hitfile = g_io_channel_new_file(CONC_LAST_ACTIVE, "w", NULL);
    g_get_current_time(&manager->time);
    g_hash_table_foreach(manager->consoles, check_activity_fe, hitfile);
    if (hitfile != NULL ) {
	g_io_channel_unref(hitfile);
    }
    cgroup_check_status(manager);
    client_send_status(manager);
    return TRUE;
}


static void
console_close(Console c) {
    g_assert(c);

    if (c->io == NULL) {
	return;
    }
    event_raise(c->manager, EVENT_CONSOLE_DISCONNECTS, c->name);
    matrix_expunge(c->manager->matrix, c->io);
    g_io_channel_shutdown(c->io, TRUE, NULL);
    g_io_channel_unref(c->io);
    c->io = NULL;
    c->status = STATUS_BROKEN;
    set_broken(c, "closed");
}

static void
console_closed(Mplex mplex, MplexChannel chan, gpointer data) {
    console_close((Console)data);
    /*reconnect?*/
}

static void
console_hash_free(gpointer data) {
    Console console = data;

    /* FIXME: do anything else? */
    console_close(console);
    if (console->name != NULL) {
	g_free(console->name);
    }
    if (console->server != NULL) {
	g_free(console->server);
    }
    if (console->config_path) {
	config_path_free(console->config_path);
    }
    if (console->type == CONSOLE_TYPE_SERIAL && console->t != NULL) {
	terminal_free(console->t);
    }
    if (console->type == CONSOLE_TYPE_IPMI && console->i != NULL) {
	ipmi_free(console->i);
    }
    set_broken(console, NULL);
    g_free(console);
}

static void
console_opened(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    Console c = data;
    GError *e;
    Mplex mplex;
    MplexChannel chan;

    e = message_arg_error_check(arg);
    if (e != NULL) {
	g_printerr("could not open console %s (%s)\n", c->name,
		e->message);
	set_broken(c, e->message);
	g_clear_error(&e);
	return;
    }
    g_assert(arg->type == MESSAGE_ARG_UINT16);
    mplex = server_mplex(c->manager, c->server);
    if (mplex == NULL) {
	/* FIXME: report error? */
	return;
    }
    c->io = mplex_open(mplex, arg->uint16, &chan, console_closed, c, &e);
    if (c->io == NULL) {
	g_printerr(_("failed to open mplex channel for console %s (%s)"),
		c->name, e->message);
	g_clear_error(&e);
	return;
    }
    if (mplex_notify(mplex, chan, &c->active, &e) == FALSE) {
	g_printerr(_("failed to set notify flag for console %s (%s)"),
		c->name, e->message);
	g_clear_error(&e);
	return;
    }
    c->status = STATUS_OK;
    set_broken(c, NULL);
    event_raise(c->manager, EVENT_CONSOLE_CONNECTS, c->name);
}

static gboolean
console_connect(Console c, GError **error) {
    MessageQueue mq;
    MessageArg *a = NULL;
    const gchar *command = NULL;
    gboolean ret;

    mq = server_message_queue(c->manager, c->server);
    if (mq == NULL) return FALSE;
    switch (c->type) {
	case CONSOLE_TYPE_SERIAL:
	    a = terminal_message_arg("terminal", c->t);
	    command = SERVER_MESSAGE_OPEN_CONSOLE;
	    break;
	case CONSOLE_TYPE_IPMI:
	    a = ipmi_message_arg("ipmi", c->i);
	    command = SERVER_MESSAGE_OPEN_IPMI;
	    break;
	default:
	    g_assert_not_reached();
    }
    ret = message_send(mq, command, a, console_opened, c, error);
    message_arg_free(a);
    return ret;
}

static void
server_disconnected_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Console c = value;
    const gchar *server = data;

    if (strcmp(server, c->server) == 0 &&
	    c->io != NULL) {
	event_raise(c->manager, EVENT_CONSOLE_DISCONNECTS, c->name);
	matrix_expunge(c->manager->matrix, c->io);
	g_io_channel_shutdown(c->io, TRUE, NULL);
	g_io_channel_unref(c->io);
	c->io = NULL;
	c->status = STATUS_BROKEN;
	set_broken(c, "server disconnected");
    }
}

static void
server_disconnected(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_hash_table_foreach(manager->consoles, server_disconnected_fe,
	    (gpointer)event->server_disconnects.server);
}

static void
server_connected_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Console c = value;
    const gchar *server = data;
    GError *e = NULL;

    if (strcmp(server, c->server) == 0) {
	if (console_connect(c,  &e) == FALSE) {
	    g_printerr("error sending message (%s)\n", e->message);
	    g_clear_error(&e);
	}
    }
}

static void
server_connected(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_hash_table_foreach(manager->consoles, server_connected_fe,
	    (gpointer)event->server_connects.server);
}

static void
server_renamed_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Console c = value;
    EventData *event = data;

    if (strcmp(event->server_renamed.server, c->server) == 0) {
	g_free(c->server);
	c->server = g_strdup(event->server_renamed.newname);
	config_unset(c->config_path, "server");
	config_set(c->config_path, "server", c->server);
    }
}


static void
server_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_hash_table_foreach(manager->consoles, server_renamed_fe,
	    (gpointer)event);
    config_file_write(manager->config, NULL);
}

extern void
console_load(Manager *manager) {
    ConfigPath cp;
    GList *consoles, *l;
    const gchar *host;
    const gchar *user;
    const gchar *passwd;
    const gchar *device;
    const gchar *parity;
    gint ispeed, ospeed, data_bits, stop_bits;
    GIOChannel *hitfile;
    GHashTable *last_active;
    guint64 *la;

    g_assert(manager);
    g_assert(manager->config);

    last_active = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
    hitfile = g_io_channel_new_file(CONC_LAST_ACTIVE, "r", NULL);
    if (hitfile != NULL) {
	char *line;
	gsize i, len;

	while (g_io_channel_read_line(hitfile, &line, &len, NULL, NULL) ==
		G_IO_STATUS_NORMAL) {
	    for (i = 0; i < len && line[i] != ' '; i++);
	    if (i == len)  {
		g_free(line);
		continue;
	    }
	    line[i++] = '\0';
	    la = g_new(guint64, 1);
	    *la = g_ascii_strtoull(&line[i], NULL, 10);
	    g_hash_table_insert(last_active, line, la);
	}
    }

    manager->consoles = g_hash_table_new_full(g_str_hash, g_str_equal,
	    g_free, console_hash_free);

    cp = config_path(manager->config, "consoles", NULL);
    if (cp == NULL) {
	consoles = NULL;
    } else {
	consoles = config_children(cp);
	config_path_free(cp);
    }
    for (l = consoles; l != NULL; l = g_list_next(l)) {
	Console c = g_new0(struct _Console, 1);
	const gchar *str;
	
	cp = l->data;
	str = config_get_value(cp, "name");
	if (str == NULL) {
	    g_printerr("configuration file corrupt (anonymous console)\n");
	    g_free(c);
	    continue;
	}
	c->name = g_strdup(str);
	c->manager = manager;
	c->config_path = cp;
	c->status = STATUS_BROKEN;
	c->active = FALSE;
	c->broken = NULL;
	set_broken(c, "not connected");

	la = g_hash_table_lookup(last_active, c->name);
	if (la != NULL) {
	    c->last_active = *la;
	} else {
	    c->last_active = 0;
	}

	c->type = CONSOLE_TYPE_SERIAL;
	str = config_get_value(cp, "type");
	if (str != NULL && strcmp(str, "ipmi") == 0) {
	    c->type = CONSOLE_TYPE_IPMI;
	}

	str = config_get_value(cp, "server");
	if (str == NULL) {
	    g_printerr("configuration file corrupt (console %s"
		    " has no server)\n", c->name);
	    g_free(c->name);
	    g_free(c);
	    continue;
	}
	c->server = g_strdup(str);

	if (c->type == CONSOLE_TYPE_SERIAL) {
	    device = config_get_value(cp, "device");
	    if (device == NULL) {
		g_printerr("configuration file corrupt (console %s"
			" has no device)\n", c->name);
		g_free(c->name);
		g_free(c->server);
		g_free(c);
		continue;
	    }

	    str = config_get_value(cp, "ispeed");
	    if (str == NULL) {
		ispeed = 9600;
	    } else {
		ispeed = (gint) g_strtod(str, NULL);
	    }

	    str = config_get_value(cp, "ospeed");
	    if (str == NULL) {
		ospeed = 9600;
	    } else {
		ospeed = (gint) g_strtod(str, NULL);
	    }

	    str = config_get_value(cp, "data_bits");
	    if (str == NULL) {
		data_bits = 8;
	    } else {
		data_bits = (gint) g_strtod(str, NULL);
	    }

	    parity = config_get_value(cp, "parity");
	    if (parity == NULL) {
		parity = "none";
	    }

	    str = config_get_value(cp, "stop_bits");
	    if (str == NULL) {
		stop_bits = 1;
	    } else {
		stop_bits = (gint) g_strtod(str, NULL);
	    }

	    c->t = terminal_new(device, ispeed, ospeed, data_bits,
		    terminal_parity(parity), stop_bits);
	} else if (c->type == CONSOLE_TYPE_IPMI) {
	    host = config_get_value(cp, "host");
	    if (host == NULL) {
		g_printerr("configuration file corrupt (console %s"
			" has no host)\n", c->name);
		g_free(c->name);
		g_free(c->server);
		g_free(c);
		continue;
	    }

	    user = config_get_value(cp, "user");
	    if (user == NULL) {
		g_printerr("configuration file corrupt (console %s"
			" has no user)\n", c->name);
		g_free(c->name);
		g_free(c->server);
		g_free(c);
		continue;
	    }

	    passwd = config_get_value(cp, "passwd");
	    if (passwd == NULL) {
		g_printerr("configuration file corrupt (console %s"
			" has no password)\n", c->name);
		g_free(c->name);
		g_free(c->server);
		g_free(c);
		continue;
	    }

	    c->i = ipmi_new(host, user, passwd);
	} else {
	    g_assert_not_reached();
	}

	c->io = NULL;

	g_hash_table_insert(manager->consoles, g_strdup(c->name), c);
    }

    g_hash_table_unref(last_active);

    event_register(manager, EVENT_SERVER_CONNECTS,
	    server_connected, NULL);
    event_register(manager, EVENT_SERVER_DISCONNECTS,
	    server_disconnected, NULL);
    event_register(manager, EVENT_SERVER_RENAMED,
	    server_renamed, NULL);
    g_timeout_add(5000, check_activity, manager);
}

extern GIOChannel *
console_get_io_channel(Console c) {
    return c->io;
}

extern Status
console_get_status(Console c) {
    return c->status;
}

extern guint64
console_get_last_active(Console c) {
    return c->last_active;
}

extern SpecConsole *
console_spec(Manager *manager,
	const gchar *console,
	GError **error) {
    Console c;
    SpecConsole *spec;

    g_assert(manager);
    g_assert(console);

    c = g_hash_table_lookup(manager->consoles, console);
    if (c == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console %s does not exists"), console);
	return NULL;
    }
    spec = g_new0(struct _SpecConsole, 1);
    spec->name = g_strdup(c->name);
    spec->server = g_strdup(c->server);
    spec->type = c->type;
    switch (c->type) {
	case CONSOLE_TYPE_SERIAL:
	    spec->terminal = terminal_new(c->t->device,
		    c->t->ispeed,
		    c->t->ospeed,
		    c->t->data_bits,
		    c->t->parity,
		    c->t->stop_bits);
	    break;
	case CONSOLE_TYPE_IPMI:
	    spec->ipmi = ipmi_new(c->i->host, c->i->user, c->i->passwd);
	    break;
	default:
	    g_assert_not_reached();
    }
    return spec;
}

static void
console_list_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Console c = value;
    GList **l = data;

    *l = g_list_prepend(*l, g_strdup(c->name));
}

extern GList *
console_list(Manager *manager) {
    GList *l = NULL;

    g_hash_table_foreach(manager->consoles, console_list_fe, &l);
    return l;
}

static void
console_list_per_server_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Console c = value;
    ListPerServer *lps = data;

    if (strcmp(lps->server, c->server) == 0) {
	lps->l = g_list_prepend(lps->l, g_strdup(c->name));
    }
}

extern GList *
console_list_per_server(Manager *manager,
	const gchar *server) {
    ListPerServer lps;

    lps.server = server;
    lps.l = NULL;
    g_hash_table_foreach(manager->consoles, console_list_per_server_fe, &lps);
    return lps.l;
}

extern gboolean
console_add(Manager *manager,
	const SpecConsole *spec,
	GError **error) {
    Console c;
    ConfigPath consoles;
    MessageQueue smq;
    gchar *s;

    g_assert(manager);
    g_assert(spec);

    c = g_hash_table_lookup(manager->consoles, spec->name);
    if (c != NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console %s already exists"), c->name);
	return FALSE;
    }
    smq = server_message_queue(manager, spec->server);
    if (smq == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_GENERAL,
		_("server %s not available"), spec->server);
	return FALSE;
    }

    c = g_new0(struct _Console, 1);
    c->name = g_strdup(spec->name);
    c->manager = manager;
    c->config_path = NULL;
    c->server = g_strdup(spec->server);
    c->type = spec->type;
    switch (c->type) {
	case CONSOLE_TYPE_SERIAL:
	    c->t = terminal_new(spec->terminal->device,
		    spec->terminal->ispeed,
		    spec->terminal->ospeed,
		    spec->terminal->data_bits,
		    spec->terminal->parity,
		    spec->terminal->stop_bits);
	    break;
	case CONSOLE_TYPE_IPMI:
	    c->i = ipmi_new(spec->ipmi->host,
		    spec->ipmi->user,
		    spec->ipmi->passwd);
	    break;
	default:
	    g_assert_not_reached();
    }
    c->io = NULL;
    c->status = STATUS_BROKEN;
    c->broken = NULL;
    set_broken(c, "not connected");
    g_hash_table_insert(manager->consoles, g_strdup(c->name), c);
    consoles = config_path(manager->config, "consoles", NULL);
    if (consoles == NULL) {
	ConfigPath root = config_path(manager->config, NULL);

	consoles = config_path_create(manager->config, root, "consoles");
	config_path_free(root);
    }
    c->config_path = config_path_create(manager->config, consoles, "console");
    config_path_free(consoles);

    config_set(c->config_path, "name", c->name);
    config_set(c->config_path, "server", c->server);
    switch (c->type) {
	case CONSOLE_TYPE_SERIAL:
	    config_set(c->config_path, "type", "serial");
	    config_set(c->config_path, "device", c->t->device);
	    s = g_strdup_printf("%d", c->t->ispeed);
	    config_set(c->config_path, "ispeed", s);
	    g_free(s);
	    s = g_strdup_printf("%d", c->t->ospeed);
	    config_set(c->config_path, "ospeed", s);
	    g_free(s);
	    s = g_strdup_printf("%d", c->t->data_bits);
	    config_set(c->config_path, "data_bits", s);
	    g_free(s);
	    config_set(c->config_path, "parity", terminal_parity_string(c->t->parity));
	    s = g_strdup_printf("%d", c->t->stop_bits);
	    config_set(c->config_path, "stop_bits", s);
	    g_free(s);
	    break;
	case CONSOLE_TYPE_IPMI:
	    config_set(c->config_path, "type", "ipmi");
	    config_set(c->config_path, "host", c->i->host);
	    config_set(c->config_path, "user", c->i->user);
	    config_set(c->config_path, "passwd", c->i->passwd);
	    break;
	default:
	    g_assert_not_reached();
    }
    config_file_write(manager->config, NULL);

    event_raise(manager, EVENT_CONSOLE_ADDED, c->name);

    return console_connect(c, NULL);
}

extern gboolean
console_edit(Manager *manager,
	const gchar *console,
	const SpecConsole *spec,
	GError **error) {
    Console c;
    MessageQueue smq;
    gchar *s;
    gchar *old_name;
    gpointer on_p, c_p;
    gboolean changed = FALSE;
    gboolean name_changed = FALSE;
    gboolean reconnect = FALSE;
    GError *e = NULL;

    g_assert(manager);
    g_assert(console);
    g_assert(spec);

    if (strcmp(console, spec->name)) {
	c = g_hash_table_lookup(manager->consoles, spec->name);
	if (c != NULL) {
	    g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		    _("console %s already exists"), c->name);
	    return FALSE;
	}
	changed = TRUE;
	name_changed = TRUE;
    }
    smq = server_message_queue(manager, spec->server);
    if (smq == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_GENERAL,
		_("server %s not available"), spec->server);
	return FALSE;
    }
    if (g_hash_table_lookup_extended(manager->consoles, console,
		&on_p, &c_p) == FALSE) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console %s does not exist"), console);
	return FALSE;
    }
    old_name = on_p;
    c = c_p;

    if (name_changed) {
	g_free(c->name);
	c->name = g_strdup(spec->name);
	config_unset(c->config_path, "name");
        config_set(c->config_path, "name", c->name);
    }
    if (strcmp(c->server, spec->server) ||
	    c->type != spec->type ||
	    (c->type == CONSOLE_TYPE_SERIAL &&
	    !terminal_match(c->t, spec->terminal)) ||
	    (c->type == CONSOLE_TYPE_IPMI &&
	     !ipmi_match(c->i, spec->ipmi))) {
	changed = TRUE;
	console_close(c);
	reconnect = TRUE;
	config_unset(c->config_path, "server");
	config_unset(c->config_path, "type");
	switch (c->type) {
	    case CONSOLE_TYPE_SERIAL:
		config_unset(c->config_path, "device");
		config_unset(c->config_path, "ispeed");
		config_unset(c->config_path, "ospeed");
		config_unset(c->config_path, "data_bits");
		config_unset(c->config_path, "parity");
		config_unset(c->config_path, "stop_bits");
		terminal_free(c->t);
		break;
	    case CONSOLE_TYPE_IPMI:
		config_unset(c->config_path, "host");
		config_unset(c->config_path, "user");
		config_unset(c->config_path, "passwd");
		ipmi_free(c->i);
		break;
	    default:
		g_assert_not_reached();
	}
	if (c->server) {
	    g_free(c->server);
	}
	c->server = g_strdup(spec->server);
	config_set(c->config_path, "server", c->server);
	c->type = spec->type;
	switch (c->type) {
	    case CONSOLE_TYPE_SERIAL:
		config_set(c->config_path, "type", "serial");
		c->t = terminal_new(spec->terminal->device,
			spec->terminal->ispeed,
			spec->terminal->ospeed,
			spec->terminal->data_bits,
			spec->terminal->parity,
			spec->terminal->stop_bits);
		config_set(c->config_path, "device", c->t->device);
		s = g_strdup_printf("%d", c->t->ispeed);
		config_set(c->config_path, "ispeed", s);
		g_free(s);
		s = g_strdup_printf("%d", c->t->ospeed);
		config_set(c->config_path, "ospeed", s);
		g_free(s);
		s = g_strdup_printf("%d", c->t->data_bits);
		config_set(c->config_path, "data_bits", s);
		g_free(s);
		config_set(c->config_path, "parity",
			terminal_parity_string(c->t->parity));
		s = g_strdup_printf("%d", c->t->stop_bits);
		config_set(c->config_path, "stop_bits", s);
		g_free(s);
		break;
	    case CONSOLE_TYPE_IPMI:
		config_set(c->config_path, "type", "ipmi");
		c->i = ipmi_new(spec->ipmi->host,
			spec->ipmi->user,
			spec->ipmi->passwd);
		config_set(c->config_path, "host", c->i->host);
		config_set(c->config_path, "user", c->i->user);
		config_set(c->config_path, "passwd", c->i->passwd);
		break;
	    default:
		g_assert_not_reached();
	}
    }
    if (changed) {
	config_file_write(manager->config, NULL);
    }
    if (name_changed) {
	g_hash_table_steal(manager->consoles, old_name);
	g_free(old_name);
	g_hash_table_insert(manager->consoles, g_strdup(c->name), c);
	event_raise(manager, EVENT_CONSOLE_RENAMED, console, c->name);
    }
    if (reconnect) {
	if (console_connect(c, &e) == FALSE) {
	    g_printerr("error sending message (%s)\n", e->message);
	    g_clear_error(&e);
	}
    }
    return TRUE;
}


extern gboolean
console_remove(Manager *manager,
	const gchar *console,
	GError **error) {
    Console c;

    g_assert(manager);
    g_assert(console);

    c = g_hash_table_lookup(manager->consoles, console);
    if (c == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console %s does not exist"), console);
	return FALSE;
    }
    config_delete(c->config_path);
    c->config_path = NULL;
    g_hash_table_remove(manager->consoles, console);
    config_file_write(manager->config, NULL);
    event_raise(manager, EVENT_CONSOLE_REMOVED, console);
    return TRUE;
}
