#include <string.h>
#include "log.h"
#include "config-file.h"
#include "message.h"
#include "server.h"
#include "svr-message.h"
#include "pipe.h"
#include "console.h"
#include "date.h"

struct _Log {
    gchar *name;
    Manager *manager;
    ConfigPath config_path;
    gchar *server;
    gchar *filename;
    GHashTable *consoles; /* hash of LogReader * */
    GIOChannel *io;
};

typedef enum {
    STATE_NORMAL,
    STATE_ESCAPE,
    STATE_IN_CODE
} LogReaderState;

typedef struct _LogReader {
    gchar *console;
    GIOChannel *io;
    GString *buf;
    guint reader;
    LogReaderState state;
    Log log;
} LogReader;

typedef struct _LogQueryData {
    Log l;
    LogQueryCB cb;
    gpointer data;
} LogQueryData;

typedef struct _ListPer {
    const gchar *target;
    GList *l;
} ListPer;

static void
log_query_opened(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    LogQueryData *lqd = data;
    GError *e;
    Mplex mplex;
    GIOChannel *c;
    
    e = message_arg_error_check(arg);
    if (e != NULL) {
	g_printerr("could not open log query %s (%s)\n", lqd->l->name,
		e->message);
	g_clear_error(&e);
	lqd->cb(lqd->l, NULL, lqd->data);
	g_free(lqd);
	return;
    }
    mplex = server_mplex(lqd->l->manager, lqd->l->server);
    if (mplex == NULL) {
	/* FIXME: report error? */
	return;
    }
    c = mplex_open(mplex, arg->uint16, NULL, NULL, NULL, &e);
    if (c == NULL) {
	g_printerr(_("failed to open mplex channel for log query %s (%s)"),
		lqd->l->name, e->message);
	g_clear_error(&e);
    }
    lqd->cb(lqd->l, c, lqd->data);
    g_free(lqd);
}

extern void
log_query(Log l,
	const LogQuery *lq,
	LogQueryCB cb,
	gpointer data) {
    MessageQueue mq;
    GError *e = NULL;

    mq = server_message_queue(l->manager, l->server);
    if (mq != NULL) {
	MessageArg *a;
	LogQueryData *lqd;
	LogQuery *llq = log_query_new(l->filename, lq->start_time,
		lq->end_time, lq->max_lines, lq-> console);

	a = log_query_message_arg("query", llq);
	log_query_free(llq);
	lqd = g_new0(struct _LogQueryData, 1);
	lqd->l = l;
	lqd->cb = cb;
	lqd->data = data;
	if (message_send(mq, SERVER_MESSAGE_QUERY_LOG,
		    a, log_query_opened, lqd, &e) == FALSE) {
	    g_printerr("error sending message (%s)\n", e->message);
	    g_clear_error(&e);
	    g_free(lqd);
	    cb(l, NULL, data);
	}
    } else {
	cb(l, NULL, data);
    }
}

static void
log_close(Log l) {
    g_assert(l);

    if (l->io == NULL) {
	return;
    }
    event_raise(l->manager, EVENT_LOG_DISCONNECTS, l->name);
    matrix_expunge(l->manager->matrix, l->io);
    g_io_channel_shutdown(l->io, TRUE, NULL);
    g_io_channel_unref(l->io);
    l->io = NULL;
}

static void
log_hash_free(gpointer data) {
    Log log = data;

    log_close(log);
    if (log->name) {
	g_free(log->name);
    }
    if (log->server) {
	g_free(log->server);
    }
    g_hash_table_unref(log->consoles);
    g_free(log);
}

static void
log_reader_done(gpointer data) {
    /* FIXME: do something! */
}

static gboolean
log_process(LogReader *lr) {
    gint i;
    GString *line;
    gchar *timestr;

    for (i = 0; i < lr->buf->len; i++) {
	if (lr->buf->str[i] == '\n') {
	    break;
	}
    }
    if (i == lr->buf->len) {
	return FALSE;
    }
    line = g_string_new("");
    timestr = date_print(date_now());
    g_string_printf(line, "%s %s: ", timestr, lr->console);
    g_free(timestr);
    if (i > 0 && lr->buf->str[i - 1] == '\r') {
	line = g_string_append_len(line, lr->buf->str, i - 1);
    } else {
	line = g_string_append_len(line, lr->buf->str, i);
    }
    line = g_string_append_c(line, '\n');
    lr->buf = g_string_erase(lr->buf, 0, i + 1);
    if (lr->log->io != NULL) {
	/* FIXME: we should check that the line wrote ok */
	g_io_channel_write_chars(lr->log->io, line->str, line->len, NULL, NULL);
    }
    g_string_free(line, TRUE);
    return TRUE;
}

static gboolean
log_reader(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    LogReader *lr = data;
    gchar buf[256];
    GError *e = NULL;
    gsize bytes;
    gint i;

    g_assert(lr);

    switch (g_io_channel_read_chars(source, buf, sizeof(buf), &bytes, &e)) {
	case G_IO_STATUS_ERROR:
	    g_printerr(_("failed to read console %s for log %s (%s)\n"),
		    lr->console, lr->log->name, e->message);
	    g_clear_error(&e);
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	default:
	    g_assert_not_reached();
    }
    for (i = 0; i < bytes; i++) {
	switch (lr->state) {
	    case STATE_NORMAL:
		if (buf[i] == 033) {
		    lr->state = STATE_ESCAPE;
		    break;
		}
		if (buf[i] < 32 &&
			buf[i] != '\r' &&
			buf[i] != '\n' &&
			buf[i] != '\t') {
		    break;
		}
		lr->buf = g_string_append_c(lr->buf, buf[i]);
		break;
	    case STATE_ESCAPE:
		if (buf[i] == '[') {
		    lr->state = STATE_IN_CODE;
		} else {
		    lr->state = STATE_NORMAL;
		}
		break;
	    case STATE_IN_CODE:
		if (g_ascii_isalpha(buf[i]) == TRUE) {
		    lr->state = STATE_NORMAL;
		}
		break;
	    default:
		g_assert_not_reached();
	}
    }
    while (log_process(lr) == TRUE);
    return TRUE;
}

static gboolean
log_reader_connect(LogReader *lr) {
    GIOChannel *p[2];
    GError *e = NULL;
    GIOChannel *console_io;
    Console c;

    if (lr->log->manager->consoles == NULL) {
	return FALSE;
    }
    c = g_hash_table_lookup(lr->log->manager->consoles, lr->console);
    if (c == NULL) {
	return FALSE;
    }
    console_io = console_get_io_channel(c);
    if (console_io == NULL) {
	return FALSE;
    }
    if (lr->reader != 0) {
	g_source_remove(lr->reader);
    }
    if (lr->io != NULL) {
	g_io_channel_unref(lr->io);
    }
    pipe_new(p);
    lr->io = p[0];
    lr->reader = g_io_add_watch_full(lr->io, G_PRIORITY_DEFAULT,
	    G_IO_IN, log_reader, lr, log_reader_done);
    if (matrix_connect(lr->log->manager->matrix,
		console_io, p[1], &e) == FALSE) {
	g_printerr(_("could not connect to matrix (%s)\n"),
		e->message);
	g_clear_error(&e);
    }
    return TRUE;
}

static void
console_connected_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Log log = value;
    const gchar *console = data;
    LogReader *lr;

    lr = g_hash_table_lookup(log->consoles, console);
    if (lr != NULL) {
	log_reader_connect(lr);
    }
}

static void
console_connected(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_hash_table_foreach(manager->logs, console_connected_fe,
	    (gpointer)event->console_connects.console);
}

static void
console_renamed_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Log log = value;
    const EventData *event = data;
    LogReader *lr;
    GHashTableIter iter;

    lr = g_hash_table_lookup(log->consoles, event->console_renamed.console);
    if (lr != NULL) {
	g_free(lr->console);
	lr->console = g_strdup(event->console_renamed.newname);
	while (config_unset(log->config_path, "console"));
	g_hash_table_iter_init(&iter, log->consoles);
	while (g_hash_table_iter_next(&iter, NULL, (gpointer)&lr) == TRUE) {
	    config_set(log->config_path, "console", lr->console);
	}
    }
}

static void
console_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_hash_table_foreach(manager->logs, console_renamed_fe,
	    (gpointer)event);
}

static void
log_opened(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    Log log = data;
    GError *e;
    Mplex mplex;

    e = message_arg_error_check(arg);
    if (e != NULL) {
	g_printerr("could not open log %s (%s)\n", log->name,
		e->message);
	g_clear_error(&e);
	return;
    }
    g_assert(arg->type == MESSAGE_ARG_UINT16);
    mplex = server_mplex(log->manager, log->server);
    if (mplex == NULL) {
	/* FIXME: report error? */
	return;
    }
    log->io = mplex_open(mplex, arg->uint16, NULL, NULL, NULL, &e);
    if (log->io == NULL) {
	g_printerr(_("failed to open mplex channel for log %s (%s)"),
		log->name, e->message);
	g_clear_error(&e);
	return;
    }
    event_raise(log->manager, EVENT_LOG_CONNECTS, log->name);
}

static void
server_disconnected_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Log log = value;
    const gchar *server = data;

    if (strcmp(server, log->server) == 0 &&
	    log->io != NULL) {
	event_raise(log->manager, EVENT_LOG_DISCONNECTS, log->name);
	matrix_expunge(log->manager->matrix, log->io);
	g_io_channel_unref(log->io);
	log->io = NULL;
    }
}

static void
server_disconnected(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_hash_table_foreach(manager->logs, server_disconnected_fe,
	    (gpointer)event->server_disconnects.server);
}


static void
server_connected_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Log log = value;
    const gchar *server = data;
    MessageQueue mq;
    GError *e = NULL;

    if (strcmp(server, log->server) == 0) {
	mq = server_message_queue(log->manager, log->server);
	if (mq != NULL) {
	    MessageArg *a;
	    GList *list = NULL;

	    list = g_list_append(list,
		    message_arg_string("file", log->filename));
	    a = message_arg_list("", list);
	    if (message_send(mq, SERVER_MESSAGE_OPEN_LOG,
			a, log_opened, log, &e) == FALSE) {
		g_printerr("error sending message (%s)\n", e->message);
		g_clear_error(&e);
	    }
	    message_arg_free(a);
	}
    }
}

static void
server_connected(Manager *manager,
        const EventData *event,
        gpointer data) {
    g_hash_table_foreach(manager->logs, server_connected_fe,
            (gpointer)event->server_connects.server);
}

static void
log_consoles_hash_free(gpointer data) {
    LogReader *lr = data;

    if (lr->console != NULL) {
	g_free(lr->console);
    }
    if (lr->buf != NULL) {
	g_string_free(lr->buf, TRUE);
    }
    if (lr->reader != 0) {
	g_source_remove(lr->reader);
    }
    if (lr->io != NULL) {
	g_io_channel_unref(lr->io);
    }
    g_free(lr);
}

static void
add_console(Log log,
	const gchar *console) {
    LogReader *lr;

    g_assert(log);
    g_assert(console);
    if (g_hash_table_lookup(log->consoles, console) != NULL) {
	return;
    }
    lr = g_new0(LogReader, 1);
    lr->console = g_strdup(console);
    lr->io = NULL;
    lr->buf = g_string_new("");
    lr->reader = 0;
    lr->state = STATE_NORMAL;
    lr->log = log;
    g_hash_table_insert(log->consoles, g_strdup(lr->console), lr);
    log_reader_connect(lr);
    /*FIXME: need an event here?*/
}

static void
remove_console(Log log,
	const gchar *console) {
    LogReader *lr;

    lr = g_hash_table_lookup(log->consoles, console);
    if (lr == NULL) {
	return;
    }
    g_hash_table_remove(log->consoles, lr->console);
    /*FIXME: need an event here?*/
}

extern void
log_load(Manager *manager) {
    ConfigPath cp;
    GList *logs, *l;

    g_assert(manager);
    g_assert(manager->config);

    manager->logs = g_hash_table_new_full(g_str_hash, g_str_equal,
	    g_free, log_hash_free);

    cp = config_path(manager->config, "logs", NULL);
    if (cp == NULL) {
	logs = NULL;
    } else {
	logs = config_children(cp);
	config_path_free(cp);
    }
    for (l = logs; l != NULL; l = g_list_next(l)) {
	Log log = g_new0(struct _Log, 1);
	const gchar *str;
	GList *consoles;

	cp = l->data;
	str = config_get_value(cp, "name");
	if (str == NULL) {
	    g_printerr("configuration file corrupt (anonymous log)\n");
	    g_free(log);
	    continue;
	}
	log->name = g_strdup(str);
	log->manager = manager;
	log->config_path = cp;
	log->consoles = g_hash_table_new_full(g_str_hash, g_str_equal,
		g_free, log_consoles_hash_free);

	str = config_get_value(cp ,"server");
	if (str == NULL) {
	    g_printerr("configuration file corrupt (log %s"
		    " has no server)\n", log->name);
	    g_free(log->name);
	    g_free(log);
	    continue;
	}
	log->server = g_strdup(str);

	str = config_get_value(cp ,"filename");
	if (str == NULL) {
	    log->filename = g_strdup(log->name);
	} else {
	    log->filename = g_strdup(str);
	}

	consoles = config_children(cp);
	while (consoles != NULL) {
	    ConfigPath ccp = consoles->data;
	    const ConfigDirective *cd = config_get(ccp);

	    consoles = g_list_remove(consoles, ccp);
	    if (strcmp(cd->name, "console") != 0 ||
		    cd->value == NULL) {
		config_path_free(ccp);
		continue;
	    }
	    add_console(log, cd->value);
	    config_path_free(ccp);
	}

	g_hash_table_insert(manager->logs, g_strdup(log->name), log);
    }

    event_register(manager, EVENT_CONSOLE_CONNECTS, console_connected, NULL);
    event_register(manager, EVENT_SERVER_CONNECTS,
	    server_connected, NULL);
    event_register(manager, EVENT_SERVER_DISCONNECTS,
	    server_disconnected, NULL);
    event_register(manager, EVENT_CONSOLE_RENAMED, console_renamed, NULL);
    /*FIXME console disconnects/removed? */

}

static void
log_list_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Log log = value;
    GList **l = data;

    *l = g_list_prepend(*l, g_strdup(log->name));
}

extern GList *
log_list(Manager *manager) {
    GList *l = NULL;

    g_hash_table_foreach(manager->logs, log_list_fe, &l);
    return l;
}

static void
log_list_per_console_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Log l = value;
    ListPer *lp = data;

    if (g_hash_table_lookup(l->consoles, lp->target) != NULL) {
	lp->l = g_list_prepend(lp->l, g_strdup(l->name));
    }
}

extern GList *
log_list_per_console(Manager *manager,
	const gchar *console) {
    ListPer lp;

    lp.target = console;
    lp.l = NULL;
    g_hash_table_foreach(manager->logs, log_list_per_console_fe, &lp);
    return lp.l;
}

static void
log_list_per_server_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Log l = value;
    ListPer *lp = data;

    if (strcmp(lp->target, l->server) == 0) {
	lp->l = g_list_prepend(lp->l, g_strdup(l->name));
    }
}

extern GList *
log_list_per_server(Manager *manager,
	const gchar *server) {
    ListPer lp;

    lp.target = server;
    lp.l = NULL;
    g_hash_table_foreach(manager->logs, log_list_per_server_fe, &lp);
    return lp.l;
}

extern SpecLog *
log_spec(Manager *manager,
	const gchar *log,
	GError **error) {
    Log l;
    SpecLog *spec;
    LogReader *lr;
    GHashTableIter iter;

    g_assert(manager);
    g_assert(log);

    l = g_hash_table_lookup(manager->logs, log);
    if (l == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("log %s does not exist"), log);
	return NULL;
    }
    spec = g_new0(struct _SpecLog, 1);
    spec->name = g_strdup(l->name);
    spec->server = g_strdup(l->server);
    spec->file = g_strdup(l->filename);
    spec->consoles = NULL;
    g_hash_table_iter_init(&iter, l->consoles);
    while (g_hash_table_iter_next(&iter, NULL, (gpointer)&lr) == TRUE) {
	spec->consoles = g_list_prepend(spec->consoles, g_strdup(lr->console));
    }
    return spec;
}

extern gboolean
log_add(Manager *manager,
	const SpecLog *spec,
	GError **error) {
    Log l;
    GList *lp;
    ConfigPath logs;
    MessageQueue smq;
    MessageArg *a;
    LogReader *lr;
    GHashTableIter iter;

    g_assert(manager);
    g_assert(spec);

    l = g_hash_table_lookup(manager->logs, spec->name);
    if (l != NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("log %s already exists"), l->name);
	return FALSE;
    }
    smq = server_message_queue(manager, spec->server);
    if (smq == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_GENERAL,
		_("server %s not available"), spec->server);
	return FALSE;
    }

    l = g_new0(struct _Log, 1);
    l->name = g_strdup(spec->name);
    l->manager = manager;
    l->config_path = NULL;
    l->server = g_strdup(spec->server);
    l->filename = g_strdup(spec->file);
    l->io = NULL;
    l->consoles = g_hash_table_new_full(g_str_hash, g_str_equal,
	    g_free, log_consoles_hash_free);
    for (lp = spec->consoles; lp != NULL; lp = g_list_next(lp)) {
	add_console(l, lp->data);
    }
    g_hash_table_insert(manager->logs, g_strdup(l->name), l);
    
    logs = config_path(manager->config, "logs", NULL);
    if (logs == NULL) {
	ConfigPath root = config_path(manager->config, NULL);

	logs = config_path_create(manager->config, root, "logs");
	config_path_free(root);
    }
    l->config_path = config_path_create(manager->config, logs, "log");
    config_path_free(logs);

    config_set(l->config_path, "name", l->name);
    config_set(l->config_path, "server", l->server);
    config_set(l->config_path, "filename", l->filename);
    g_hash_table_iter_init(&iter, l->consoles);
    while (g_hash_table_iter_next(&iter, NULL, (gpointer)&lr) == TRUE) {
	config_set(l->config_path, "console", lr->console);
    }
    config_file_write(manager->config, NULL);

    event_raise(manager, EVENT_LOG_ADDED, l->name);

    lp = g_list_prepend(NULL,
	    message_arg_string("file", l->filename));
    a = message_arg_list("", lp);
    if (message_send(smq, SERVER_MESSAGE_OPEN_LOG,
		a, log_opened, l, error) == FALSE) {
	message_arg_free(a);
	return FALSE;
    }
    message_arg_free(a);

    return TRUE;
}

extern gboolean
log_edit(Manager *manager,
	const gchar *log,
	const SpecLog *spec,
	GError **error) {
    Log l;
    MessageQueue smq;
    gchar *old_name;
    gpointer on_p, l_p;
    MessageArg *a;
    GList *lp;
    GList *m;
    gboolean changed = FALSE;
    gboolean name_changed = FALSE;
    gboolean consoles_changed = FALSE;
    gboolean reconnect = FALSE;
    LogReader *lr;
    GHashTableIter iter;

    g_assert(manager);
    g_assert(log);
    g_assert(spec);

    if (strcmp(log, spec->name)) {
	l = g_hash_table_lookup(manager->logs, spec->name);
	if (l != NULL) {
	    g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		    _("log %s already exists"), l->name);
	    return FALSE;
	}
	changed = TRUE;
	name_changed = TRUE;
    }
    smq = server_message_queue(manager, spec->server);
    if (smq == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_GENERAL,
		_("server %s not available"), spec->server);
	return FALSE;
    }
    if (g_hash_table_lookup_extended(manager->logs, log,
		&on_p, &l_p) == FALSE) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("log %s does not exist"), log);
	return FALSE;
    }
    old_name = on_p;
    l = l_p;

    if (name_changed) {
	g_free(l->name);
	l->name = g_strdup(spec->name);
	config_unset(l->config_path, "name");
        config_set(l->config_path, "name", l->name);
    }

    for (lp = spec->consoles; lp != NULL; lp = g_list_next(lp)) {
	if (g_hash_table_lookup(l->consoles, lp->data) == NULL) {
	    add_console(l, lp->data);
	    consoles_changed = TRUE;
	}
    }
    g_hash_table_iter_init(&iter, l->consoles);
    while (g_hash_table_iter_next(&iter, NULL, (gpointer)&lr) == TRUE) {
	for (m = spec->consoles; m != NULL; m = g_list_next(m)) {
	    if (strcmp(m->data, lr->console) == 0) {
		break;
	    }
	}
	if (m == NULL) {
	    g_hash_table_iter_remove(&iter);
	    consoles_changed = TRUE;
	}
    }
    if (consoles_changed) {
	while (config_unset(l->config_path, "console"));
	g_hash_table_iter_init(&iter, l->consoles);
	while (g_hash_table_iter_next(&iter, NULL, (gpointer)&lr) == TRUE) {
	    config_set(l->config_path, "console", lr->console);
	}
	changed = TRUE;
    }

    if (strcmp(l->server, spec->server) ||
	    strcmp(l->filename, spec->file)) {
	changed = TRUE;
	log_close(l);
	reconnect=TRUE;
	if (l->server) {
	    g_free(l->server);
	}
	l->server = g_strdup(spec->server);
	config_unset(l->config_path, "server");
	config_set(l->config_path, "server", l->server);
	if (l->filename) {
	    g_free(l->filename);
	}
	l->filename = g_strdup(spec->file);
	config_unset(l->config_path, "server");
	config_set(l->config_path, "server", l->filename);
    }
    if (changed) {
	config_file_write(manager->config, NULL);
    }
    if (name_changed) {
	g_hash_table_steal(manager->logs, old_name);
	g_free(old_name);
	g_hash_table_insert(manager->logs, g_strdup(l->name), l);
	event_raise(manager, EVENT_LOG_RENAMED, log, l->name);
    }
    if (reconnect) {
	GError *e = NULL;

	lp = g_list_append(lp,
		message_arg_string("file", l->filename));
	a = message_arg_list("", lp);
	if (message_send(smq, SERVER_MESSAGE_OPEN_LOG,
		    a, log_opened, l, &e) == FALSE) {
	    g_printerr("error sending message (%s)\n", e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
    }
    return TRUE;
}

extern gboolean
log_remove(Manager *manager,
	const gchar *log,
	GError **error) {
    Log l;

    g_assert(manager);
    g_assert(log);

    l = g_hash_table_lookup(manager->logs, log);
    if (l == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("log %s does not exist"), log);
	return FALSE;
    }
    config_delete(l->config_path);
    l->config_path = NULL;
    g_hash_table_remove(manager->logs, log);
    config_file_write(manager->config, NULL);
    event_raise(manager, EVENT_LOG_REMOVED, log);
    return TRUE;
}
