#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include "config.h"
#include "mgr.h"
#include "server.h"
#include "server-connection.h"
#include "console.h"
#include "cgroup.h"
#include "log.h"
#include "client.h"
#include "user.h"
#include "date.h"

GIOChannel *log_file;

typedef struct _EventRecord {
    EventCB cb;
    gpointer data;
} EventRecord;

extern GQuark
manager_error_quark(void) {
    static GQuark error_quark = 0;

    if (error_quark == 0) {
	error_quark = g_quark_from_static_string("MANAGER_ERROR");
    }
    return error_quark;
}

extern void
event_raise(Manager *manager,
	EventType type,
	...) {
    EventData e;
    va_list ap;
    GList *l;

    e.type = type;
    va_start(ap, type);
    switch (type) {
	case EVENT_SERVER_CONNECTS:
	    e.server_connects.server = va_arg(ap, const gchar *);
	    break;
	case EVENT_SERVER_DISCONNECTS:
	    e.server_disconnects.server = va_arg(ap, const gchar *);
	    break;
	case EVENT_CONSOLE_CONNECTS:
	    e.console_connects.console = va_arg(ap, const gchar *);
	    break;
	case EVENT_CONSOLE_DISCONNECTS:
	    e.console_disconnects.console = va_arg(ap, const gchar *);
	    break;
	case EVENT_LOG_CONNECTS:
	    e.log_connects.log = va_arg(ap, const gchar *);
	    break;
	case EVENT_LOG_DISCONNECTS:
	    e.log_disconnects.log = va_arg(ap, const gchar *);
	    break;
	case EVENT_CGROUP_CHANGED:
	    e.cgroup_changed.group = va_arg(ap, const gchar *);
	    e.cgroup_changed.console = va_arg(ap, const gchar *);
	    e.cgroup_changed.added = va_arg(ap, int);
	    break;
	case EVENT_CGROUP_ADDED:
	    e.cgroup_added.cgroup = va_arg(ap, const gchar *);
	    break;
	case EVENT_CGROUP_REMOVED:
	    e.cgroup_removed.cgroup = va_arg(ap, const gchar *);
	    break;
	case EVENT_CGROUP_RENAMED:
	    e.cgroup_renamed.cgroup = va_arg(ap, const gchar *);
	    e.cgroup_renamed.newname = va_arg(ap, const gchar *);
	    break;
	case EVENT_CONSOLE_ADDED:
	    e.console_added.console = va_arg(ap, const gchar *);
	    break;
	case EVENT_CONSOLE_REMOVED:
	    e.console_removed.console = va_arg(ap, const gchar *);
	    break;
	case EVENT_CONSOLE_RENAMED:
	    e.console_renamed.console = va_arg(ap, const gchar *);
	    e.console_renamed.newname = va_arg(ap, const gchar *);
	    break;
	case EVENT_LOG_ADDED:
	    e.log_added.log = va_arg(ap, const gchar *);
	    break;
	case EVENT_LOG_REMOVED:
	    e.log_removed.log = va_arg(ap, const gchar *);
	    break;
	case EVENT_LOG_RENAMED:
	    e.log_renamed.log = va_arg(ap, const gchar *);
	    e.log_renamed.newname = va_arg(ap, const gchar *);
	    break;
	case EVENT_USER_ADDED:
	    e.user_added.user = va_arg(ap, const gchar *);
	    break;
	case EVENT_USER_REMOVED:
	    e.user_removed.user = va_arg(ap, const gchar *);
	    break;
	case EVENT_USER_RENAMED:
	    e.user_renamed.user = va_arg(ap, const gchar *);
	    e.user_renamed.newname = va_arg(ap, const gchar *);
	    break;
	case EVENT_SERVER_ADDED:
	    e.server_added.server = va_arg(ap, const gchar *);
	    break;
	case EVENT_SERVER_REMOVED:
	    e.server_removed.server = va_arg(ap, const gchar *);
	    break;
	case EVENT_SERVER_RENAMED:
	    e.server_renamed.server = va_arg(ap, const gchar *);
	    e.server_renamed.newname = va_arg(ap, const gchar *);
	    break;
	default:
	    g_assert_not_reached();
    }
    va_end(ap);
    for (l = manager->event[type]; l != NULL; l = g_list_next(l)) {
	EventRecord *er = l->data;

	er->cb(manager, &e, er->data);
    }
}

extern void
event_register(Manager *manager,
	EventType type,
	EventCB cb,
	gpointer data) {
    EventRecord *er = g_new0(EventRecord,1);

    er->cb = cb;
    er->data = data;
    manager->event[type] = g_list_prepend(manager->event[type], er);
}

static void
log_file_write(const gchar *message) {
    GString *s;
    gchar *timestr;
    gint i;
    gboolean newline = TRUE;

    g_assert(message);
    s = g_string_new("");
    timestr = date_print(date_now());
    for (i = 0; message[i] != '\0'; i++) {
	if (newline == TRUE) {
	    s = g_string_append(s, timestr);
	    s = g_string_append_c(s, ' ');
	    newline = FALSE;
	}
	s = g_string_append_c(s, message[i]);
	if (message[i] == '\n') {
	    newline = TRUE;
	}
    }
    g_free(timestr);
    g_io_channel_write_chars(log_file, s->str, s->len, NULL, NULL);
    g_io_channel_flush(log_file, NULL);
    g_string_free(s, TRUE);
}

int
main(int argc, char *argv[]) {
    Manager *manager = g_new0(Manager, 1);
    GError *e = NULL;
    gchar *config_filename = CONCSYSCONFDIR "/conc.conf";
    gint debug = 0;
    gint i;
    GOptionEntry options[] = {
	{ "debug", 'd', 0, G_OPTION_ARG_NONE, &debug,
	    N_("enter debug mode"), NULL },
	{ "config-file", 'c', 0, G_OPTION_ARG_STRING, &config_filename,
	    N_("specify location of configuration file"), N_("FILE") },
	{ NULL }
    };
    GOptionContext *option_context;

    option_context = g_option_context_new(NULL);
    g_option_context_add_main_entries(option_context, options,
	GETTEXT_PACKAGE);
    manager->program = gnome_program_init("concmgr", PACKAGE_VERSION,
	    LIBGNOME_MODULE, argc, argv,
	    GNOME_PARAM_GOPTION_CONTEXT, option_context,
	    GNOME_PARAM_NONE);
    if (debug == 1) {
	g_printerr("entering debug mode\n");
	g_log_set_always_fatal(G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL |
	    G_LOG_LEVEL_WARNING);
    }
    manager->main_loop = g_main_loop_new(NULL, FALSE);
    manager->config = config_file_open(config_filename, &e);
    if (manager->config == NULL) {
	g_printerr("could not load configuration file %s (%s)\n",
		config_filename, e->message);
	exit(EXIT_FAILURE);
    }
    for (i = 0; i < EVENT_LAST; i++) {
	manager->event[i] = NULL;
    }
    g_get_current_time(&manager->time);
    manager->matrix = matrix_new();
    if (debug == 0) {
	log_file = g_io_channel_new_file(CONCLOGDIR "/conc.log", "a", &e);
	if (log_file == NULL) {
	    g_printerr("could not open log file " CONCLOGDIR "/conc.log (%s)\n",
		    e->message);
	    exit(EXIT_FAILURE);
	}
	g_set_print_handler(log_file_write);
	g_set_printerr_handler(log_file_write);
	/* FIXME: daemon() is not very portable */
	daemon(0, 0);
    }
    log_load(manager);
    console_load(manager);
    cgroup_load(manager);
    server_load(manager);
    client_load(manager);
    user_load(manager);
    g_main_loop_run(manager->main_loop);
    exit(EXIT_SUCCESS);
}
