#ifndef __MGR_H
#define __MGR_H

#include <gnet.h>
#include <libgnome/libgnome.h>
#include "i18n.h"
#include "config-file.h"
#include "matrix.h"

/*
 * ManagerError - errors generated within the manager
 *
 */
typedef enum {
    MANAGER_ERROR_DOES_NOT_EXIST,
    MANAGER_ERROR_PERMISSION_DENIED,
    MANAGER_ERROR_GENERAL
} ManagerError;
#define MANAGER_ERROR manager_error_quark()
extern GQuark manager_error_quark(void);

/*
 * EventType - an emueration of all possible events
 *
 */
typedef enum {
    EVENT_SERVER_CONNECTS,
    EVENT_SERVER_DISCONNECTS,
    EVENT_CONSOLE_CONNECTS,
    EVENT_CONSOLE_DISCONNECTS,
    EVENT_LOG_CONNECTS,
    EVENT_LOG_DISCONNECTS,
    EVENT_CGROUP_CHANGED,
    EVENT_CGROUP_ADDED,
    EVENT_CGROUP_REMOVED,
    EVENT_CGROUP_RENAMED,
    EVENT_CONSOLE_ADDED,
    EVENT_CONSOLE_REMOVED,
    EVENT_CONSOLE_RENAMED,
    EVENT_LOG_ADDED,
    EVENT_LOG_REMOVED,
    EVENT_LOG_RENAMED,
    EVENT_USER_ADDED,
    EVENT_USER_REMOVED,
    EVENT_USER_RENAMED,
    EVENT_SERVER_ADDED,
    EVENT_SERVER_REMOVED,
    EVENT_SERVER_RENAMED,
    EVENT_LAST
} EventType;

/*
 * Manager - the root structure of a concmgr - no globals here mate!
 *
 */
typedef struct _Manager {
    GnomeProgram *program;
    GMainLoop *main_loop;
    ConfigFile config;
    Matrix matrix;
    GTcpSocket *sock; /* socket for listening for clients */
    GHashTable *servers; /* hash of Server */
    GHashTable *consoles; /* hash of Console */
    GHashTable *cgroups; /* hash of CGroup */
    GHashTable *logs; /* hash of Log */
    GList *clients; /* list of Client */
    GHashTable *users; /* hash of User */
    GList *event[EVENT_LAST];
    GTimeVal time;
} Manager;

/*
 * EventData - data associtaed with an event
 *
 * The valid member of the union is dictate by type.
 *
 */
typedef struct _EventData {
    EventType type;
    union {
	struct {
	    const gchar *server;
	} server_connects;
	struct {
	    const gchar *server;
	} server_disconnects;
	struct {
	    const gchar *console;
	} console_connects;
	struct {
	    const gchar *console;
	} console_disconnects;
	struct {
	    const gchar *log;
	} log_connects;
	struct {
	    const gchar *log;
	} log_disconnects;
	struct {
	    const gchar *group;
	    const gchar *console;
	    guint8 added;
	} cgroup_changed;
	struct {
	    const gchar *cgroup;
	} cgroup_added;
	struct {
	    const gchar *cgroup;
	} cgroup_removed;
	struct {
	    const gchar *cgroup;
	    const gchar *newname;
	} cgroup_renamed;
	struct {
	    const gchar *console;
	} console_added;
	struct {
	    const gchar *console;
	} console_removed;
	struct {
	    const gchar *console;
	    const gchar *newname;
	} console_renamed;
	struct {
	    const gchar *log;
	} log_added;
	struct {
	    const gchar *log;
	} log_removed;
	struct {
	    const gchar *log;
	    const gchar *newname;
	} log_renamed;
	struct {
	    const gchar *user;
	} user_added;
	struct {
	    const gchar *user;
	} user_removed;
	struct {
	    const gchar *user;
	    const gchar *newname;
	} user_renamed;
	struct {
	    const gchar *server;
	} server_added;
	struct {
	    const gchar *server;
	} server_removed;
	struct {
	    const gchar *server;
	    const gchar *newname;
	} server_renamed;
    };
} EventData;

/*
 * EventCB - callback initiated when an event occurs
 *
 */
typedef void (*EventCB) (Manager *manager,
	const EventData *event,
	gpointer data);

/*
 * event_raise - raise an event
 *
 * The arguments are the memebers of the apropriate struct in EventData, as
 * indicated by the type. Where an event refers to some resource becoming
 * unavailable, the event should be raised before that resource is freed.
 *
 */
extern void event_raise(Manager *manager,
	EventType type,
	...);

/*
 * event_register - register a callback on an event.
 *
 */
extern void event_register(Manager *manager,
	EventType type,
	EventCB cb,
	gpointer data);

#endif
