#include <string.h>
#include <gnet.h>
#include "config.h"
#include "stdinout.h"
#include "mplex.h"
#include "message.h"
#include "server.h"
#include "server-connection.h"

typedef void (*ConnectionDataFreeCB) (ServerConnection c);

struct _ServerConnection {
    gint sin;
    gint sout;
    GIOChannel *io;
    Mplex mplex;
    MessageQueue msg_q;
    ConnectionDataFreeCB connection_data_free;
    gpointer connection_data;
    ServerConnectionClosedCB cb;
    gpointer user_data;
};

static void
server_quits(Mplex mplex,
	gpointer data) {
    ServerConnection server = data;

    mplex_free(server->mplex);
    server->mplex = NULL;
    g_io_channel_unref(server->io);
    server->io = NULL;
    server->cb(server, server->user_data, NULL);
}

static ServerConnection
server_build(GIOChannel *io,
	ServerConnectionClosedCB cb,
	gpointer user_data,
	GError **error) {
    ServerConnection server = g_new0(struct _ServerConnection, 1);
    GIOChannel *mq_io;

    server->io = io;
    server->cb = cb;
    server->user_data = user_data;
    g_io_channel_set_close_on_unref(server->io, TRUE);
    server->mplex = mplex_new(server->io, server_quits, server, error);
    if (server->mplex == NULL) {
	g_io_channel_unref(server->io);
	g_free(server);
	return NULL;
    }
    mq_io = mplex_open(server->mplex, CONC_MSG_CHANNEL, NULL, NULL, NULL, error);
    if (mq_io == NULL) {
	mplex_free(server->mplex);
	g_io_channel_unref(server->io);
	g_free(server);
	return NULL;
    }
    server->msg_q = message_queue_new(mq_io, error);
    if (server->msg_q == NULL) {
	mplex_free(server->mplex);
	g_io_channel_unref(server->io);
	g_free(server);
	return NULL;
    }
    /* register handlers */
    message_queue_start(server->msg_q);
    mplex_start(server->mplex);
    return server;
}
 
extern ServerConnection
server_exec(const gchar *path,
	ServerConnectionClosedCB cb,
	gpointer user_data,
	GError **error) {
    gchar *argv[3];
    gint sin, sout;

    g_assert(cb);
    
    if (path == NULL) {
	argv[0] = "concsvr";
    } else {
	argv[0] = (gchar *)path;
    }
    argv[1] = "--pipe";
    argv[2] = NULL;
    if (g_spawn_async_with_pipes(NULL, argv, NULL,
		G_SPAWN_SEARCH_PATH, NULL, NULL, NULL,
		&sin, &sout, NULL, error) == FALSE) {
	return NULL;
    }
    return server_build(stdinout_new_with_fds(sout, sin),
	    cb, user_data, error);
}

extern ServerConnection
server_connect(const gchar *host,
	gint port,
	ServerConnectionClosedCB cb,
	gpointer user_data,
	GError **error) {
    ServerConnection c;
    GTcpSocket *sock;

    g_assert(host);
    g_assert(cb);

    if (port < 0) {
	port = CONC_PORT_SVR;
    }
    sock = gnet_tcp_socket_connect(host, port);
    if (sock == NULL) {
	g_set_error(error, SERVER_CONNECTION_ERROR, SERVER_CONNECTION_FAILED,
		"failed to connect a TCP socket to %s", host);
	return NULL;
    }
    c = server_build(gnet_tcp_socket_get_io_channel(sock),
	    cb, user_data, error);
    /*
     no need to delete socket - it us unreffed elswhere
    if (c) {
	c->connection_data = sock;
	c->connection_data_free = (ConnectionDataFreeCB)gnet_tcp_socket_delete;
    }
    */
    return c;
}
 
extern ServerConnection
server_ssh(const gchar *host,
	const gchar *program,
	const gchar *user,
	const gchar *identity_file,
	const gchar *remote_cmd,
	ServerConnectionClosedCB cb,
	gpointer user_data,
	GError **error) {
    const gchar *argv[8];
    gint argc = 0;
    gint sin, sout;

    g_assert(host);
    g_assert(program);
    g_assert(remote_cmd);
    g_assert(cb);
    
    argv[argc++] = program;
    if (user != NULL) {
	argv[argc++] = "-l";
	argv[argc++] = user;
    }
    if (identity_file != NULL) {
	argv[argc++] = "-i";
	argv[argc++] = identity_file;
    }
    argv[argc++] = host;
    argv[argc++] = remote_cmd;
    argv[argc++] = NULL;

    if (g_spawn_async_with_pipes(NULL, (gchar **) argv, NULL,
		G_SPAWN_SEARCH_PATH, NULL, NULL, NULL,
		&sin, &sout, NULL, error) == FALSE) {
	return NULL;
    }
    return server_build(stdinout_new_with_fds(sout, sin),
	    cb, user_data, error);
}

extern void
server_connection_free(ServerConnection s) {
    if (s->io != NULL) {
	g_io_channel_unref(s->io);
    }
    if (s->msg_q != NULL) {
	message_queue_free(s->msg_q);
    }
    if (s->mplex != NULL) {
	mplex_free(s->mplex);
    }
    if (s->connection_data_free != NULL) {
	s->connection_data_free(s->connection_data);
    }
    g_free(s);
}

extern MessageQueue
server_connection_message_queue(ServerConnection s) {
    g_assert(s);
    return s->msg_q;
}

extern Mplex
server_connection_mplex(ServerConnection s) {
    g_assert(s);
    return s->mplex;
}

extern GQuark
server_connection_error_quark(void) {
    static GQuark error_quark = 0;

    if (error_quark == 0) {
	error_quark = g_quark_from_static_string("SERVER_CONNECTION_ERROR");
    }
    return error_quark;
}
