#include <string.h>
#include "server.h"
#include "server-connection.h"
#include "svr-message.h"
#include "config-file.h"
#include "retry.h"
#include "spec-server.h"
#include "console.h"
#include "log.h"

#define RETRY_INTERVAL 10
#define RETRY_MULTIPLIER 1.5
#define RETRY_MAX 20
#define PING_TIME 5

struct _Server {
    gchar *name;
    Manager *manager;
    ServerType type;
    Retry connection_retry;
    Retry ping_retry;
    gint ping_count;
    gint connection_complete;
    ServerConnection connection;
    ConfigPath config_path;
    union {
	ServerExec exec;
	ServerConnect connect;
	ServerSsh ssh;
    };
};

static void
server_hash_free(gpointer data) {
    Server server = data;

    if (server->connection_retry != NULL) {
	retry_free(server->connection_retry);
    }
    if (server->ping_retry != NULL) {
	retry_free(server->ping_retry);
    }
    if (server->connection != NULL) {
	server_connection_free(server->connection);
    }
    g_free(server);
}

static void
server_config_write(Server s,
	gboolean flush) {
    gchar *tmp;
    ConfigPath servers;

    g_assert(s);

    servers = config_path(s->manager->config, "servers", NULL);
    if (servers == NULL) {
	ConfigPath root = config_path(s->manager->config, NULL);

	servers = config_path_create(s->manager->config, root, "servers");
	config_path_free(root);
    }
    if (s->config_path != NULL) {
	config_delete(s->config_path);
    }
    s->config_path = config_path_create(s->manager->config, servers, "server");
    config_path_free(servers);
    config_set(s->config_path, "name", s->name);
    switch (s->type) {
	case SERVER_TYPE_EXEC:
	    config_set(s->config_path, "type", "exec");
	    if (s->exec.path) {
		config_set(s->config_path, "path", s->exec.path);
	    }
	    break;
	case SERVER_TYPE_CONNECT:
	    config_set(s->config_path, "type", "connect");
	    if (s->connect.host) {
		config_set(s->config_path, "host", s->connect.host);
	    }
	    tmp = g_strdup_printf("%d", s->connect.port);
	    config_set(s->config_path, "port", tmp);
	    g_free(tmp);
	    break;
	case SERVER_TYPE_SSH:
	    config_set(s->config_path, "type", "ssh");
	    if (s->ssh.host) {
		config_set(s->config_path, "host", s->ssh.host);
	    }
	    if (s->ssh.ssh_program) {
		config_set(s->config_path, "ssh_program", s->ssh.ssh_program);
	    }
	    if (s->ssh.ssh_user) {
		config_set(s->config_path, "ssh_user", s->ssh.ssh_user);
	    }
	    if (s->ssh.ssh_identity) {
		config_set(s->config_path, "ssh_identity", s->ssh.ssh_identity);
	    }
	    if (s->ssh.remote_cmd) {
		config_set(s->config_path, "remote_cmd", s->ssh.remote_cmd);
	    }
	    break;
	default:
	    g_assert_not_reached();
    }
    if (flush) {
	config_file_write(s->manager->config, NULL);
    }
}

static gboolean
connection_retry_attempt(Retry retry,
	gboolean attempt_no,
	gpointer user_data) {
    Server server = user_data;
    GError *e = NULL;

    if (server->connection != NULL) {
	if (server->connection_complete == FALSE) {
	    g_printerr("awaiting ping reply from server %s\n",
			server->name);
	    return FALSE;
	} else {
	    return TRUE;
	}
    }
    g_printerr("attempting reconnection of server %s (attempt %d)\n",
		server->name, attempt_no);
    if (server_make_connection(server, &e) == FALSE) {
	g_printerr("could not open connection to server %s (%s)\n",
		server->name, e->message);
	g_clear_error(&e);
    }
    return FALSE;
}

static void
connection_retry_done(Retry retry,
	gboolean successful,
	gint attempt_no,
	gpointer user_data) {
    Server server = user_data;

    server->connection_retry = NULL;
    retry_free(retry);
    if (successful == FALSE) {
	g_printerr("giving up reconnecting to server %s after %d attempts\n",
		server->name, attempt_no);
    }
}

static void
server_close_and_retry(Server s) {
    if (s->ping_retry != NULL) {
	retry_free(s->ping_retry);
	s->ping_retry = NULL;
    }
    event_raise(s->manager, EVENT_SERVER_DISCONNECTS, s->name);
    /* FIXME
     * is some circumstances s->connection->io has one ref too few and
     * this causes a segfault. 
    if (s->connection != NULL) server_connection_free(s->connection);*/
    s->connection = NULL;
    if (s->connection_retry == NULL) {
	s->connection_retry = retry_new(connection_retry_attempt,
		connection_retry_done, RETRY_INTERVAL, RETRY_MULTIPLIER,
		RETRY_MAX, s);
    }
}

static void
server_closed(ServerConnection s,
	gpointer user_data,
	GError *err) {
    Server server = user_data;

    if (err) {
	g_printerr("server %s closed connection unexpectedly (%s)\n",
		server->name, err->message);
	g_error_free(err);
    } else {
	g_printerr("server %s closed connection unexpectedly\n", server->name);
    }
    server_close_and_retry(server);
}

static void
server_pong(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    Server s = data;

    s->ping_count = 0;
    s->connection_complete = TRUE;
}

static gboolean
server_ping(Retry retry,
	gint attempt_no,
	gpointer user_data) {
    Server s = user_data;
    MessageQueue mq;
    MessageArg *a;
    GError *e = NULL;

    if (s->connection == NULL || s->ping_count > 1) {
	return TRUE;
    }
    mq = server_connection_message_queue(s->connection);
    if (mq == NULL) {
	return TRUE;
    }
    a = message_arg_none("");
    if (message_send(mq, SERVER_MESSAGE_PING, a, server_pong, s, &e) == FALSE) {
	g_clear_error(&e);
	message_arg_free(a);
	return TRUE;
    }
    message_arg_free(a);
    s->ping_count++;
    return FALSE;
}

static void
server_ping_failed(Retry retry,
	gboolean successful,
	gint attempt_no,
	gpointer user_data) {
    Server s = user_data;

    g_printerr("server %s failed to respond to ping, reconnecting\n", s->name);
    server_close_and_retry(s);
}

extern gboolean
server_make_connection(Server s,
	GError **error) {
    if (s->connection != NULL) {
	return TRUE;
    }
    switch (s->type) {
	case SERVER_TYPE_EXEC:
	    s->connection = server_exec(s->exec.path,
		    server_closed, s, error);
	    break;
	case SERVER_TYPE_CONNECT:
	    s->connection = server_connect(s->connect.host,
		    s->connect.port, server_closed, s, error);
	    break;
	case SERVER_TYPE_SSH: 
	    s->connection = server_ssh(s->ssh.host,
		    s->ssh.ssh_program, s->ssh.ssh_user,
		    s->ssh.ssh_identity, s->ssh.remote_cmd,
		    server_closed, s, error);
	    break;
    }
    if (s->connection != NULL) {
	event_raise(s->manager, EVENT_SERVER_CONNECTS, s->name);
	s->ping_retry = retry_new(server_ping, server_ping_failed,
	    PING_TIME, 1.0, -1, s);
	s->ping_count = 0;
	s->connection_complete = FALSE;
	return TRUE;
    } else {
	return FALSE;
    }
}

extern void
server_load(Manager *manager) {
    ConfigPath cp;
    GList *servers, *l;
    GError *e = NULL;

    g_assert(manager);
    g_assert(manager->config);

    manager->servers = g_hash_table_new_full(g_str_hash, g_str_equal,
	    g_free, server_hash_free);

    cp = config_path(manager->config, "servers", NULL);
    if (cp == NULL) {
	servers = NULL;
    } else {
	servers = config_children(cp);
	config_path_free(cp);
    }
    for (l = servers; l != NULL; l = g_list_next(l)) {
	Server s = g_new0(struct _Server, 1);
	const gchar *str;

	cp = l->data;
	str = g_strdup(config_get_value(cp, "name"));
	if (str == NULL) {
	    g_printerr("configuration file corrupt (anonymous server)\n");
	    continue;
	}
	s->name =  g_strdup(str);
	s->manager = manager;
	s->config_path = cp;
	s->connection_retry = NULL;
	s->ping_retry = NULL;

	str = config_get_value(cp, "type");
	if (str == NULL) {
	    str = "exec";
	}
	if (strcmp(str, "exec") == 0) {
	    s->type = SERVER_TYPE_EXEC;
	}
	if (strcmp(str, "connect") == 0) {
	    s->type = SERVER_TYPE_CONNECT;
	}
	if (strcmp(str, "ssh") == 0) {
	    s->type = SERVER_TYPE_SSH;
	}

	switch (s->type) {
	    case SERVER_TYPE_EXEC:
		str = config_get_value(cp, "path");
		if (str == NULL) {
		    s->exec.path = NULL;
		} else {
		    s->exec.path = g_strdup(str);
		}
		break;
	    case SERVER_TYPE_CONNECT:
		str = config_get_value(cp, "host");
		if (str == NULL) {
		    s->connect.host = g_strdup("localhost");
		} else {
		    s->connect.host = g_strdup(str);
		}
		str = config_get_value(cp, "port");
		if (str == NULL) {
		    s->connect.port = CONC_PORT_SVR;
		} else {
		    s->connect.port = (gint) g_strtod(str, NULL);
		}
		break;
	    case SERVER_TYPE_SSH:
		str = config_get_value(cp, "host");
		if (str == NULL) {
		    s->ssh.host = g_strdup("localhost");
		} else {
		    s->ssh.host = g_strdup(str);
		}
		str = config_get_value(cp, "ssh_program");
		if (str == NULL) {
		    s->ssh.ssh_program = g_strdup("ssh");
		} else {
		    s->ssh.ssh_program = g_strdup(str);
		}
		str = config_get_value(cp, "ssh_user");
		if (str == NULL) {
		    s->ssh.ssh_user = NULL;
		} else {
		    s->ssh.ssh_user = g_strdup(str);
		}
		str = config_get_value(cp, "ssh_identity");
		if (str == NULL) {
		    s->ssh.ssh_identity = NULL;
		} else {
		    s->ssh.ssh_identity = g_strdup(str);
		}
		str = config_get_value(cp, "remote_cmd");
		if (str == NULL) {
		    s->ssh.remote_cmd = g_strdup(CONCBINDIR "/concsvr --pipe");
		} else {
		    s->ssh.remote_cmd = g_strdup(str);
		}
		break;
	}

	g_hash_table_insert(manager->servers, g_strdup(s->name), s);
	
	if (server_make_connection(s, &e) == FALSE) {
	    g_printerr("could not open connection to server %s (%s)\n",
		    s->name, e->message);
	    g_clear_error(&e);
	    s->connection_retry = retry_new(connection_retry_attempt,
		connection_retry_done, RETRY_INTERVAL, RETRY_MULTIPLIER,
		RETRY_MAX, s);
	}
    }
    g_list_free(servers);
}

extern MessageQueue
server_message_queue(Manager *manager,
	const gchar *server) {
    Server s = g_hash_table_lookup(manager->servers, server);

    if (s == NULL || s->connection == NULL) {
	return NULL;
    }
    return server_connection_message_queue(s->connection);
}

extern Mplex
server_mplex(Manager *manager,
	const gchar *server) {
    Server s = g_hash_table_lookup(manager->servers, server);

    if (s == NULL || s->connection == NULL) {
	return NULL;
    }
    return server_connection_mplex(s->connection);
}

static void
server_list_fe(gpointer key,
	gpointer value,
	gpointer data) {
    Server s = value;
    GList **l = data;

    *l = g_list_prepend(*l, g_strdup(s->name));
}

extern GList *
server_list(Manager *manager) {
    GList *l = NULL;

    g_hash_table_foreach(manager->servers, server_list_fe, &l);
    return l;
}

extern SpecServer *
server_spec(Manager *manager,
	const gchar *server,
	GError **error) {
    Server s;
    SpecServer *ss;

    g_assert(manager);
    g_assert(server);

    s = g_hash_table_lookup(manager->servers, server);
    if (s == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("server %s does not exist"), server);
	return FALSE;
    }
    ss = g_new0(struct _SpecServer, 1);
    ss->name = g_strdup(s->name);
    ss->type = s->type;
    switch (ss->type) {
	case SERVER_TYPE_EXEC:
	    ss->exec.path = g_strdup(s->exec.path?s->exec.path:"");
	    break;
	case SERVER_TYPE_CONNECT:
	    ss->connect.host = g_strdup(s->connect.host?s->connect.host:"");
	    ss->connect.port = s->connect.port;
	    break;
	case SERVER_TYPE_SSH:
	    ss->ssh.host = g_strdup(s->ssh.host?s->ssh.host:"");
	    ss->ssh.ssh_program =
		g_strdup(s->ssh.ssh_program?s->ssh.ssh_program:"");
	    ss->ssh.ssh_user = g_strdup(s->ssh.ssh_user?s->ssh.ssh_user:"");
	    ss->ssh.ssh_identity =
		g_strdup(s->ssh.ssh_identity?s->ssh.ssh_identity:"");
	    ss->ssh.remote_cmd =
		g_strdup(s->ssh.remote_cmd?s->ssh.remote_cmd:"");
	    break;
    }
    return ss;
}

extern gboolean
server_add(Manager *manager,
	const SpecServer *spec,
	GError **error) {
    Server s;

    g_assert(manager);
    g_assert(spec);

    s = g_hash_table_lookup(manager->servers, spec->name);
    if (s != NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("server %s already exists"), s->name);
	return FALSE;
    }

    s = g_new0(struct _Server, 1);
    s->name = g_strdup(spec->name);
    s->manager = manager;
    s->config_path = NULL;
    s->type = spec->type;
    switch (s->type) {
	case SERVER_TYPE_EXEC:
	    if (spec->exec.path && strlen(spec->exec.path) > 0) {
		s->exec.path = g_strdup(spec->exec.path);
	    } else {
		s->exec.path = NULL;
	    }
	    break;
	case SERVER_TYPE_CONNECT:
	    if (spec->connect.host && strlen(spec->connect.host) > 0) {
		s->connect.host = g_strdup(spec->connect.host);
	    } else {
		s->connect.host = g_strdup("localhost");
	    }
	    s->connect.port = spec->connect.port;
	    break;
	case SERVER_TYPE_SSH:
	    if (spec->ssh.host && strlen(spec->ssh.host) > 0) {
		s->ssh.host = g_strdup(spec->ssh.host);
	    } else {
		s->ssh.host = g_strdup("localhost");
	    }
	    if (spec->ssh.ssh_program && strlen(spec->ssh.ssh_program) > 0) {
		s->ssh.ssh_program = g_strdup(spec->ssh.ssh_program);
	    } else {
		s->ssh.ssh_program = g_strdup("ssh");
	    }
	    if (spec->ssh.ssh_user && strlen(spec->ssh.ssh_user) > 0) {
		s->ssh.ssh_user = g_strdup(spec->ssh.ssh_user);
	    } else {
		s->ssh.ssh_user = NULL;
	    }
	    if (spec->ssh.ssh_identity && strlen(spec->ssh.ssh_identity) > 0) {
		s->ssh.ssh_identity = g_strdup(spec->ssh.ssh_identity);
	    } else {
		s->ssh.ssh_identity = NULL;
	    }
	    if (spec->ssh.remote_cmd && strlen(spec->ssh.remote_cmd) > 0) {
		s->ssh.remote_cmd = g_strdup(spec->ssh.remote_cmd);
	    } else {
		s->ssh.remote_cmd = g_strdup("concsvr");
	    }
	    break;
	default:
	    g_assert_not_reached();
    }

    g_hash_table_insert(manager->servers, g_strdup(s->name), s);
    if (server_make_connection(s, error) == FALSE) {
	g_hash_table_remove(manager->servers, s->name);
	return FALSE;
    }
    server_config_write(s, TRUE);

    event_raise(manager, EVENT_SERVER_ADDED, s->name);
    return TRUE;
}

static gboolean
string_changed(const gchar *a,
	const gchar *b) {
    if ((a == NULL && b != NULL && strlen(b) > 0) ||
	    (b == NULL && a != NULL && strlen(a) > 0) ||
	    (a != NULL && a != NULL && strcmp(a, b) != 0)) {
	return TRUE;
    }
    return FALSE;
}

extern gboolean
server_edit(Manager *manager,
	const gchar *server,
	const SpecServer *spec,
	GError **error) {
    Server s;
    gchar *old_name;
    gpointer on_p, s_p;
    gboolean spec_changed = FALSE;
    gboolean name_changed = FALSE;

    g_assert(manager);
    g_assert(server);
    g_assert(spec);

    if (strcmp(server, spec->name)) {
	s = g_hash_table_lookup(manager->servers, spec->name);
	if (s != NULL) {
	    g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		    _("server %s already exists"), s->name);
	    return FALSE;
	}
	name_changed = TRUE;
    }
    if (g_hash_table_lookup_extended(manager->servers, server,
		&on_p, &s_p) == FALSE) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("server %s does not exist"), server);
	return FALSE;
    }
    old_name = on_p;
    s = s_p;

    if (spec->type != s->type) {
	spec_changed = TRUE;
    } else {
	switch(s->type) {
	    case SERVER_TYPE_EXEC:
		if (string_changed(s->exec.path, spec->exec.path)) {
		    spec_changed = TRUE;
		}
		break;
	    case SERVER_TYPE_CONNECT:
		if (string_changed(s->connect.host, spec->connect.host) ||
			s->connect.port != spec->connect.port) {
		    spec_changed = TRUE;
		}
		break;
	    case SERVER_TYPE_SSH:
		if (string_changed(s->ssh.host, spec->ssh.host) ||
			string_changed(s->ssh.ssh_program,
			    spec->ssh.ssh_program) ||
			string_changed(s->ssh.ssh_user, spec->ssh.ssh_user) ||
			string_changed(s->ssh.ssh_identity,
			    spec->ssh.ssh_identity) ||
			string_changed(s->ssh.remote_cmd,
			    spec->ssh.remote_cmd)) {
		    spec_changed = TRUE;
		}
		break;
	    default:
		g_assert_not_reached();
	}
    }
    if (spec_changed) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_GENERAL,
		_("cannot edit server spec (try adding new server)"));
	return FALSE;
    }
    if (name_changed == FALSE) {
	return TRUE;
    }
    g_free(s->name);
    s->name = g_strdup(spec->name);
    config_unset(s->config_path, "name");
    config_set(s->config_path, "name", s->name);
    config_file_write(manager->config, NULL);
    g_hash_table_steal(manager->servers, old_name);
    g_free(old_name);
    g_hash_table_insert(manager->servers, g_strdup(s->name), s);
    event_raise(manager, EVENT_SERVER_RENAMED, server, s->name);
    return TRUE;
}

extern
gboolean server_remove(Manager *manager,
	const gchar *server,
	GError **error) {
    Server s;
    GList *l;

    g_assert(manager);
    g_assert(server);

    s = g_hash_table_lookup(manager->servers, server);
    if (s == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("server %s does not exist"), server);
	return FALSE;
    }
    l = console_list_per_server(manager, server);
    if (l != NULL) {
	GString *msg = g_string_new("");

	while (l != NULL) {
	    msg = g_string_append(msg, l->data);
	    msg = g_string_append_c(msg, ' ');
	    g_free(l->data);
	    l = g_list_remove(l, l->data);
	}
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("server %s still has these consoles connected: %s"), server,
		msg->str);
	g_string_free(msg, TRUE);
	return FALSE;
    }
    l = log_list_per_server(manager, server);
    if (l != NULL) {
	GString *msg = g_string_new("");

	while (l != NULL) {
	    msg = g_string_append(msg, l->data);
	    msg = g_string_append_c(msg, ' ');
	    g_free(l->data);
	    l = g_list_remove(l, l->data);
	}
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("server %s still has these logs connected: %s"), server,
		msg->str);
	g_string_free(msg, TRUE);
	return FALSE;
    }
    config_delete(s->config_path);
    s->config_path = NULL;
    g_hash_table_remove(manager->servers, server);
    config_file_write(manager->config, NULL);
    event_raise(manager, EVENT_SERVER_REMOVED, server);
    return TRUE;
}
