#include <string.h>
#include <gnet.h>
#include "user.h"
#include "config-file.h"

struct _User {
    gchar *name;
    Manager *manager;
    ConfigPath config_path;
    gchar *encrypted_password;
};

static void
user_hash_free(gpointer data) {
    User user = data;

    if (user->name != NULL) {
	g_free(user->name);
    }
    if (user->encrypted_password != NULL) {
	g_free(user->encrypted_password);
    }
    if (user->config_path != NULL) {
	config_path_free(user->config_path);
    }
    g_free(user);
}

static void
user_config_write(User u,
	gboolean flush) {
    ConfigPath users;

    users = config_path(u->manager->config, "users", NULL);
    if (users == NULL) {
	ConfigPath root = config_path(u->manager->config, NULL);

	users = config_path_create(u->manager->config, root, "users");
	config_path_free(root);
    }
    u->config_path = config_path_create(u->manager->config, users, "user");
    config_path_free(users);
    config_set(u->config_path, "name", u->name);
    config_set(u->config_path, "password", u->encrypted_password);
    if (flush) {
	config_file_write(u->manager->config, NULL);
    }
}

extern void
user_load(Manager *manager) {
    ConfigPath cp;
    GList *users, *l;
    User admin;

    g_assert(manager);
    g_assert(manager->config);

    manager->users = g_hash_table_new_full(g_str_hash, g_str_equal,
	    g_free, user_hash_free);
    cp = config_path(manager->config, "users", NULL);
    if (cp == NULL) {
	users = NULL;
    } else {
	users = config_children(cp);
	config_path_free(cp);
    }
    for (l = users; l != NULL; l = g_list_next(l)) {
	User user = g_new0(struct _User, 1);
	const gchar *str;

	cp = l->data;
	str = config_get_value(cp, "name");
	if (str == NULL) {
	    g_printerr("configuration file corrupt (anonymous user)\n");
	    g_free(user);
	    continue;
	}
	user->name = g_strdup(str);
	user->manager = manager;
	user->config_path = cp;

	str = config_get_value(cp, "password");
	if (str == NULL) {
	    user->encrypted_password = NULL;
	} else {
	    user->encrypted_password = g_strdup(str);
	}

	g_hash_table_insert(manager->users, g_strdup(user->name), user);
    }
    admin = g_hash_table_lookup(manager->users, "admin");
    if (admin == NULL) {
	admin = g_new0(struct _User, 1);

	admin->name = g_strdup("admin");
	admin->manager = manager;
	admin->config_path = NULL;
	admin->encrypted_password = NULL;
	user_config_write(admin, FALSE);
	g_hash_table_insert(manager->users, g_strdup("admin"), admin);
    }
}

extern gboolean
user_verify(Manager *manager,
	const gchar *user,
	const gchar *password) {
    User u;
    GSHA *sha;
    gchar *encrypted_password;

    g_assert(user);
    g_assert(password);

    u = g_hash_table_lookup(manager->users, user);
    if (u == NULL) {
	return FALSE;
    }
    if (u->encrypted_password == NULL) {
	return TRUE;
    }
    sha = gnet_sha_new(password, strlen(password));
    encrypted_password = gnet_sha_get_string(sha);
    gnet_sha_delete(sha);
    if (strcmp(encrypted_password, u->encrypted_password) == 0) {
	g_free(encrypted_password);
	return TRUE;
    }
    g_free(encrypted_password);
    return FALSE;
}

extern gboolean
user_change_password(Manager *manager,
	const gchar *user,
	const gchar *password) {
    User u;
    GSHA *sha;
    gchar *encrypted_password;

    g_assert(user);
    g_assert(password);

    u = g_hash_table_lookup(manager->users, user);
    if (u == NULL) {
	return FALSE;
    }
    sha = gnet_sha_new(password, strlen(password));
    encrypted_password = gnet_sha_get_string(sha);
    gnet_sha_delete(sha);
    g_free(u->encrypted_password);
    u->encrypted_password = encrypted_password;
    config_unset(u->config_path, "password");
    config_set(u->config_path, "password", encrypted_password);
    config_file_write(manager->config, NULL);
    return TRUE;
}

extern gboolean
user_add(Manager *manager,
	const SpecUser *spec,
	GError **error) {
    User u;
    GSHA *sha;

    g_assert(manager);
    g_assert(spec);
    g_assert(spec->user);
    g_assert(spec->password);

    u = g_hash_table_lookup(manager->users, spec->user);
    if (u != NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("user %s already exists"), u->name);
	return FALSE;
    }
    u = g_new0(struct _User, 1);
    u->name = g_strdup(spec->user);
    u->manager = manager;
    sha = gnet_sha_new(spec->password, strlen(spec->password));
    u->encrypted_password = gnet_sha_get_string(sha);
    gnet_sha_delete(sha);
    g_hash_table_insert(manager->users, g_strdup(u->name), u);
    user_config_write(u, TRUE);
    event_raise(manager, EVENT_USER_ADDED, u->name);
    return TRUE;
}

extern gboolean
user_edit(Manager *manager,
	const gchar *user,
	const SpecUser *spec,
	GError **error) {
    User u;
    gchar *old_name;
    gpointer on_p, u_p;
    gboolean changed = FALSE;
    gboolean name_changed = FALSE;

    g_assert(manager);
    g_assert(user);

    if (g_hash_table_lookup_extended(manager->users, user,
		&on_p, &u_p) == FALSE) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("user %s does not exists"), user);
	return FALSE;
    }
    old_name = on_p;
    u = u_p;
    if (strcmp(u->name, spec->user)) {
	name_changed = changed = TRUE;
	g_free(u->name);
	u->name = g_strdup(spec->user);
	config_unset(u->config_path, "name");
        config_set(u->config_path, "name", u->name);
    }
    if (changed) {
	config_file_write(manager->config, NULL);
    }
    if (name_changed) {
	g_hash_table_steal(manager->users, old_name);
	g_free(old_name);
	g_hash_table_insert(manager->users, g_strdup(u->name), u);
	event_raise(manager, EVENT_USER_RENAMED, user, u->name);
    }
    return TRUE;
}


extern gboolean
user_remove(Manager *manager,
	const gchar *user,
	GError **error) {
    User u;

    g_assert(manager);
    g_assert(user);

    u = g_hash_table_lookup(manager->users, user);
    if (u == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("user %s does not exist"), user);
	return FALSE;
    }
    config_delete(u->config_path);
    u->config_path = NULL;
    g_hash_table_remove(manager->users, user);
    config_file_write(manager->config, NULL);
    event_raise(manager, EVENT_USER_REMOVED, user);
    return TRUE;
}

extern SpecUser *
user_spec(Manager *manager,
	const gchar *user,
	GError **error) {
    User u;
    SpecUser *spec;

    g_assert(manager);
    g_assert(user);

    u = g_hash_table_lookup(manager->users, user);
    if (u == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("user %s does not exists"), user);
	return NULL;
    }
    spec = g_new0(struct _SpecUser, 1);
    spec->user = g_strdup(u->name);
    spec->password = g_strdup("******");
    return spec;
}

static void
user_list_fe(gpointer key,
	gpointer value,
	gpointer data) {
    User u = value;
    GList **l = data;

    *l = g_list_prepend(*l, g_strdup(u->name));
}

extern GList *
user_list(Manager *manager) {
    GList *l = NULL;

    g_hash_table_foreach(manager->users, user_list_fe, &l);
    return l;
}
