#include <string.h>
#include "svr.h"
#include "console.h"
#include "svr-message.h"
#include "terminal.h"
#include "ipmi.h"
#include "spec-console.h"

typedef struct _Console {
    ConsoleType type;
    union {
	Terminal *t;
	IPMI *i;
    };
    GIOChannel *io;
    MplexChannel chan;
    Server *server;
} Console;

static void
console_hash_free(gpointer data) {
    Console *c = data;
    GError *e = NULL;

    if (mplex_close(c->server->mplex, c->chan, &e) == FALSE) {
	g_printerr(_("failed to close mplex channel (%s)\n"), e->message);
	g_clear_error(&e);
    }
    switch (c->type) {
	case CONSOLE_TYPE_SERIAL:
	    terminal_free(c->t);
	    break;
	case CONSOLE_TYPE_IPMI:
	    ipmi_free(c->i);
	    break;
	default:
	    g_assert_not_reached();
    }
    g_free(c);
}

static void
console_closed(Mplex mplex,
	MplexChannel chan,
	gpointer data) {
    Console *c = data;

    g_printerr(_("console %s closed\n"), c->t->device);
    g_hash_table_remove(c->server->consoles, c->t->device);
}

static void
console_open(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Server *server = data;
    char *key;
    Terminal *terminal = NULL;
    IPMI *ipmi = NULL;
    MessageArg *a;
    GIOChannel *io = NULL;
    GError *e = NULL;
    MplexChannel chan;
    Console *c;
    ConsoleType type;

    g_assert(server);

    if (strcmp(name, SERVER_MESSAGE_OPEN_IPMI) == 0) {
	type = CONSOLE_TYPE_IPMI;
	ipmi = ipmi_new_from_arg(arg);
	g_assert(ipmi);
	key = ipmi->host;
	g_printerr("attempting to open sol console %s\n", key);
    } else {
	type = CONSOLE_TYPE_SERIAL;
	terminal = terminal_new_from_arg(arg);
	g_assert(terminal);
	key = terminal->device;
	g_printerr("attempting to open serial console %s\n", key);
    }
    if (g_hash_table_lookup(server->consoles, key) != NULL) {
	g_printerr("console %s already open\n", key);
	a = message_arg_error(1, "already open");
	if (message_respond(q, id, a, &e) == FALSE) {
	    g_printerr(_("failed to send reponse (%s)\n"), e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
	return;
    }
    switch (type) {
	case CONSOLE_TYPE_SERIAL:
	    io = terminal_open_io_channel(terminal, &e);
	    break;
	case CONSOLE_TYPE_IPMI:
	    io = ipmi_open_io_channel(ipmi, &e);
	    break;
	default:
	    g_assert_not_reached();
    }
    if (io == NULL) {
	g_printerr(_("failed to open console %s\n"), key);
	a = message_arg_error(1, "%s", e->message);
	g_clear_error(&e);
	if (message_respond(q, id, a, &e) == FALSE) {
	    g_printerr(_("failed to send reponse (%s)\n"), e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
	return;
    }
    c = g_new0(struct _Console, 1);
    if (mplex_connect(server->mplex, 0, &chan, io, console_closed, c, &e)
	    == FALSE) {
	g_printerr(_("failed to connect console to mplex (%s)\n"), e->message);
	g_clear_error(&e);
	g_io_channel_unref(io);
	g_free(c);
	return;
    }
    c->type = type;
    switch (type) {
	case CONSOLE_TYPE_SERIAL:
	    c->t = terminal;
	    break;
	case CONSOLE_TYPE_IPMI:
	    c->i = ipmi;
	    break;
	default:
	    g_assert_not_reached();
    }
    c->io = io;
    c->chan = chan;
    c->server = server;
    g_hash_table_insert(server->consoles, g_strdup(key), c);
    a = message_arg_uint16("channel", chan);
    if (message_respond(q, id, a, &e) == FALSE) {
	g_printerr(_("failed to send reponse (%s)\n"), e->message);
	g_clear_error(&e);
    }
}

extern void
console_register_message_handlers(Server *server) {
    GError *e = NULL;

    g_assert(server);

    if (server->consoles == NULL) {
	server->consoles = g_hash_table_new_full(g_str_hash, g_str_equal,
		g_free, console_hash_free);
    }
    if (server->msg_q == NULL) {
	return;
    }
    if (message_handler_register(server->msg_q,
		SERVER_MESSAGE_OPEN_CONSOLE, console_open,
		server, &e) == FALSE) {
	g_printerr(_("could not register handler %s (%s)\n"),
		SERVER_MESSAGE_OPEN_CONSOLE, e->message);
    }
    if (message_handler_register(server->msg_q,
		SERVER_MESSAGE_OPEN_IPMI, console_open,
		server, &e) == FALSE) {
	g_printerr(_("could not register handler %s (%s)\n"),
		SERVER_MESSAGE_OPEN_IPMI, e->message);
    }
}

