#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "svr.h"
#include "logquery.h"
#include "log-query.h"
#include "svr-message.h"
#include "date.h"

typedef enum {
    QS_HUNT_LINESEEK,
    QS_HUNT_LINEREAD,
    QS_READ_LINESEEK,
    QS_READ_LINEREAD,
} QueryState;

typedef struct _Query {
    guint query_id;
    LogQuery *lq;
    GIOChannel *io; /* io channel to log file */
    GIOChannel *rio; /* io channel for query results */
    MplexChannel chan;
    Server *server;
    QueryState state;
    guint line_count;
    gint64 hunt_start;
    gint64 hunt_end;
    gint64 pos;
    gint64 line_pos;
    gint64 line_end;
    GString *buf;
    guint read_source_id;
    guint write_source_id;
    gsize target_length;
} Query;

static void logquery_start_read(Query *q, guint length);
static void logquery_hunt(Query *q);
static void logquery_read(Query *q);

static void
logquery_tree_free(gpointer data) {
    Query *q = data;

    log_query_free(q->lq);
    if (q->read_source_id > 0) {
	g_source_remove(q->read_source_id);
    }
    if (q->write_source_id > 0) {
	g_source_remove(q->write_source_id);
    }
    g_io_channel_unref(q->io);
    g_io_channel_unref(q->rio);
    mplex_close(q->server->mplex, q->chan, NULL);
    if (q->buf != NULL) {
	g_string_free(q->buf, TRUE);
    }
    /* FIXME: send message to manger? */
    g_free(q);
}

static void
logquery_done(Query *q) {
    g_tree_remove(q->server->logqueries, GINT_TO_POINTER(q->query_id));
}

static gboolean
logquery_writer(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    Query *q = data;
    gsize count;
    gsize bytes_sent = 0;
    GString *s;

    g_assert(q);
    g_assert(q->buf);
    count = q->buf->len;
    if (count > 1024) {
	count = 1024;
    }
    if (count > 0) {
	g_io_channel_write_chars(source, q->buf->str, count, &bytes_sent, NULL);
    }
    if (bytes_sent == 0) {
	logquery_read(q);
	q->write_source_id = 0;
	return FALSE;
    }
    s = q->buf;
    q->buf = g_string_new_len(&s->str[bytes_sent], s->len - bytes_sent);
    g_string_free(s, TRUE);
    return TRUE;
}

static gboolean
logquery_reader(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    Query *q = data;
    gchar buf[1024];
    gsize count;
    gsize bytes_read = 0;
    Date line_time;
    guint dateend;

    g_assert(q);
    count = q->target_length - q->buf->len;
    if (count >= 1024) {
	count = 1024;
    }
    g_io_channel_read_chars(source, buf, count, &bytes_read, NULL);
    q->pos += bytes_read;
    if (bytes_read > 0) {
	q->buf = g_string_append_len(q->buf, buf, bytes_read);
	if (q->buf->len < q->target_length) {
	    return TRUE;
	}
    }
    q->read_source_id = 0;
    switch (q->state) {
	case QS_HUNT_LINESEEK:
	    if (q->buf->len == 0 || q->buf->str[0] == '\n') {
		q->line_pos = q->pos;
		q->state = QS_HUNT_LINEREAD;
		logquery_start_read(q, 1024);
	    } else {
		q->pos -=  2;
		if (q->pos < 0) {
		    q->pos = 0;
		    g_io_channel_seek_position(q->io, q->pos, G_SEEK_SET, NULL);
		    q->line_pos = q->pos;
		    q->state = QS_HUNT_LINEREAD;
		    logquery_start_read(q, 1024);
		} else {
		    g_io_channel_seek_position(q->io, q->pos, G_SEEK_SET, NULL);
		    logquery_start_read(q, 1);
		}
	    }
	    break;
	case QS_HUNT_LINEREAD:
	    line_time = date_read(q->buf->str, q->buf->len, NULL);
	    if (line_time == 0) {
		q->hunt_end = q->hunt_start;
	    } else {
		if (line_time > q->lq->end_time) {
		    q->hunt_end = ((q->hunt_end - q->hunt_start) / 2)
			+ q->hunt_start;
		} else {
		    q->hunt_start = ((q->hunt_end - q->hunt_start) / 2)
			+ q->hunt_start;
		}
	    }
	    logquery_hunt(q);
	    break;
	case QS_READ_LINESEEK:
	    if (q->buf->len == 0 || q->buf->str[0] == '\n') {
		q->line_pos = q->pos;
		q->state = QS_READ_LINEREAD;
		logquery_start_read(q, q->line_end - q->pos + 1);
	    } else {
		if (q->pos < 2) {
		    logquery_done(q);
		} else {
		    q->pos -=  2;
		    g_io_channel_seek_position(q->io, q->pos, G_SEEK_SET, NULL);
		    logquery_start_read(q, 1);
		}
	    }
	    break;
	case QS_READ_LINEREAD:
	    line_time = date_read(q->buf->str, q->buf->len, &dateend);
	    if (line_time == 0) {
		    logquery_done(q);
	    } else {
		if (line_time < q->lq->start_time) {
		    logquery_done(q);
		} else {
		    q->write_source_id = g_io_add_watch_full(q->rio,
			    G_PRIORITY_LOW, G_IO_OUT, logquery_writer, q, NULL);
		}
	    }
	    break;
    }
    return FALSE;
}

static void
logquery_start_read(Query *q,
	guint length) {
    if (q->buf != NULL) {
	q->buf = g_string_assign(q->buf, "");
    } else {
	q->buf = g_string_new("");
    }
    q->target_length = length;
    q->read_source_id = g_io_add_watch_full(q->io, G_PRIORITY_LOW, G_IO_IN,
	    logquery_reader, q, NULL);
}

static void
logquery_read(Query *q) {
    q->line_count++;
    if (q->lq->max_lines > 0 &&  q->line_count > q->lq->max_lines) {
	logquery_done(q);
	return;
    }
    q->state = QS_READ_LINESEEK;
    if (q->pos < 2) {
	logquery_done(q);
	return;
    }
    q->line_end = q->line_pos - 1;
    q->pos = q->line_pos - 2;
    g_io_channel_seek_position(q->io, q->pos, G_SEEK_SET, NULL);
    logquery_start_read(q,1);
}

static void
logquery_hunt(Query *q) {
    q->state = QS_HUNT_LINESEEK;
    if (q->hunt_end - q->hunt_start < 16) {
	logquery_read(q);
	return;
    }
    q->pos = ((q->hunt_end - q->hunt_start) / 2) + q->hunt_start;
    g_io_channel_seek_position(q->io, q->pos, G_SEEK_SET, NULL);
    if (q->pos == 0) {
	logquery_read(q);
	return;
    }
    logquery_start_read(q,1);
}

static void
logquery_new(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    Query *query;
    Server *server = data;
    LogQuery *lq;
    MessageArg *a;
    GIOChannel *io;
    GIOChannel *rio;
    GError *e = NULL;
    MplexChannel chan;
    static gint query_id = 0;
    struct stat buf;

    g_assert(server);
    lq = log_query_new_from_arg(arg);
    g_assert(lq);
    io = g_io_channel_new_file(lq->log, "r", &e);
    if (io == NULL) {
	int errcode = 1;

	g_printerr(_("failed to open log file %s\n"), lq->log);
	if (g_error_matches(e, G_FILE_ERROR, G_FILE_ERROR_EXIST) == TRUE) {
	    errcode = 0;
	}
	a = message_arg_error(errcode, "%s", e->message);
	g_clear_error(&e);
	if (message_respond(q, id, a, &e) == FALSE) {
	    g_printerr(_("failed to send reponse (%s)\n"), e->message);
	    g_clear_error(&e);
	}
	message_arg_free(a);
	log_query_free(lq);
	return;
    }
    rio = mplex_open(server->mplex, 0, &chan, NULL, NULL, &e);
    if (rio == NULL) {
	g_printerr(_("failed to open mplex channel (%s)\n"), e->message);
	/* FIXME: send response? */
	g_io_channel_unref(io);
	g_clear_error(&e);
	log_query_free(lq);
	return;
    }
    query = g_new0(Query, 1);
    query->lq = lq;
    query->io = io;
    query->rio = rio;
    query->chan = chan;
    query->server = server;
    query->hunt_start = 0;
    stat(lq->log, &buf);
    query->hunt_end = buf.st_size;
    query->buf = NULL;
    query->query_id = query_id++;
    g_tree_insert(server->logqueries, GINT_TO_POINTER(query->query_id), query);
    a = message_arg_uint16("channel", chan);
    if (message_respond(q, id, a, &e) == FALSE) {
	g_printerr(_("failed to send reponse (%s)\n"), e->message);
	g_clear_error(&e);
    }
    message_arg_free(a);
    logquery_hunt(query);
}

extern void
logquery_register_message_handlers(Server *server) {
    GError *e = NULL;

    g_assert(server);

    if (server->msg_q == NULL) {
	return;
    }
    if (message_handler_register(server->msg_q, SERVER_MESSAGE_QUERY_LOG,
		logquery_new, server, &e) == FALSE) {
	g_printerr(_("could not register handler %s (%s)\n"),
		SERVER_MESSAGE_QUERY_LOG, e->message);
    }
}

static gint
logquery_compare(gint a,
	gint b) {
    return b-a;
}

extern GTree *
logquery_tree_init(void) {
    return g_tree_new_full((GCompareDataFunc)logquery_compare, NULL, NULL,
	    logquery_tree_free);
}
