#include <stdlib.h>
#include <unistd.h>
#include "config.h"
#include "svr.h"
#include "console.h"
#include "log.h"
#include "logquery.h"
#include "stdinout.h"
#include "ping.h"

static void
end_service(Mplex mplex, gpointer data) {
    Server *server = data;

    g_printerr("manager closed connection\n");
    if (server->msg_q != NULL) {
	message_queue_free(server->msg_q);
	server->msg_q = NULL;
    }
    if (server->mplex != NULL) {
	mplex_free(server->mplex);
	server->mplex = NULL;
    }
    if (server->accept_sock != NULL) {
	gnet_tcp_socket_delete(server->accept_sock);
	server->accept_sock = NULL;
	server->chan = NULL;
    } else {
	if (server->chan != NULL) {
	    g_io_channel_unref(server->chan);
	    server->chan = NULL;
	}
    }
    if (server->listen_sock == NULL) {
	g_main_loop_quit(server->main_loop);
    }
}

static void
start_service(Server *server) {
    GError *e = NULL;
    GIOChannel *io;

    server->mplex = mplex_new(server->chan, end_service, server, &e);
    if (e != NULL) {
	g_printerr("cannot start mplex! (%s)\n", e->message);
	g_clear_error(&e);
	g_io_channel_unref(server->chan);
	server->chan = NULL;
	if (server->accept_sock) {
	    gnet_tcp_socket_unref(server->accept_sock);
	    server->accept_sock = NULL;
	}
	return;
    }
    io = mplex_open(server->mplex, CONC_MSG_CHANNEL, NULL, NULL, NULL, &e);
    if (e != NULL) {
	g_printerr("cannot open message channel! (%s)\n", e->message);
	g_clear_error(&e);
	mplex_free(server->mplex);
	g_io_channel_unref(server->chan);
	server->chan = NULL;
	if (server->accept_sock) {
	    gnet_tcp_socket_unref(server->accept_sock);
	    server->accept_sock = NULL;
	}
	return;
    }
    server->msg_q = message_queue_new(io, &e);
    if (e != NULL) {
	g_printerr("cannot open message queue! (%s)\n", e->message);
	g_clear_error(&e);
	mplex_free(server->mplex);
	g_io_channel_unref(server->chan);
	server->chan = NULL;
	if (server->accept_sock) {
	    gnet_tcp_socket_unref(server->accept_sock);
	    server->accept_sock = NULL;
	}
	return;
    }
    server->logqueries = logquery_tree_init();
    console_register_message_handlers(server);
    log_register_message_handlers(server);
    logquery_register_message_handlers(server);
    ping_register_message_handlers(server);
    message_queue_start(server->msg_q);
    mplex_start(server->mplex);
    return;
}

static void
accept_con(GTcpSocket *gserver,
	GTcpSocket *sock,
	gpointer user_data) {
    Server *server = user_data;

    if (sock == NULL) {
	/* FIXME: there was an error.
	 * server->accept_sock will need to be set up again.
	 */
	return;
    }
    if (server->chan != NULL) {
	GInetAddr *addr = gnet_tcp_socket_get_remote_inetaddr(sock);

	g_printerr("warning: an attempt make to connect by %s when server"
	    " already has an active connection!\n",
	    gnet_inetaddr_get_name(addr));
	gnet_inetaddr_unref(addr);
	gnet_tcp_socket_unref(sock);
	return;
    }
    server->accept_sock = sock;
    server->chan = gnet_tcp_socket_get_io_channel(server->accept_sock);
    start_service(server);
    g_printerr("manager connects\n");
}

int
main(int argc, char *argv[]) {
    Server *server = g_new0(Server, 1);
    gint debug = 0, mode_daemon = 0, mode_stdin = 0;
    GOptionEntry options[] = {
	{ "debug", 'd', 0, G_OPTION_ARG_NONE, &debug,
	    N_("enter debug mode"), NULL },
	{ "daemon", 'l', 0, G_OPTION_ARG_NONE, &mode_daemon,
	    N_("run as a daemon (listening for connection from manager)"),
	    NULL },
	{ "pipe", 'p', 0, G_OPTION_ARG_NONE, &mode_stdin,
	    N_("communicate with manager using stdin/out"), NULL },
	{ NULL }
    };
    GOptionContext *option_context;

    chdir(CONCLOGDIR);
    option_context = g_option_context_new(NULL);
    g_option_context_add_main_entries(option_context, options,
	GETTEXT_PACKAGE);
    server->program = gnome_program_init("concsvr", PACKAGE_VERSION,
	LIBGNOME_MODULE, argc, argv,
	GNOME_PARAM_GOPTION_CONTEXT, option_context,
	GNOME_PARAM_NONE);
    if (debug == 1) {
	g_printerr("entering debug mode\n");
	g_log_set_always_fatal(G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL |
	    G_LOG_LEVEL_WARNING);
    }
    server->main_loop = g_main_loop_new(NULL, FALSE);
    server->mode = SERVER_MODE_DEFAULT;
    if (mode_daemon == 1) {
	server->mode = SERVER_MODE_DAEMON;
    }
    if (mode_stdin == 1) {
	server->mode = SERVER_MODE_STDIN;
    }
    if (server->mode == SERVER_MODE_DEFAULT) {
	if (isatty(0) == 1) {
	    server->mode = SERVER_MODE_DAEMON;
	} else {
	    server->mode = SERVER_MODE_STDIN;
	}
    }
    if (server->mode == SERVER_MODE_DAEMON) {
	gnet_init();
	server->listen_sock =
	    gnet_tcp_socket_server_new_with_port(CONC_PORT_SVR);
	server->accept_sock = NULL;
	if (server->listen_sock == NULL) {
	    g_printerr("error: cannot create server socket\n");
	}
    } else {
	server->listen_sock = NULL;
	server->accept_sock = NULL;
	server->chan = stdinout_new();
	start_service(server);
    }
    if (server->listen_sock != NULL) {
	gnet_tcp_socket_server_accept_async(server->listen_sock,
		accept_con, server);
    }
    g_main_loop_run(server->main_loop);
    exit(EXIT_SUCCESS);
}
